#ifdef USE_ARTE
// *****************************************************************************
// 
//  Author:  Michael Schmelling                Created:  27-Oct-00
//           MPI-Heidelberg                    Modified: 31-Oct-00
//  Email:   Michael.Schmelling@mpi-hd.mpg.de  HERA-B Collaboration
//
// anl_evselect.C
// basic event selection function
// based on earlier work by Michael Dressel and Wouter Hulsbergen 
// Events are required to have a certain minimum number of VDS segments
// found, at least one candidate primary vertex with 3 tracks, to be
// from a nominally filled BX and to have proper timing in the OTR. 
//
// changes:
// 30 Oct 2000 (MS) introduced enhanced monitoring functionality
// 31 Oct 2000 (MS) speed up and improved primary vertex estimate
//
// *****************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <iostream.h>
#include <vector.h>
#include "arte/arte.hh"
#include "analysis/anl_evselect.hh"

// use locally define constants to describe the pattern of empty bunches

#define WRONG_BX  \
(BX==4||BX==15||BX==26||BX==37||(BX>=48&&BX<=52)||BX==63||BX==74||\
BX==85||BX==96||BX==107||(BX>=118&&BX<=122)||BX==133||BX==144||\
BX==155||BX==166||BX==177||(BX>=188&&BX<=202)||BX==213)


int anl_evselect(int target, anl_evselect_info & info) 
/*=====================================================================
 * basic event selection function
 *===================================================================*/
{
  // initialization at first call

  static int    FIRST = 1;  
  static double s0w, s1w, s2w, s0a, s1a, s2a;
  static double nevtot, facc, freject, fwrongbx, fvdstracks, fnovertex, foutoftime; 

  // set the event selection parameters

  info.evselectflag = 0;    // initialize the event selection flag
  info.BXOFFSET     = 1;    // offset between BX count and bunch pattern
  info.MINHITS      = 6;    // mimimum number of VDS hits per track
  info.NTRACKMIN    = 3;    // minumum number of VDS tracks required
  info.DVERTEX      = 0.1;  // maximum distance between tracks in cand. vtx
  info.OTTMEAN_MIN  = 165.; // minimum of average drift time
  info.OTTMEAN_MAX  = 185.; // maximum of average drift time
  info.OTTWIDTH_MAX =  50.; // maximum rms width of drift time distribution

  // auxiliary variable

  info.D2VERTEX = info.DVERTEX*info.DVERTEX;

  // print parameter settings

  if(FIRST) {
    FIRST  = 0;
    s0w    = s1w  = s2w = s0a = s1a = s2a = 0;
    nevtot = facc = freject = fwrongbx = fvdstracks = fnovertex = foutoftime = 0; 
    cout << " === INFO: >>>>\n";
    cout << " === INFO: anl_evselect " 
         << xcvsVERSION_evselect << " (" << xcvsDAYTIME_evselect << " UTC)\n";
    cout << " === INFO: minimum number of hits/VDS-track        : " 
         << info.MINHITS      << endl; 
    cout << " === INFO: minimum number of VDS-tracks            : " 
         << info.NTRACKMIN    << endl; 
    cout << " === INFO: maximum xy-distance in vertex candidate : " 
         << info.DVERTEX      << endl; 
    cout << " === INFO: minimum value for average OTR-drifttime : " 
         << info.OTTMEAN_MIN  << endl; 
    cout << " === INFO: maximum value for average OTR-drifttime : " 
         << info.OTTMEAN_MAX  << endl; 
    cout << " === INFO: maximum value for rms OTR-drifttime     : " 
         << info.OTTWIDTH_MAX << endl; 
    cout << " === INFO: <<<<\n";
  }


  // ============================================================
  // get a list of non-clone rtras
  // ============================================================

  vector< ArtePointer<RTRA> > vrtra;  

  for( ArteTable<RTRA>::iterator rtra = ArteTable<RTRA>::begin(); 
       rtra != ArteTable<RTRA>::end(); ++rtra ){
    if( (*rtra)->hsi < info.MINHITS) continue;
    int OK = 1; 
    for(unsigned int i=0; i!=vrtra.size(); ++i) {
      if( (*rtra)->csvd == (vrtra[i])->csvd ) {
	OK = 0;
	break;
      }
    }      
    if(OK) vrtra.push_back(*rtra);
  }
  info.ntrvds = vrtra.size();

  if( info.ntrvds < info.NTRACKMIN ) info.evselectflag |= EVSEL_VDSTRACKS;

  // ============================================================
  // BX selection and calculation of coasting beam index
  // ============================================================
 
  ArtePointer<DIHD> dihd(1);
  info.bx = dihd->bx;
  int BX  = (info.bx + info.BXOFFSET)%220;

  if( WRONG_BX ) {
    info.evselectflag |= EVSEL_WRONGBX;
    s0w += 1.;
    s1w += double(info.ntrvds);
    s2w += double(info.ntrvds*info.ntrvds);
  }
  else {
    s0a += 1.;
    s1a += double(info.ntrvds);
    s2a += double(info.ntrvds*info.ntrvds);
  }

  double avw = 0;
  double sdw = 0;
  if(s0w > 0) {
   avw = s1w/s0w;
   sdw = sqrt(fabs(s2w/s0w - (s1w/s0w)*(s1w/s0w))/s0w);
  }
  double avr = 0;
  double sdr = 0;
  if(s0a > 0) {
   avr = s1a/s0a;
   sdr = sqrt(fabs(s2a/s0a - (s1a/s0a)*(s1a/s0a))/s0a);
  }
  info.cb  = 0;
  info.dcb = 0;
  if(s0w>0 && s0a>0 && avw>0 && avr>0) {
    info.cb  = float(avw/avr);
    info.dcb = info.cb*float(sqrt((sdw/avw)*(sdw/avw)+ (sdr/avr)*(sdr/avr)));
  }

  // ============================================================
  // check for a candidate primary vertex
  // ============================================================
  
  // collect z-positions of target wires

  vector< float > vztar; 

  for( ArteTableIterator<GTAR> gtar=ArteTable<GTAR>::begin(); 
       gtar!=ArteTable<GTAR>::end(); ++gtar) {
    char *name = (char*)&((*gtar)->name);
    if((strncmp(name,"1 a",3)==0)&&(target&EVSEL_ABOVE1)) vztar.push_back((*gtar)->z);
    if((strncmp(name,"1 b",3)==0)&&(target&EVSEL_BELOW1)) vztar.push_back((*gtar)->z);
    if((strncmp(name,"1 i",3)==0)&&(target&EVSEL_INNER1)) vztar.push_back((*gtar)->z);
    if((strncmp(name,"1 o",3)==0)&&(target&EVSEL_OUTER1)) vztar.push_back((*gtar)->z);
    if((strncmp(name,"2 a",3)==0)&&(target&EVSEL_ABOVE2)) vztar.push_back((*gtar)->z);
    if((strncmp(name,"2 b",3)==0)&&(target&EVSEL_BELOW2)) vztar.push_back((*gtar)->z);
    if((strncmp(name,"2 i",3)==0)&&(target&EVSEL_INNER2)) vztar.push_back((*gtar)->z);
    if((strncmp(name,"2 o",3)==0)&&(target&EVSEL_OUTER2)) vztar.push_back((*gtar)->z);
  }

  // look for 3-track vertex candidate

  info.d2vertex = info.D2VERTEX + 1.;

  if( vrtra.size() > 2) {
    for(unsigned int l=0; l!=vztar.size(); ++l) {
      for(unsigned int i=2; i!=vrtra.size(); ++i) {
	float xi = vrtra[i]->xf + (vztar[l] - vrtra[i]->zf) * (vrtra[i]->txf);
	float yi = vrtra[i]->yf + (vztar[l] - vrtra[i]->zf) * (vrtra[i]->tyf);
	for(unsigned int j=1; j!=i; ++j) {
	  float xj   = vrtra[j]->xf + (vztar[l] - vrtra[j]->zf) * (vrtra[j]->txf);
	  float yj   = vrtra[j]->yf + (vztar[l] - vrtra[j]->zf) * (vrtra[j]->tyf);
	  float d2ij = (xi-xj)*(xi-xj) + (yi-yj)*(yi-yj);
	  if(d2ij >= info.D2VERTEX) continue; 
	  for(unsigned int k=0; k!=j; ++k) {
	    float xk   = vrtra[k]->xf + (vztar[l] - vrtra[k]->zf) * (vrtra[k]->txf);
	    float yk   = vrtra[k]->yf + (vztar[l] - vrtra[k]->zf) * (vrtra[k]->tyf);
	    float d2ik = (xi-xk)*(xi-xk) + (yi-yk)*(yi-yk);
	    if(d2ik >= info.D2VERTEX) continue;  
	    float d2jk = (xj-xk)*(xj-xk) + (yj-yk)*(yj-yk);
            if(d2jk >= info.D2VERTEX) continue; 
            float d2max = d2ij;
            if( d2ik > d2max) d2max = d2ik;
            if( d2jk > d2max) d2max = d2jk;
	    if(d2max < info.d2vertex) {
              info.d2vertex = d2max;
	      info.xver     = (xi+xj+xk)/3.;
	      info.yver     = (yi+yj+yk)/3.;
	      info.zver     = vztar[l];
	    }            
          }
        }
      }
    }
  }
  if(info.d2vertex >= info.D2VERTEX) info.evselectflag |= EVSEL_NOVERTEX;
 
  info.dvertex = sqrt(info.d2vertex);

  // ============================================================
  // OTR selection based on drifttime distribution
  // ============================================================
 
  if(!ArteTable<MCIN>::booked()) {
    ArtePointer<EVRC> evrc( ArteTable<EVRC>::size() ) ;
    info.ottmean  = evrc->ottmean;
    info.ottwidth = evrc->ottwidth;
    if( (info.ottmean  < info.OTTMEAN_MIN) || 
        (info.ottmean  > info.OTTMEAN_MAX) ||
	(info.ottwidth > info.OTTWIDTH_MAX)  )  info.evselectflag |= EVSEL_OUTOFTIME; 
  }

  // do some statistics and return

  nevtot += 1.;
  if(info.evselectflag == 0) facc    += 1.;
  else                       freject += 1.;
  if(info.evselectflag & EVSEL_WRONGBX  ) fwrongbx   += 1.;
  if(info.evselectflag & EVSEL_VDSTRACKS) fvdstracks += 1.;
  if(info.evselectflag & EVSEL_NOVERTEX ) fnovertex  += 1.;
  if(info.evselectflag & EVSEL_OUTOFTIME) foutoftime += 1.;

  info.facc       = float(facc/nevtot);
  info.freject    = float(freject/nevtot);
  info.fwrongbx   = float(fwrongbx/nevtot);
  info.fvdstracks = float(fvdstracks/nevtot);
  info.fnovertex  = float(fnovertex/nevtot);
  info.foutoftime = float(foutoftime/nevtot);
  info.nevtot     = int(nevtot);

  return info.evselectflag;
}

#undef WRONG_BX  
#endif








