/************************************************************************
* This file contains the routines which implement the cell model for
* the outer tracker digitization.
*
* All routines called from dgomod
*
* Written by                                 Wouter Hulsbergen  June 2000
* From original code written by S.Nowak, C.Stegmann, A.Lanyov
*
* Modified:
* - Addition of routines for generation of cross-talk and noise 
*                                            Wouter    November 2000
* - Added `rt-calibration' for MC rtrelation           April 2001
* - Removed bug in routine that calculates pathlength     "
* - Added default arguments for Estimatedradius/Sigma for slt simu.
* - Fixed buf in SetEfficiency/GausSmearing (Thanks to Denis)
************************************************************************/

/************************************************************************
* The application of the cell model in the outer tracker digitization
* exists of the following steps:
*
*  Given a MIMP and a module
*   1) calculate distance $dd$ to wires in module that are `close' to track
*   2) check whether channel exists and is good (level >=3 only)
*   3) calculate length $l$ of segment of track through cell 
*   4) generate cluster on segment closest to the distance of closest
*      approach to wire and check whether closer than $0.5*l$
*   5) calculate distance of closest cluster to wire, convert distance
*      to drifttime and apply gaussian smearing. This is the measurement.
*
*  After processing all MIMPs in a module
*   6) create cross-talk and noise hits
*
*  After creating all hits:
*   7) create hitb in cell corresponding to smallest measured time.
*   8) convert measured time to expectation value for distance 
*     (hitb->q) and for the sqrt(variance) of the distance (hitb->r)
*  
*  Step 8 is new with repect to older versions of the digitization. It
*  mimics the effect of the calibration as applied to real data.
*
*  Wouter Hulsbergen (June 2000)
*
* ---------------------------------------
*
* With release of April 2001 
*
* Although the cell model is not very complicated, it is not
* straightforward to obtain the distance and error for a certain hit
* after the hit has been `digitized'. The value of hitb->q and hitb->r
* should be filled such that the pool ditribution
*     (hitb->q - `true MC distance')/hitb->r
* is centered at zero and has unity width for all values of a certain
* digitized time `tdc'. The corresponding r(t) and sigma(t) relation
* with which hitb->q and hitb->r are filled, depend on the parameters
* used in the digitization. Up to now some quasi-analytic formula was
* used to take this dependence into account. This approach is not very
* flexible: If something is added to the cell mode, the formulas will
* have to be adjusted. As of this release, a new approach has been
* implemented: The r(t) and sigma(t) relation are calibrated in the
* initialization step by generating hits and comparing to the monte
* carlo truth.
* 
************************************************************************/

#include "digi/dgocell.hh"
#include <math.h>
#include <iostream>

#ifdef USE_ARTE
#include <arte/arte.hh>
#endif

// This is not really an example of nice C++ programming but for now
// it should just work. 

extern "C" 
{
  void ranmar_(float[], const int &);
  void grndm_(float [], const int &);
  void granor_(float &, float &);
}

namespace dgocell
{
  // default parameters of the model
  float gPrimIonDensity = 20 ;      // prim. ion. density [1/cm]
  float gAttachDensity  = 0.0 ;     // attachment density [1/cm]
  float gGausSmearing      = -1 ;   // gaussian smearing       (default taken from gede->acc)
  float gNominalEfficiency = -1 ;   // max of cell-eff plateau (default taken from gede->eff)
  float gDriftVelocity = 0.01 ;     // [cm/ns]
  bool gIsInitialized = false ;
  float gTdcGain      = -256/96. ;  // [count/ns]
  float gTdcOffset    = 210 ;       // [count]
  float gXTalkProbability = 0. ;       // 0.05  from Ilija
  float gRandomNoiseProbability = 0. ; // 0.003 from empty events from run14577
  float *gRTRelation05mm=0;
  float *gSTRelation05mm=0;
  float *gRTRelation10mm=0;
  float *gSTRelation10mm=0;

  void Print()
  { 
    std::cout << "**************************************************" << std::endl 
	      << "* Initializing otr digitization                  *" << std::endl
	      << "*                                                *" << std::endl
	      << "* Model parameters:                              *" << std::endl
	      << "*  primary ion density [1/cm]: " << gPrimIonDensity << std::endl
	      << "*  attachment density [1/cm]:  " << gAttachDensity << std::endl 
	      << "*  nominal efficiency :        " << gNominalEfficiency << std::endl
	      << "*  gaussian smearing :         " << gGausSmearing << std::endl
	      << "*  drift velocity [cm/ns] :    " << gDriftVelocity << std::endl
	      << "*  tdc gain [count/ns] :       " << gTdcGain << std::endl
	      << "*  tdc offset [count] :        " << gTdcOffset << std::endl
	      << "*  x talk:                     " << gXTalkProbability << std::endl
	      << "*  random noise:               " << gRandomNoiseProbability << std::endl
	      << "* Model output (averaged over modules):          *" << std::endl 
	      << "*  total efficiency 5 mm (at wire) : " 
	      << gNominalEfficiency * (1-exp(-gPrimIonDensity*0.5)) << std::endl
	      << "*  total efficiency 10 mm (at wire) : " 
	      << gNominalEfficiency * (1-exp(-gPrimIonDensity*1.0)) << std::endl ;
    float tdc = gTdcOffset + 0.1*gTdcGain/gDriftVelocity ;
    cout      << "*  total resolution 5 mm (at 1 mm) : " 
	      << EstimatedSigma(tdc,0.87,gGausSmearing) << std::endl
	      << "*  total resolution 10 mm (at 1 mm) : " 
	      << EstimatedSigma(tdc,2*0.87,gGausSmearing) << std::endl
	      << "**************************************************" << std::endl ;
  }
  
  void ForceInitialization() {
    gIsInitialized=false ;
  }

  void Initialize()
  {
    gIsInitialized=true ;
    double acc2sum(0),effsum(0) ;
    unsigned int n(0) ;  
    // this looks a bit clumsy but the idea is: if
    // gNominalEfficiency/gGausSmearing>0, they are applied to all
    // entries in gede bank; otherwise they are set to the average of
    // the values that are currently in the gede bank.
    for(ArteTable<GEDE>::iterator gedeIt = ArteTable<GEDE>::begin() ;
	gedeIt != ArteTable<GEDE>::end() ; ++gedeIt) 
      if(gedeIt->cmp==2 && gedeIt->sec>0) { 
	if(gNominalEfficiency>0) gedeIt->eff = gNominalEfficiency ;
	if(gGausSmearing>0)      gedeIt->acc = gGausSmearing ;
	acc2sum += gedeIt->acc*gedeIt->acc ;
	effsum += gedeIt->eff ;
	n++;
      }
    if(gNominalEfficiency<=0) gNominalEfficiency = effsum/n;
    if(gGausSmearing<=0)      gGausSmearing = sqrt(acc2sum/n) ;
    CalibrateRT() ;
    Print() ;
   }


  float PathLengthHC(float pitch, float x, float tx, float ty)
  {
    // This routine calculates the path lenth of a track trough a
    // honeycomb cell.
    // pitch: the wire pitch as in gede->pitch
    // x:  (x_track - x_wire) in plane z_wire
    // tx: tx_track in system of wire (rotated)
    // ty: ty_track in system of wire (rotated)
    
    // The coordinates normalized to the pitch are called xp and zp.
    // These are the coordinates of the 6 sides, normalized to the pitch:
    static const float xp[6] = {-1/3.,-1/6.,1/6.,1/3.,1/6.,-1/6.} ;
#define REC2SQRT3 1/(2*sqrt(3))
    static const float zp[6] = {0,REC2SQRT3,REC2SQRT3,0,-REC2SQRT3,-REC2SQRT3} ;
#undef REC2SQRT3
    
    // Loop over all sides and find points of intersection
    const float xptr = x/pitch; // the xp-coordinate of the track
    float xpent[2];             // the xp-coordinates of the intersections
    float zpent[2];             // the zp-coordinate of the intersections
    int n = 0;                  // counter for the intersections
    for(int i1=0; i1<6; i1++) {
      int i2=(i1+1)%6 ;
      // This is zp,xp of point of intersect
      float zis = ((zp[i2]-zp[i1])*xptr + xp[i2]*zp[i1] - xp[i1]*zp[i2]) /
	( (xp[i2]-xp[i1]) - tx* (zp[i2]-zp[i1]) ) ;
      float xis = xptr + zis*tx ;
      // Is this between boundaries ?
      if( (xis - xp[i1])*(xp[i2] - xis) >0 ) {
	if(n<=1) {
	  xpent[n] = xis ;
	  zpent[n] = zis ;
	}
	n++;
      }
    }
      
    // Calculate the path length from the intersections
    float len=0;
    if(n==2) {
      float dz = pitch * fabs(zpent[1]-zpent[0]) ;
      len = dz * sqrt(1 + tx*tx + ty*ty) ;
    }
    return len;
  }
  
  inline float EffectivePrimIonDensity(float dd) {
    float dav = sqrt(dd*dd+0.25/(gPrimIonDensity*gPrimIonDensity)) ;
    return gPrimIonDensity * exp(-dav*gAttachDensity) ;
  }

  float FirstCluster(float dd /* distance to wire */, float eff) 
  {
    /* This routine simulates the distance of the closest cluster to any
       point on the track. The probability for `no attachment' is a
       function of the distance to the wire. A lower estimate is given
       by exp(fabs(dd)*attachDensity) ; This results in a `lam' effective
       ionization density.  Given this effective density we use a poison
       estimate of the closest cluster to the point, following the
       probability P(dlen) = exp(-2*dlen*lam) (The factor 2 comes in
       because the `closest cluster' can be in two directions, thus
       halving the average distance.)  */
    
    int i1 = 1;
    float dlen(0), rndm[1];
    // adjust prim ion for expected attachment probability
    float lam = EffectivePrimIonDensity(dd) ;
    // throw a distance
    grndm_(rndm,i1);
    dlen = -log(rndm[0])/(2*lam) ;
    
    if(eff<1) {
      grndm_(rndm,i1);
      if(rndm[0]>eff) dlen += 100 ; // this takes care hit is not valid
    }  
    return dlen;
  }
  
  float MeasuredTime( float d /* distance of first cluster */, float acc)
  {
    /* This is the t(r) relation. It transforms the monte carlo
       distance to the first cluster into a tdc value. Note that the
       gaussian smearing is applied here.*/

    if(!gIsInitialized) Initialize() ;
    float rndm[2] ;
    granor_(rndm[0],rndm[1]);
    float dm = d + acc*rndm[0] ;
    return (gTdcOffset + dm*gTdcGain/gDriftVelocity) ;
  }

  // The next two routines somehow mimic the effect of
  // calibration. They should take care that the average residual is
  // more or less zero and that hitb->r is a good measurement of the
  // resolution.

  float EstimatedRadius( float time, float pit, float acc) {
    float r;
    if(pit<1) r = gRTRelation05mm[int(time+0.5)] ;
    else      r = gRTRelation10mm[int(time+0.5)] ;
    return r;
  }
  
  float EstimatedSigma( float time, float pit, float acc) { 
    float stot ;
    if(pit<1) stot = gSTRelation05mm[int(time+0.5)] ;
    else      stot = gSTRelation10mm[int(time+0.5)] ;
    // this is to account for the fact that some modules might have
    // different resolution. just scale
    if(gGausSmearing>0 && acc>0) stot *= acc/gGausSmearing ;
    return stot ;
  }

  void GenerateXTalkHits( vector<digx>& dgvec, ArtePointer<GEDE> gf) {
    // Generates a cross-talk hit in nearest neighbour only. Neighbours are
    // in the same single layer (to keep things simple).
    if(gXTalkProbability > 0 ) {
      // generate random numbers
      vector<float> rndm(2*dgvec.size()) ;
      ranmar_(rndm.begin(),rndm.size()) ;
      
      // make hits
      vector<digx> newhits ;
      vector<float>::const_iterator rndmIt = rndm.begin() ;
      for(vector<digx>::const_iterator hitIt = dgvec.begin() ;
	  hitIt != dgvec.end() ; hitIt++) {
	int mono = hitIt->strip >> 16  ;
	int cell = hitIt->strip%256 -1 ; // starting at 1 !!
	
	// what are neighbours (nb) ? I'll define them : neighbouring 
	// in the same _single_ layer.
	int nbmono  = mono + 1 - 2*(mono%2) ;
	int nbcell1 = cell - (mono%2) ; // on +X side
	if(gf->xfw[0] < gf->xfw[1]) nbcell1 = cell-1 + (mono%2); // on -X side
	
	// loop over both neighbours and generate Xtalk
	for(int nbcellIt = nbcell1 ; nbcellIt <= nbcell1+1;
	    nbcellIt++, rndmIt++) {
	  if( nbcellIt >= 0 && nbcellIt < gf->nwi[nbmono] &&
	      (*rndmIt) < gXTalkProbability ) {
	    digx newhit ;
	    newhit.charge = hitIt->charge ;
	    newhit.strip  = (nbmono<<16) + nbcellIt + 1;
	    newhit.mimp   = 0 ; // don't assign mimp to X talk hits
	    newhits.push_back(newhit) ;
	  }
	}
      }
      dgvec.insert(dgvec.end(),newhits.begin(),newhits.end()) ;
    }
  }
  
  void GenerateNoiseHits( vector<digx>& dgvec, ArtePointer<GEDE> gf) {
    // Generates noise for every cell in the module. 
    // The routine is relatively slow, but still negligible with
    // respect to the total event generation time.

    if(gRandomNoiseProbability>0 ) {
      // generate random numbers by one call
      int nwires(0) ;
      for(int mono=0; mono < gf->nsl; mono++) nwires += gf->nwi[mono];
      vector<float> rndm(nwires) ;
      ranmar_(rndm.begin(),nwires) ;
      
      // loop over all cells   
      vector<digx> newhits ;
      vector<float>::const_iterator rndmIt = rndm.begin() ;
      for(int mono=0; mono < gf->nsl; mono++) 
	for(int cell=0; cell < gf->nwi[mono]; cell++, rndmIt++) {
	  if( (*rndmIt) < gRandomNoiseProbability ) {
	    // random tdc count
	    int tdccount = int(256 * (*rndmIt)/gRandomNoiseProbability) ;
	    digx newhit ;
	    newhit.charge = tdccount ;
	    newhit.strip  = (mono<<16) + cell + 1;
	    newhit.mimp   = 0 ; // don't assign mimp to noise hits
	    newhits.push_back(newhit) ;
	  }
	}
      dgvec.insert(dgvec.end(),newhits.begin(),newhits.end()) ;
    }
  }

  /* here follows the code for `calibratint' the MC r(t)-relation */

  class StatBin {
  public:
    StatBin(): sum_(0),sum2_(0),entries_(0) {} 
    const StatBin& operator+=(const StatBin& rhs) { 
      sum_+=rhs.sum_ ; sum2_+=rhs.sum2_; entries_+= rhs.entries_ ;
      return *this ; }
    void add(double x) { sum_+=x; sum2_+=x*x; entries_++; }
    double mean()  const { return sum_/entries_; }
    double width() const { return sqrt( (sum2_-sum_*sum_/entries_)/entries_) ; }
    unsigned int entries() const { return entries_ ; }
  private:
    double sum_,sum2_ ;
    unsigned int entries_ ;
  };

  void CalibrateRT(float pitch, float rtr[256], float str[256]) {
    // This routines calibrates the r(t) and s(t) relations used in
    // the digitization. We just throw a lot of `monte carlo distance
    // to wire', apply the cell mode and calculate the drift time. Then
    // we calculate for every drift-time bin the mean and width of the
    // corresponding `distance to wire' distribution.
    
    // if gIsInialized is not set to true, we get some strange loops:
    bool saveinitialized = gIsInitialized ;
    gIsInitialized = true ;

    bool verbose = false ;

    float rmin = 0 ;
    float rmax = pitch/3. ;    
    int nbins = 256 ;
    vector<StatBin> statvec(nbins) ;
    const int ntries= 200000 ;
    float rndm[ntries] ;
    grndm_(rndm,ntries);

    for(int i=0; i<ntries; ++i) {
      float dd  = rmin + rndm[i]*(rmax-rmin) ;
      float dpath = PathLengthHC(pitch,dd,0,0) ;
      if(dpath>0) {
	float dlen   = FirstCluster(dd,1) ;
	if( dlen < 0.5*dpath) {
	  float r = sqrt(dd*dd+dlen*dlen) ;
	  int bin = int(MeasuredTime(r,gGausSmearing)+0.5)%256 ;
	  statvec[bin].add(dd) ;
	}
      }
    }

    // calculate the mean and width in each tdc bin. smooth if
    // necessary.
    const unsigned int minentries = 1000 ;
    int firstbin(nbins),lastbin(0) ;
    for(int bin=0; bin<nbins; bin++) {
      StatBin thisstat = statvec[bin] ;
      int binbelow(bin-1),binabove(bin+1) ;
      while(thisstat.entries() < minentries && (binbelow>=0 || binabove<nbins)) {
	if(binbelow>=0 )  thisstat+=statvec[binbelow] ;
	if(binabove<nbins) thisstat+=statvec[binabove] ;
	binbelow-- ; binabove++ ;
      }
      rtr[bin] = thisstat.mean() ;
      str[bin] = thisstat.width() ;
      if(statvec[bin].entries()>100) {
	if(firstbin>bin) firstbin=bin ;
	if(lastbin<bin) lastbin=bin ;
      }
    } 
    // let's do a linear extrapolation for the empty bins (it doesn't
    // make sense, but it looks better)
    int span = 5;
    if(verbose) cout << "first/lastbin: " << firstbin << " " << lastbin << endl ;
    float drdt = (rtr[firstbin+span]-rtr[firstbin])/float(span) ;
    for(int bin=0; bin<firstbin; ++bin) 
      rtr[bin] = rtr[firstbin] + float(bin-firstbin)*drdt ;
    drdt = (rtr[lastbin]-rtr[lastbin-span])/float(span) ;
    for(int bin=lastbin+1; bin<nbins; ++bin)
      rtr[bin] = rtr[lastbin]  + float(bin-lastbin)*drdt ;
    if(verbose) 
      for(int bin=0; bin<nbins; bin++) 
	cout << bin << " " << rtr[bin] << " " << str[bin] << endl ;
    gIsInitialized = saveinitialized ;
  }
  
  void CalibrateRT() {
    /* This routine calibrates the r(t) and sigma(t) relations. There
       are only two of those: one for 5~mm and one for 10~mm
       cells. Note that if the smearing is not the same for each
       module, this will not be found in the sigma(t) relation. */
    cout << "* Calibrating r(t) and s(t) relations for MC" << endl ;
    
    if(!gRTRelation05mm) gRTRelation05mm = new float[256] ;
    if(!gRTRelation10mm) gRTRelation10mm = new float[256] ;
    if(!gSTRelation05mm) gSTRelation05mm = new float[256] ;
    if(!gSTRelation10mm) gSTRelation10mm = new float[256] ;
    
    // Generate a flat distance to the wire, generate a path length, a
    // tdc value. Store for every tdc bin.
    
    CalibrateRT(0.5*sqrt(3.0),gRTRelation05mm,gSTRelation05mm) ;
    CalibrateRT(1.0*sqrt(3.0),gRTRelation10mm,gSTRelation10mm) ;
  }
}

  
// the kuip interface for setting parameters
extern "C" 
{  
  double ku_getr();
  char  *ku_path();  

  void dgocell_kuip() {
     char ch[65]; 
     strncpy(ch,ku_path(),64);
     char *ptr = &ch[0];
     while (strtok(ptr,"/") != NULL) { 
       strcpy(ch,ptr);   ptr = strtok(NULL,"/");
     }

     for (unsigned int i =0; i<strlen(ch); i++) ch[i] = toupper(ch[i]); 

     if     (!strcmp(ch,"OTRPRIMIONDENSITY"))
                                 dgocell::gPrimIonDensity = ku_getr();
     else if(!strcmp(ch,"OTRATTACHDENSITY"))
                                 dgocell::gAttachDensity = ku_getr();
     else if(!strcmp(ch,"OTRGAUSSMEARING"))
                                 dgocell::gGausSmearing = ku_getr();
     else if(!strcmp(ch,"OTRNOMINALEFFICIENCY"))
                                 dgocell::gNominalEfficiency = ku_getr();
     else if(!strcmp(ch,"OTRXTALK"))
                                 dgocell::gXTalkProbability = ku_getr();
     else if(!strcmp(ch,"OTRRANDOMNOISE"))
                                 dgocell::gRandomNoiseProbability = ku_getr();

     dgocell::gIsInitialized=false ;
  }
}

