// ********************************************************************
//
// Routine: DQ_RECO_MATCH()
//
// Author:  P.Conde
//
// Description: some counting on the matchings.
//
// Filling of bins for Match histogram:
//
//   Bin 0: number of events.
//   Bin 1: number of Matches.
//   Bin 2: VDS-OTR matches.
//   Bin 3: VDS-ECAL matches.
//   Bin 4: VDS-RICH matches.
//   Bin 5: OTR-ECAL matches.
//   Bin 6: OTR-RICH matches.
//
// Subdetectors histograms: for each segment store the number of times
//                          that it's used into a Match:
//  Hist.: VDS segments used in OTR matchings.
//  Hist.: VDS segments used in ECAL matchings.
//  Hist.: VDS segments used in RICH matchings.
//  Hist.: OTR segments used in VDS matchings.
//  Hist.: OTR segments used in ECAL matchings.
//  Hist.: OTR segments used in RICH matchings.
//  Hist.: RICH segments used in VDS matchings.
//  Hist.: RICH segments used in OTR matchings.
//  Hist.: ECAL segments used in VDS matchings.
//  Hist.: ECAL segments used in OTR matchings.
//
// Histogram for kind of matchings VDS-Tracker
//
//  Bin0: PC-only segments. OTR only segments.
//  Bin1: MCPC segments. OTR only segments.
//  Bin2: PCTC segments. OTR only segments.
//  Bin3: MCPCTC segments. OTR only segments.
//
//  Bin5: Otr only Matches VDS-OTR.
//  Bin6: Itr only Matches VDS-OTR.
//
// ********************************************************************

// standrad includes
#include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <algo.h>


#include "arte/arte.hh"
#include "auto/auto.hh"
#include "arte/RSEG_cmp.hh"
#include "farm/farm.hh"
#include "rhp.h" 
#include "dq/DQ.h"

#define HOFF             2000
#define HIST_MATCH(I)     DQ_ID(COMP_RECO,LVL_REC,HOFF+I)

#define MATCH_H_ID    HIST_MATCH(15)
#define VDS_OTR_ID    HIST_MATCH(16)
#define VDS_ECAL_ID   HIST_MATCH(17)
#define VDS_RICH_ID   HIST_MATCH(18)
#define OTR_VDS_ID    HIST_MATCH(19)
#define OTR_ECAL_ID   HIST_MATCH(20)
#define OTR_RICH_ID   HIST_MATCH(21)
#define RICH_OTR_ID   HIST_MATCH(22)
#define RICH_VDS_ID   HIST_MATCH(23)
#define ECAL_OTR_ID   HIST_MATCH(24)
#define ECAL_VDS_ID   HIST_MATCH(25)

#define TRK_KIND_ID   HIST_MATCH(26)

//
//
//
int
DQ_RECO_MATCH_init()
{
  cout << "DQ_RECO_MATCH_init" << endl;

  rhp_hbook1(MATCH_H_ID,"MATCHED SEGMENTS",20,-0.5,19.5); 
  rhp_hbook1(VDS_OTR_ID,"VDS segm used in a VDS-OTR match",10,-0.5,9.5);
  rhp_hbook1(VDS_ECAL_ID,"VDS segm used in a VDS-ECAL match",10,-0.5,9.5);
  rhp_hbook1(VDS_RICH_ID,"VDS segm used in a VDS-RICH match",10,-0.5,9.5);
  rhp_hbook1(OTR_VDS_ID,"OTR segm used in a VDS-OTR match",10,-0.5,9.5);
  rhp_hbook1(OTR_ECAL_ID,"OTR segm used in a OTR-ECAL match",10,-0.5,9.5);
  rhp_hbook1(OTR_RICH_ID,"OTR segm used in a OTR-RICH match",10,-0.5,9.5);

  rhp_hbook1(RICH_OTR_ID,"RICH segm used in a OTR-RICH match",10,-0.5,9.5);
  rhp_hbook1(RICH_VDS_ID,"RICH segm used in a VDS-RICH match",10,-0.5,9.5);

  rhp_hbook1(ECAL_OTR_ID,"ECAL segm used in a OTR-ECAL match",10,-0.5,9.5);
  rhp_hbook1(ECAL_VDS_ID,"ECAL segm used in a VDS-ECAL match",10,-0.5,9.5);

  rhp_hbook1(TRK_KIND_ID,"Kind of TRK segments in OTR-VDS match",7,-0.5,6.5);

  return 0;
}

//
//
//
int
DQ_RECO_MATCH_evnt()
{

  rhp_hf1(MATCH_H_ID,0.,1.);

  // Loop over segments. 

   
  for(ArteTable<RSEG>::iterator<RSEG> rseg = ArteTable<RSEG>::begin();
      rseg != ArteTable<RSEG>::end(); ++rseg){

    // Count how many segments are used in each kind of matching.

    if((*rseg)->fit != 7) {

      // Vds Segments.

      if((*rseg)->cmp & Rsegc::vxd){
	int VdsOtr=0, VdsEcal=0, VdsRich=0;
	for(ArteRelationIterator<RSEG> matVds = (*rseg)->rbeg.begin();
	    matVds != (*rseg)->rbeg.end(); matVds++){
	  if((*matVds)->fit != 7 ) continue;
	  if((*matVds)->cmp & Rsegc::patt) VdsOtr++;
	  if((*matVds)->cmp & Rsegc::ecal) VdsEcal++;
	  if((*matVds)->cmp & Rsegc::rich) VdsRich++;
	}
	if(VdsOtr < 9 ) rhp_hf1(VDS_OTR_ID,VdsOtr,1.);
	else rhp_hf1(VDS_OTR_ID,9.,1.);
	if(VdsEcal < 9 ) rhp_hf1(VDS_ECAL_ID,VdsEcal,1.);
	else rhp_hf1(VDS_ECAL_ID,9.,1.);
	if(VdsRich < 9 ) rhp_hf1(VDS_RICH_ID,VdsRich,1.);
	else rhp_hf1(VDS_RICH_ID,9.,1.);
      }

      // Otr Segments.

      if(((*rseg)->cmp & Rsegc::patt)){

	int OtrVds=0, OtrRich=0, OtrEcal=0;
	for(ArteRelationIterator<RSEG> matOtr = (*rseg)->rbeg.begin();
	    matOtr != (*rseg)->rbeg.end(); matOtr++){
	  if((*matOtr)->fit != 7 ) continue;
	  if((*matOtr)->cmp & Rsegc::vxd) OtrVds++;
	  if((*matOtr)->cmp & Rsegc::rich) OtrRich++;
	}
	if(OtrVds < 9 ) rhp_hf1(OTR_VDS_ID,OtrVds,1.);
	else rhp_hf1(OTR_VDS_ID,9.,1.);
	if(OtrRich < 9 ) rhp_hf1(OTR_RICH_ID,OtrRich,1.);
	else rhp_hf1(OTR_RICH_ID,9.,1.);

	for(ArteRelationIterator<RCCL> ClusOtr = (*rseg)->rccl.begin();
	    ClusOtr != (*rseg)->rccl.end(); ClusOtr++){
	  OtrEcal++;
	}
	if(OtrEcal < 9 ) rhp_hf1(OTR_ECAL_ID,OtrEcal,1.);
	else rhp_hf1(OTR_ECAL_ID,9.,1.);
	// Number of OTR-ECAL matches:
	if( OtrEcal > 0) rhp_hf1(MATCH_H_ID,5.,1.);

      }


      // Rich Segments:

      if((*rseg)->cmp & Rsegc::rich){

	int RichOtr=0, RichVds=0;
	for(ArteRelationIterator<RSEG> matOtr = (*rseg)->rbeg.begin();
	    matOtr != (*rseg)->rbeg.end(); matOtr++){
	  if((*matOtr)->fit != 7 ) continue;
	  if((*matOtr)->cmp & Rsegc::vxd) RichVds++;
	  if(((*matOtr)->cmp & Rsegc::patt) ||((*matOtr)->cmp & Rsegc::trit))
	    RichOtr++;
	}
	if(RichOtr < 9 ) rhp_hf1(RICH_OTR_ID,RichOtr,1.);
	else rhp_hf1(RICH_OTR_ID,9.,1.);
	if(RichVds < 9 ) rhp_hf1(RICH_VDS_ID,RichVds,1.);
	else rhp_hf1(RICH_VDS_ID,9.,1.);
      }

    } // End of if(fit != 7){

    // Now count all the matchings of each kind.

    if((*rseg)->fit != 7 ) continue;

    rhp_hf1(MATCH_H_ID,1.,1.);
    
    if(((*rseg)->cmp & Rsegc::vxd) && ((*rseg)->cmp & Rsegc::patt)){

       rhp_hf1(MATCH_H_ID,2.,1.);

       // Tracker segments. Divide them by different kinds:

       ArtePointer<RSEG> Otr=0;
       for(ArteRelationIterator<RSEG> otrseg = (*rseg)->rend.begin();
	   otrseg!= (*rseg)->rend.end(); otrseg++){
	 if((*otrseg)->fit == 7 ) continue;
	 if((*otrseg)->cmp & Rsegc::patt) Otr = (*otrseg);
       }
       if(Otr == 0 ) continue;
       // Search for the TC part:
       
       ArtePointer<RSEG> TC=0;
       for(ArteRelationIterator<RSEG> tcit = Otr->rend.begin();
	   tcit != Otr->rend.end(); tcit++){
	 if((*tcit)->cmp & Rsegc:: trit) TC = (*tcit);
       }

       if( (Otr->cmp & Rsegc::wch) && !(Otr->cmp & Rsegc::msg))
	 rhp_hf1(TRK_KIND_ID,5.,1.);
       if( !(Otr->cmp & Rsegc::wch) && (Otr->cmp & Rsegc::msg))
	 rhp_hf1(TRK_KIND_ID,6.,1.);

       if(TC == 0 && !(Otr->cmp & Rsegc::magt))
	 rhp_hf1(TRK_KIND_ID,0.,1.);
       if(TC == 0 && (Otr->cmp & Rsegc::magt))
	 rhp_hf1(TRK_KIND_ID,1.,1.);
       if( TC != 0 && !(Otr->cmp & Rsegc::magt))
	 rhp_hf1(TRK_KIND_ID,2.,1.);
       if( TC != 0  && (Otr->cmp & Rsegc::magt))
	 rhp_hf1(TRK_KIND_ID,3.,1.);
 
    }
    if(((*rseg)->cmp & Rsegc::vxd) && ((*rseg)->cmp & Rsegc::ecal)) 
       rhp_hf1(MATCH_H_ID,3.,1.);
    if(((*rseg)->cmp & Rsegc::vxd) && ((*rseg)->cmp & Rsegc::rich)) 
       rhp_hf1(MATCH_H_ID,4.,1.);
    if(((*rseg)->cmp & Rsegc::patt) && ((*rseg)->cmp & Rsegc::rich))
      rhp_hf1(MATCH_H_ID,6.,1.);
    
  } // End of loop in segments.


  // Loop over ECAL clusters.

   for(ArteTable<RCCL>::iterator<RCCL> rccl = ArteTable<RCCL>::begin();
      rccl != ArteTable<RCCL>::end(); ++rccl){
     int EcalVds=0, EcalOtr=0;
     for(ArteRelationIterator<RSEG> matClus = (*rccl)->rseg.begin();
	 matClus != (*rccl)->rseg.end(); matClus++){
       if((*matClus)->fit != 7 && 
	  (((*matClus)->cmp & Rsegc::patt) ||((*matClus)->cmp & Rsegc::trit))) 
	 EcalOtr++;

       if((*matClus)->fit != 7 ) continue;
       if(((*matClus)->cmp & Rsegc::ecal) && ((*matClus)->cmp & Rsegc::vxd)) 
	 EcalVds++;
     }
     if(EcalVds < 9) rhp_hf1(ECAL_VDS_ID,EcalVds,1.);
     else rhp_hf1(ECAL_VDS_ID,9.,1.);
     if(EcalOtr < 9) rhp_hf1(ECAL_OTR_ID,EcalOtr,1.);
     else rhp_hf1(ECAL_OTR_ID,9.,1.);

   } // End of loop in ecal clusters.

  return 0;

}

