
// *****************************************************************************
//
// source:    faEmg.C
//
// created:   30 Jul 1997
//
// Author:    Andreas Gellrich
//            HERA-B Collaboration
//            DESY Zeuthen
//            Platanenallee 6
//            D-15738 Zeuthen
//            Germany
//            email: Andreas.Gellrich@desy.de
//
// purpose:   Handle read-in of raw data events from EMG.
//
// symbols:   faReadEmg
//            faInitEmg
//            faStopEmg
//
// called by: faRead
//            faInit
//            faStop
//
// changes:         
// 12 Jan 1999 (AG) removed emg status flag
// 14 Jan 1999 (AG) polishing
// 02 Feb 1999 (AG) polishing
// 22 Feb 1999 (AG) no READALL in emg for cycle mode
// 25 Mar 1999 (AG) correct emg_consum() and emg_get_event()
// 01 Jun 1999 (AG) put includes
// 09 Sep 1999 (AG) always use cout
// 03 Nov 1999 (AG) renamed CYCLE to NOBLOCK
// 23 Nov 1999 (AG) do not wait for emg
// 25 Nov 1999 (AG) introduce SMC
// 08 Dec 1999 (AG) count events
// 10 Dec 1999 (AG) event counting is done by Arte
// 24 Jan 2000 (AG) correct usleep()
// 29 Feb 2000 (AG) change sleep time to 50 msec
// 22 Mar 2000 (AG) polish emg
// 23 Mar 2000 (AG) polish emg
// 27 Apr 2000 (AG) use ARTE_mark()
// 09 May 2000 (AG) change printout
// 15 May 2000 (AG) polish printout
// 23 Jun 2000 (AG) polishing
// 11 Jul 2000 (AG) use faCheckRhp()
// 12 Jul 2000 (AG) polished comments
// 07 Aug 2000 (AG) downscaled printout
// 09 Aug 2000 (AG) more downscaled printout
// 06 Nov 2000 (AG) prepared for emg flag EMG_NSCAVGE
// 09 Nov 2000 (AG) enabled emg blocking
// 17 Jan 2001 (AG) added comments
//
// *****************************************************************************

// standard includes
#include <iostream.h>
#include <unistd.h>

// ONLINE includes
#include "rpm.h"
#include "smc.h"
#include "rhp.h"
#include "emg.h"
#include "grr.h"
#include "par.h"

// Arte includes
#include "farm/fa.hh"
#include "farm/farm.hh"

// externals
extern int  FADEB;     // debug flag
extern int  NOBLOCK;   // no block emg
extern int  RPMINIT;   // init rpm/rhp
extern char *PARNAME;  // emg partition name;
extern int  SEG_NUM;   // emg partitions
extern int  SEG_LEN;   // emg segment size

extern char *evtBuf;   // event buffer

// constants
//const int MUSEC_SLEEP = 500000;
const int MUSEC_SLEEP = 50000; // 50000 musec = 50 msec = 0.05 sec

// prototypes
#ifdef __linux
extern "C" void usleep( unsigned long musec );
#endif

int faCheckRhp( int print ); // handle rhp

// *****************************************************************************
//
// name:      faReadEmg
//
// purpose:   read event from emg
//
// called by: faReadEmg
//
// input:     void
// return:    int  rc
//
// *****************************************************************************

int
faReadEmg()
{
  ARTE_mark( "faReadEmg", 0 );

  // pre-sets
  int        rc    = GRR_OK;
  static int count = 0;
  static int loop  = 0;
  
  // print
  if( count < 10 ){
    cout << " EMG: Waiting for event ....." << endl;
  }

  //
  // get event
  //
  int flags  = NOWAIT;
  while( 1 ){

    // read emg
    if( ( rc = emg_get_event( (int *) evtBuf, SEG_LEN, flags ) ) == GRR_OK ){

      // print
      if( count < 10 ){
	count++;
	cout << " EMG: Got event" << endl;
      }

      // rhp 
      if( RPMINIT ){
	if( smc_if_trans_pend() == SMC_OK ) smc_complete_trans(); // complete transition 
	faCheckRhp( 1 );                                          // handle rhp
      }

      break;
    }
    // empty emg
    else if( rc == GRR_EMPTY ){

      loop++;
      int print = !( loop%100 );
      if( print ) cout << " EMG: Queue empty. Sleep " << MUSEC_SLEEP << " musec and retry." << endl;
      
      // rhp
      if( RPMINIT ){
	if( smc_if_trans_pend() == SMC_OK ) smc_complete_trans(); // complete transition
	faCheckRhp( print ); 	                                          // handle rhp 
      }
      
      // sleep
      usleep( MUSEC_SLEEP );
    }
    // emg error
    else {
      cout << " EMG: ERROR: getting event rc=" << rc << endl;
      break;
    }
  }

  return rc;
}

// *****************************************************************************
//
// name:      faInitEmg
//
// purpose:   initialize emg
//
// called by: fainit_
//
// input:     void
// return:    int  error code
//
// *****************************************************************************

int
faInitEmg()
{
  ARTE_mark( "faInitEmg", 0 );

  //
  // flags
  //
  int consumFlags = READALL;
  if( NOBLOCK ){
    consumFlags = 0;
  }

  //int attachFlags = 0;           // create OR use
  int attachFlags = EMG_NSCAVGE; // enables blocking

  //
  // init emg
  //
  int rc = GRR_OK;
  if( ( rc = emg_init() ) != GRR_OK ){ // initialize emg
    cout << "EMG: ERROR: init rc=" << rc << endl;
    return rc;
  }

  //
  // emg ctrl
  //
  cout << "EMG: "
       << "partition=" << PARNAME     << " "
       << "segments="  << SEG_NUM     << " "
       << "length="    << SEG_LEN     << " "
       << "attach="    << consumFlags << " "
       << "consum="    << attachFlags << endl;

  //
  // SEG_NUM
  //
  if( ( rc = emg_ctrl( NSEGS, &SEG_NUM ) ) != GRR_OK ){ // set parameters
    cout << "EMG: ERROR: segments rc=" << rc << endl;
    return rc;
  }

  //
  // SEG_LEN
  //
  if( ( rc = emg_ctrl( SEGLEN, &SEG_LEN ) ) != GRR_OK ){
    cout << "EMG: ERROR: length rc=" << rc << endl;
    return rc;
  }

  //
  // attach to emg
  //
  if( ( rc = emg_attach( PARNAME, attachFlags ) ) == GRR_FISHY ){
    cout << "EMG: ERROR: (GRR_FISHY) partition NOT attached." << endl;
    ARTE_error( GRR_SFATAL, "EMG: ERROR: (GRR_FISHY)", rc );
    ARTE_exit ( GRR_SFATAL, "EMG: ERROR: (GRR_FISHY)", rc );
  }
  else if( rc == GRR_UNKNOWN ){
    cout << "EMG: ERROR: (GRR_UNKNOWN) partition NOT found." << endl;
    ARTE_error( GRR_SFATAL, "EMG: ERROR: (GRR_UNKNOWN)", rc );
    ARTE_exit ( GRR_SFATAL, "EMG: ERROR: (GRR_UNKNOWN)", rc );
  }
  else if( rc == GRR_INCOMPATIBLE ){
    cout << "EMG: ERROR: (GRR_INCOMPATIBLE) partition USED somewhere else?" << endl;
    ARTE_error( GRR_SFATAL, "EMG: ERROR: (GRR_INCOMPATIBLE)", rc );
    ARTE_exit ( GRR_SFATAL, "EMG: ERROR: (GRR_INCOMPATIBLE)", rc );
  }
  else if( rc != GRR_OK ){
    cout << "EMG: ERROR: rc=" << rc << endl;
    ARTE_error( GRR_SFATAL, "EMG: ERROR", rc );
    ARTE_exit ( GRR_SFATAL, "EMG: ERROR", rc );
  }

  //
  // consum
  //
  if( ( rc = emg_consum( -1, -1, 1, consumFlags ) ) != GRR_OK ){ // consum emg
    cout << "EMG: ERROR: consuming "
	 << "consumFlags=" << consumFlags << " "
	 << "rc="    << rc    << endl;
    return rc; // error
  }
  
  return rc; // okay
}

// *****************************************************************************
//
// name:      faStopEmg
//
// purpose:   terminate
//
// called by: fastop_
//
// input:     void
// return:    int  error code
//                 =0: ok
//                 <0: error
//
// *****************************************************************************

int
faStopEmg()
{
  ARTE_mark( "faStopEmg", 0 );

  int rc = GRR_OK;
  
  // release
  if( ( rc = emg_release( PARNAME, 0 ) ) == GRR_OK ){ // realease emg
    cout << "EMG: released (1)." << endl;
  }
  // error
  else {
    cout << "EMG: ERROR: releasing (1) rc=" << rc << endl;
    ARTE_error( GRR_SERROR, "EMG: emg_release()", rc );
  }

  // sleep
  sleep( 1 );
  
  // release again
  rc = emg_release( PARNAME, 0 ); // release emg
  
  return rc;
}
