#include <iostream>

#ifndef GROVER_WITHOUT_ARTE
#include "arte/arte.hh"
#endif

#include "grover/DumpFormatter.hh"
#include "grover/grover.hh"

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                  GroverWire member functions                       //
//                                                                    //
////////////////////////////////////////////////////////////////////////

//
// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// ------------------------------------------------------------------------
// 00-10-29  M.Sang   Added constructor for RTAR table
// 01-01-12  M.Sang   Added operator <<
// 01-02-14  M.Sang   Added symmetric operations DistanceTo(...) 
//                    and Chi2DistTo(...) where meaningful
// 01-02-28  M.Sang   Added operators == and !=
// 01-03-07  M.Sang   Corrected wire names in operator<<
// 01-03-13  Max      Added preprocessor directives
// 01-04-24  Michael Added sx, sy, sz simular as in RTAR 
//

//
//
// SEE ALSO grover_inlines.hh !!!!!!!
//
//


GroverWire::GroverWire(void)
  : wnr(-1), 
  x(0.),     y(0.),  z(0.), 
  dx(0.025), dy(1.), dz(0.025),
  sx(0.025), sy(1.), sz(0.025),
  shape(1),
  horizontal(false),   vertical(true),
  intarget(false), inlocal(false), inglobal(true),
  ntr(0), nprim(0), nsec(0), mobile(1)
{ // Default constructor is a vertical wire at 0,0,0 with thickness 250*250 um
}




#ifndef GROVER_WITHOUT_ARTE
GroverWire::GroverWire(const ArtePointer<GTAR> gtar)
  : wnr(gtar->wnr),
    x(gtar->x),  y(gtar->y), z(gtar->z),
    dx(gtar->dx),  dy(gtar->dy), dz(gtar->dz),
    sx(gtar->dx),  sy(gtar->dy), sz(gtar->dz),
    shape(gtar->shap),
    intarget(false), inlocal(false), inglobal(true),
    ntr(0), nprim(0), nsec(0), mobile(1) 
{ // Constructor from GTAR entry
  for (int i=0;i<9;i++) rot[i] = gtar->rot[i];
  horizontal = (fabs(dx)>fabs(dy));
  vertical = !horizontal;
}
#endif



#ifndef GROVER_WITHOUT_ARTE
GroverWire::GroverWire(const ArtePointer<RTAR> rtar)
  : wnr(rtar->wnr),
    x(rtar->x),  y(rtar->y), z(rtar->z),
    dx(rtar->dx),  dy(rtar->dy), dz(rtar->dz),
    sx(rtar->sx),  sy(rtar->sy), sz(rtar->sz),
    shape(0),
    intarget(false), inlocal(false), inglobal(true),
    ntr(0), nprim(0), nsec(0), mobile(1)
{ // Constructor from RTAR entry added 29.10.00 MS
  for (int i=0;i<9;i++) rot[i] = FAILURE;
  horizontal = (fabs(dx)>fabs(dy));
  vertical = !horizontal;
}
#endif




GroverWire::~GroverWire(void) 
{ // default destructor
}

int GroverWire::GlobalToLocal(void)
{
  // Covariance unchanged (not a bad approximation, in fact).
  // Momentum should be re-calculated (uncomment line below if/when
  // the EstimateMomentum() member function becomes available!)
  if (inglobal && !inlocal) {

#ifndef GROVER_WITHOUT_ARTE
    float oldpos[5], newpos[5];
    oldpos[0] = x; oldpos[1] = y; oldpos[2] = z; oldpos[3] = 0; oldpos[4] = 0;
    acGlobalToLocal(newpos, oldpos, 1);
    x = newpos[0]; y = newpos[1]; z = newpos[2];
#endif

    inlocal = true; inglobal = false;
  }
  return SUCCESS;
}

int GroverWire::LocalToGlobal(void)
{
  // Covariance unchanged (not a bad approximation, in fact).
  // Momentum should be re-calculated (uncomment line below if/when
  // the EstimateMomentum() member function becomes available!)
  if (inlocal && !inglobal) {

#ifndef GROVER_WITHOUT_ARTE
    float oldpos[5], newpos[5];
    oldpos[0] = x; oldpos[1] = y; oldpos[2] = z; oldpos[3] = 0; oldpos[4] = 0;
    acLocalToGlobal(oldpos, newpos, 1);
    x = newpos[0]; y = newpos[1]; z = newpos[2];
#endif

    inlocal = false; inglobal = true;
  }
  return SUCCESS;
}  



float GroverWire::DistanceTo(const GroverParticle* that) const 
{
  return that->DistanceTo(this);
}



float GroverWire::Chi2DistTo(const GroverParticle* that) const 
{
  return that->Chi2DistTo(this);
}



float GroverWire::DistanceTo(const GroverVertex* that) const 
{
  return that->DistanceTo(this);
}



float GroverWire::Chi2DistTo(const GroverVertex* that) const 
{
  return that->Chi2DistTo(this);
}



float GroverWire::DistanceTo(const GroverWire* that) const 
{
  if (this->Horizontal() && that->Horizontal()) {
    float dy = this->Y() - that->Y();
    float dz = this->Z() - that->Z();
    return static_cast<float>(sqrt(dy*dy+dz*dz));
  }
  else if (this->Vertical() && that->Vertical()) {
    float dx = this->X() - that->X();
    float dz = this->Z() - that->Z();
    return static_cast<float>(sqrt(dx*dx+dz*dz));
  }
  else {
    return this->Z() - that->Z();
  }
}



float GroverWire::DistanceTo(const float xx, const float yy, const float zz) const 
{
  if (this->Horizontal()) {
    float dy = this->Y() - yy;
    float dz = this->Z() - zz;
    return static_cast<float>(sqrt(dy*dy+dz*dz));
  }
  else if (this->Vertical()) {
    float dx = this->X() - xx;
    float dz = this->Z() - zz;
    return static_cast<float>(sqrt(dx*dx+dz*dz));
  }
  else return -1e9; // should never happen!
}



void GroverWire::Dump(void) const { // Dump members to stdout
  cout << "*** Grover: Dump() of GroverWire object" << endl;
  DumpFormatter Nice;
  Nice.DumpClassName("GroverWire");
  Nice.DumpComment("Part of the Grover vertexing package");
  Nice.DumpComment("Please send questions about this class "
		    "to the Grover authors");
  Nice.DumpHeader();
  Nice.DumpPrint("wnr",&wnr);
  Nice.DumpPrint("x",&x);
  Nice.DumpPrint("y",&y);
  Nice.DumpPrint("z",&z);
  Nice.DumpPrint("dx",&dx);
  Nice.DumpPrint("dy",&dy);
  Nice.DumpPrint("dz",&dz);
  Nice.DumpPrint("sx",&sx);
  Nice.DumpPrint("sy",&sy);
  Nice.DumpPrint("sz",&sz);
  Nice.DumpPrint("horizontal",&horizontal);
  Nice.DumpPrint("vertical",&vertical);
  Nice.DumpPrint("rot",rot,9);
  Nice.DumpPrint("intarget",&intarget);
  Nice.DumpPrint("inlocal",&inlocal);
  Nice.DumpPrint("inglobal",&inglobal);
  Nice.DumpPrint("ntr",&ntr);
  Nice.DumpPrint("nprim",&nprim);
  Nice.DumpPrint("nsec",&nsec);
  Nice.DumpPrint("mobile",&mobile);
  Nice.DumpFooter();  
}

///////////////////////////////////////////////////////////////////
//
// Operators
//


bool operator==(const GroverWire& a, const GroverWire& b)
{
  return (a.Number()==b.Number());
}

bool operator!=(const GroverWire& a, const GroverWire& b)
{
  return !(a==b);
}

ostream& operator<<(ostream& s, GroverWire& w)
{
  s << "[";
  int n = w.Number();
  s << "wnr=" << n;
  switch (n) {
  case 1: {s << "(a1)"; break;}
  case 2: {s << "(b1)"; break;}
  case 3: {s << "(i1)"; break;}
  case 4: {s << "(o1)"; break;}
  case 5: {s << "(a2)"; break;}
  case 6: {s << "(b2)"; break;}
  case 7: {s << "(i2)"; break;}
  case 8: {s << "(o2)"; break;}
  default:{break;}
  }
  s << " x,y,z={"    << w.X()     << "," << w.Y()     << "," << w.Z()     << "}";
  s << " dx,dy,dz={" << w.SizeX() << "," << w.SizeY() << "," << w.SizeZ() << "}";
  s << " sx,sy,sz={" << w.SX() << "," << w.SY() << "," << w.SZ() << "}";
  
  if (w.InLocal()) s << " crd=LOC";
  else s << " crd=GLO";
  s << "]";
  return s;
}


