//**********************************************************************/
// Created: 1.03.98  by  Olya Igonkina   for   MARPLE package
//
// find the closest target for RcPoint
// take into account a rotation of the wires around beam.
// check c_rotation_matrix for correct rotation 
//
//**********************************************************************/
#include <math.h>

#include "marple/marple.hh"
#include "arte/arte.hh"
#include "arte/ArteMessage.hh"

extern void maFindTarget( RcPoint & rp, 
			  HepDouble &tarX,   HepDouble &tarY,  HepDouble &tarZ,
			  HepDouble &covTarXX, HepDouble &covTarYY, 
			  HepDouble &covTarXY, 
			  int & iErr);
extern void maFindTarget( RcPoint & rp, 
			  HepDouble &tarX,   HepDouble &tarY,  HepDouble &tarZ,
			  HepSymMatrix & covTar , 
			  int & iErr);


struct MaTarget
{
  ArtePointer<GTAR> gtar;
  HepMatrix rot;

  int closeToBeam;

  HepDouble xc;
  HepDouble yc;
  HepDouble zc;

  HepSymMatrix covc;
};

void maFindTarget( RcPoint & rp, 
		   HepDouble &tarX,   HepDouble &tarY,  HepDouble &tarZ,
		   HepDouble &covTarXX, HepDouble &covTarYY, 
		   HepDouble &covTarXY, 
		   int & iErr)
{
  tarX = 0;
  tarY = 0;
  tarZ = 0;
  covTarXX = 0.000025; // 0.005*0.005
  covTarYY = 0.000025;
  covTarXY = 0;

  HepSymMatrix covTar(3,1);
  maFindTarget( rp, tarX, tarY, tarZ, covTar, iErr );
  if( !iErr )
    {
      covTarXX = covTar(1,1);
      covTarYY = covTar(2,2);
      covTarXY = covTar(1,2);
    }
}

void maFindTarget( RcPoint & rp, 
		   HepDouble &tarX,   HepDouble &tarY,  HepDouble &tarZ,
		   HepSymMatrix & covTar , 
		   int & iErr)
{
  iErr = 0;
  //-----------------------------------------------------------------
  // check z position
  //-----------------------------------------------------------------
  if( rp.z() > 200 )
    {
      arteMsg << "MARPLE: maFindTarget: RcPoint is too far from target - z =" 
	   << rp.z() << endl << "prolongate to target first " ;
      arteMsg.allp();
      iErr = 2;
      return;
    }

  //-----------------------------------------------------------------
  // check X, Y position, may be {0,0,-3} is the best approximation
  //-----------------------------------------------------------------
  float x0 = rp.x()+(3-rp.z())*rp.tx();
  float y0 = rp.y()+(3-rp.z())*rp.ty();
  if( fabs(x0)>100 || fabs(y0)>100)
    {
      static int idMesg = 2738982;
      arteMsg << "MARPLE findTarget Warning :{0,0,-3} is accepted as target\n";
      arteMsg.allp(idMesg,10);
      tarX = 0;
      tarY = 0;
      tarZ = -3;
      
      covTar[0][0] = 0.25;  
      covTar[1][1] = 0.25;  
      covTar[0][1] = 0;
      return;
    }

  //-----------------------------------------------------------------
  // initialization of necessary parameters
  //-----------------------------------------------------------------
  static int eventKeeper = (*ArteTable<EVHD>::begin())->evt;

  static int firstCallMaFindTarget = 1;
  unsigned int nrow = ArteTable<GTAR>::size();
  static std::vector<MaTarget> maTarget;
  if( maTarget.size() != nrow ) // after first call the GTAR table is changed
    {
      maTarget.erase(maTarget.begin(), maTarget.end() );
      firstCallMaFindTarget = 1;
      maTarget.insert(maTarget.begin(), nrow, MaTarget());  
    }

  //--------------------------------------------------------------
  // initialization
  //--------------------------------------------------------------
  if( firstCallMaFindTarget )
    {
      firstCallMaFindTarget = 0; 
      if( ArteTable<GTAR>::begin() ==  ArteTable<GTAR>::end() ||
	  !(*ArteTable<GTAR>::begin() != 0)  )
	{
	  arteMsg << "MARPLE: maFindTarget : GTAR bank is empty ";
	  arteMsg.allp();
	  iErr = 1;
	  return;
	}
      // it is supposed that GTAR table order is constant for all events
      int iCount=0;
      for( ArteTable<GTAR>::iterator gtarI = ArteTable<GTAR>::begin();
	    gtarI != ArteTable<GTAR>::end(); ++gtarI, iCount++ )
	{
	  maTarget[iCount].gtar = *gtarI;

	  HepMatrix rot(3,3,1);
	  for( int i=0; i<3; i++){
	    for(int j=0; j<3; j++){
	      rot[i][j] = HepDouble((*gtarI)->rot[c_rotation_matrix[i][j]]); 
	    }
	  }
	  maTarget[iCount].rot = rot;

	  HepSymMatrix tmpMatrix(3,1);
	  maTarget[iCount].covc = tmpMatrix;
	  maTarget[iCount].covc(1,1) = (*gtarI)->dx * (*gtarI)->dx / 4;
	  maTarget[iCount].covc(2,2) = (*gtarI)->dy * (*gtarI)->dy / 4;
	  maTarget[iCount].covc(3,3) = (*gtarI)->dz * (*gtarI)->dz / 4;
	  (maTarget[iCount].covc).similarity(maTarget[iCount].rot);

	  maTarget[iCount].xc  = 
	    rot(1,1)*(*gtarI)->x + rot(1,2)*(*gtarI)->y + rot(1,3)*(*gtarI)->z;
	  maTarget[iCount].yc  =  
	    rot(2,1)*(*gtarI)->x + rot(2,2)*(*gtarI)->y + rot(2,3)*(*gtarI)->z;
	  maTarget[iCount].zc  =  
	    rot(3,1)*(*gtarI)->x + rot(3,2)*(*gtarI)->y + rot(3,3)*(*gtarI)->z;

	  if( fabs( maTarget[iCount].xc ) < 1.5 &&
	      fabs( maTarget[iCount].yc ) < 1.5 &&
	      fabs((*gtarI)->eff) > .1e-10 )
	                maTarget[iCount].closeToBeam = 1;
	  else          maTarget[iCount].closeToBeam = 0;

	}
    }else if ( eventKeeper != (*ArteTable<EVHD>::begin())->evt ){
      //-----------------------------------------------------------------
      //  update wires position for new event
      //-----------------------------------------------------------------
      eventKeeper = (*ArteTable<EVHD>::begin())->evt;
      for( unsigned int iTar = 0; iTar < maTarget.size(); iTar++ )
	{
	  ArtePointer<GTAR> gtarI = maTarget[iTar].gtar;
	  maTarget[iTar].xc  = 
	    maTarget[iTar].rot(1,1)*gtarI->x + 
	    maTarget[iTar].rot(1,2)*gtarI->y + 
	    maTarget[iTar].rot(1,3)*gtarI->z;
	  maTarget[iTar].yc  =  
	    maTarget[iTar].rot(2,1)*gtarI->x + 
	    maTarget[iTar].rot(2,2)*gtarI->y + 
	    maTarget[iTar].rot(2,3)*gtarI->z;
	  maTarget[iTar].zc  =  
	    maTarget[iTar].rot(3,1)*gtarI->x + 
	    maTarget[iTar].rot(3,2)*gtarI->y + 
	    maTarget[iTar].rot(3,3)*gtarI->z;	  

	  if( fabs( maTarget[iTar].xc ) < 1.5 &&
	      fabs( maTarget[iTar].yc ) < 1.5 && 
	      fabs( gtarI->eff) > .1e-10 )
	    maTarget[iTar].closeToBeam = 1;
	  else          maTarget[iTar].closeToBeam = 0;
	}	 
    }


  //-----------------------------------------------------------------
  // search for closest wire
  //-----------------------------------------------------------------

  HepDouble distanceOld = 100000.;
  int iTarOld = -1;
  for( unsigned int iTar=0; iTar < maTarget.size(); iTar++ )
    {
      if( !maTarget[iTar].closeToBeam ) continue;
      HepDouble Mx, My, Mz;
      if( fabs(maTarget[iTar].gtar->x)< 0.1e-8 )
	{
	  //             horizontal wire
	  Mx =  rp.ty()*maTarget[iTar].rot(3,1) - maTarget[iTar].rot(2,1);
	  My = -rp.tx()*maTarget[iTar].rot(3,1) + maTarget[iTar].rot(1,1);
	  Mz =  rp.tx()*maTarget[iTar].rot(2,1) - 
	        rp.ty()*maTarget[iTar].rot(1,1);
	}else{
	  //             vertical wire
	  Mx =  rp.ty()*maTarget[iTar].rot(3,2) - maTarget[iTar].rot(2,2);
	  My = -rp.tx()*maTarget[iTar].rot(3,2) + maTarget[iTar].rot(1,2);
	  Mz =  rp.tx()*maTarget[iTar].rot(2,2) - 
	        rp.ty()*maTarget[iTar].rot(1,2);
	}
      HepDouble distance = fabs( (rp.x() - maTarget[iTar].xc)*Mx + 
				 (rp.y() - maTarget[iTar].yc)*My + 
				 (rp.z() - maTarget[iTar].zc)*Mz )/
	sqrt( Mx*Mx + My*My + Mz*Mz );
      if( distance < distanceOld )
	{
	  distanceOld = distance;
	  iTarOld = (int)iTar;
	}
    }
  if( iTarOld < 0 || iTarOld >= (int)maTarget.size() ){ iErr = 1; cout << "!!!!!! No Target is found "<< iTarOld << " " << rp<< "\n"; return; }
  
  tarX = maTarget[iTarOld].xc;
  tarY = maTarget[iTarOld].yc;
  tarZ = maTarget[iTarOld].zc;

  covTar=maTarget[iTarOld].covc;

  //cout << "!!!!!!!!!!!! Pick Target " <<  maTarget[iTarOld].gtar << endl;

}

//void maTargetPrint()
//{
//cout << " maTarget " << iCount 
//     << " for GTAR" <<  maTarget[iCount].gtar << endl;
//cout << " xc = " << maTarget[iCount].xc 
//     << " yc = " << maTarget[iCount].yc 
//     << " zc = " << maTarget[iCount].zc << endl; 
//cout << " Mx = " << maTarget[iCount].Mx 
//     << " My = " << maTarget[iCount].My
//     << " Mz = " << maTarget[iCount].Mz << endl;
//cout << " rotation by" << maTarget[iCount].rot
//     << " covariant matrix " << maTarget[iCount].covc << endl;
//}
