// See HPT_DAQ_Abstract.hh for details.

// Correction made on 30.09.00: in HPT_DAQ_Abstract::process_event
//  reset_fired_cell_vectors_in(hpt_setup); was substituted by
//  reset_fired_cells_in(hpt_setup, false); /see fired_cell_tools.hh for
// explanations of these methods/


// Author: Vladislav Balagura (balagura@hera-b.desy.de, 03.11.1999)

#include "ptset/HPT_DAQ_Abstract.hh"

#include "ptset/high_pt.hh"
#include "ptset/HPT_cable_maker_based_on_horizontal_rows.hh"
#include "ptset/FEDs.hh"
#include "ptset/fired_cell_tools.hh"
#include <fstream.h>
#include <stdio.h>
#include "ptset/setup_menu.hh"
#include "ptset/io_cell_states.hh"

HPT_DAQ_Abstract::HPT_DAQ_Abstract(const string& hpt_geo_file,
				   const string& hpt_cables_file,
				   const string& hpt_hit_map_file)
  : cables(0), feds(0) {
    // ------------------------------ Setup ------------------------------
    cout << flush;
    if ( hpt_setup.empty() ) { // if hpt_setup is not initialized:
      // initialize it using the "hpt_geo.dat" from the current directory
      string setup_file_name;
      if ( &hpt_geo_file == 0 ) setup_file_name = "hpt_geo.dat";
      else setup_file_name = hpt_geo_file;
      ifstream setup_file(setup_file_name.c_str(),
			  ios::in | ios::bin | ios::nocreate);
      if ( !setup_file ) {
	fprintf(stderr, "can not open geometry file %s\n",
		setup_file_name.c_str());
	return;
      }
      printf("Geometry file \"%s\" opened\n", setup_file_name.c_str());
      hpt_setup.read(setup_file);
      if ( global_error() ) { printf_global_error(); return; }
      printf("Geometry defined\n");
    }
    // ------------------------------ Cabling ------------------------------
    cables = new Cables<HPT_Setup::IChamber>();
    feds = new FEDs<HPT_Setup::IChamber>();
    
    // create interactive commands in "menu" which will fill cables and feds
    SetupBaseMenu<HPT_Setup> menu(&hpt_setup);
    menu.chamber("define_cable") = make_interactive_command_1par
      (HPT_CableMakerBasedOnHorizontalRows(cables),
       (HPT_Setup::IChamber*)0);
    // HPT_CableMakerBasedOnHorizontalRows fills "cables" for 
    // one chamber. The chamber iterator HPT_Setup::IChamber is
    // the "parameter" for the interactive command
    if ( global_error() ) { printf_global_error(); return; }
      
    menu("define_fed") = make_interactive_command
      (FEDs_Filler<HPT_Setup::IChamber>(feds, *cables));
    // this is usual interactive command (without parameters)

    {
      string cables_file_name;
      if ( &hpt_cables_file == 0 ) cables_file_name = "hpt_cables.dat";
      else cables_file_name = hpt_cables_file;
      ifstream is_cables(cables_file_name.c_str(),
			 ios::in | ios::bin | ios::nocreate);
      if ( !is_cables ) { 
	fprintf(stderr, "can not open cabling description file %s\n",
		cables_file_name.c_str());
	return;
      }
      printf("Cabling description file %s opened\n", cables_file_name.c_str());
   
      
      // here new file was opened. However the cabling information 
      // and the geometry information can be supplied together in one file
      // since we defined one "menu". In this case one should only make sure
      // that the cells are created before cables are connected to them
      // (otherwise a global_error() will be produced). This means
      // the /make_cells command should be issued **before**
      // /define_cable command in the chamber menu.
      
      menu(&is_cables);
      if ( global_error() ) { printf_global_error(); return; }
      printf("Cabling defined\n");
      /*
      {
	ostrstream tmp;
	feds->dump(tmp);
	printf(tmp.str());
      } // print out which cables are connected to FEDs
      */
    }
    
    // ------------------------------ Hit map -------------------------------
    menu.section("BAD") = make_interactive_command_1par
      (BAD_NOISY_Cells_Marker<HPT_Chamber::ISection>(Fired_Cell::BAD),
       (HPT_Setup::ISection*)0);
    menu.section("NOISY") = make_interactive_command_1par
      (BAD_NOISY_Cells_Marker<HPT_Chamber::ISection>(Fired_Cell::NOISY),
       (HPT_Setup::ISection*)0);
    //  if ( global_error() ) { cerr << global_error(); return 1; }  
    {
      string hit_map_file_name;
      if ( &hpt_hit_map_file == 0 ) hit_map_file_name = "hpt_hit_map.dat";
      else hit_map_file_name = hpt_hit_map_file;
      
      if ( hit_map_file_name != "" ) {
	ifstream is_hit_map(hit_map_file_name.c_str(),
			    ios::in | ios::bin | ios::nocreate);
	if ( !is_hit_map ) { 
	  fprintf(stderr, "can not open BAD/NOISY map file %s\n",
		  hit_map_file_name.c_str());
	  return;
	}
	printf("BAD/NOISY map file %s opened\n", hit_map_file_name.c_str());
	menu(&is_hit_map);
	if ( global_error() ) { printf_global_error(); return; }
	printf("BAD/NOISY cell states defined\n");
      }
      else printf("BAD/NOISY map file is not given\n");
      //  write_cell_states(*HERA_B_detector, "inner", cout);
    }
}

inline void write_byte_of_data 
(Cable<HPT_Setup::IChamber>& cable, unsigned int& ichannel,
 unsigned char byte) {
  if ( cable.is_connected(ichannel) ) 
    write_data_to_fired_setup_cell(bool(byte & 0x01), cable[ichannel]);
  ++ichannel;
  if ( cable.is_connected(ichannel) ) 
    write_data_to_fired_setup_cell(bool(byte & 0x02), cable[ichannel]);
  ++ichannel;
  if ( cable.is_connected(ichannel) ) 
    write_data_to_fired_setup_cell(bool(byte & 0x04), cable[ichannel]);
  ++ichannel;
  if ( cable.is_connected(ichannel) ) 
    write_data_to_fired_setup_cell(bool(byte & 0x08), cable[ichannel]);
  ++ichannel;
  if ( cable.is_connected(ichannel) ) 
    write_data_to_fired_setup_cell(bool(byte & 0x10), cable[ichannel]);
  ++ichannel;
  if ( cable.is_connected(ichannel) ) 
    write_data_to_fired_setup_cell(bool(byte & 0x20), cable[ichannel]);
  ++ichannel;
  if ( cable.is_connected(ichannel) ) 
    write_data_to_fired_setup_cell(bool(byte & 0x40), cable[ichannel]);
  ++ichannel;
  if ( cable.is_connected(ichannel) ) 
    write_data_to_fired_setup_cell(bool(byte & 0x80), cable[ichannel]);
  ++ichannel;
}

void HPT_DAQ_Abstract::process_event(short int BX, unsigned int* fed_numbers, 
				     unsigned char** data, int n_feds) {
  //  if ( n_feds != 1 ) cout << "******** n_feds: " << n_feds << endl;
  reset_fired_cells_in(hpt_setup, false);

  typedef FED<HPT_Setup::IChamber> HPT_FED;
  FEDs<HPT_Setup::IChamber>::iterator iter_fed;
  
  for (int i_fed=0; i_fed<n_feds; ++i_fed) {
    if ((iter_fed = feds->find(fed_numbers[i_fed])) == feds->end()) {
      /*
      string tmp;
      tmp << "data accepted from the FED with #" << fed_numbers[i_fed]
	  << " which is not in the FEDs table";
      global_error()(tmp);
      return;
      */
      continue;
    }
    HPT_FED& fed = (*iter_fed).second;
    unsigned char* p_data = data[i_fed]; 

    for (unsigned int islot=0; islot<HPT_FED::N_SLOTS; ++islot) {
      if ( fed.is_connected(islot) == false ) {
	p_data += 2; // skip 16 bits
	continue;
      }
      Cable<HPT_Setup::IChamber>& cable = (*fed[islot]).second;

      unsigned int ichannel = 0;
      write_byte_of_data(cable, ichannel, *p_data++);
      write_byte_of_data(cable, ichannel, *p_data++);
    }
  }

  run(BX);
}

HPT_DAQ_Abstract::~HPT_DAQ_Abstract() {
  delete cables;
  delete feds;
}
  
