/* This file is analogous to intractive.hh.
   It defines the analogous classes: IntractiveItem, ...Command, ...Menu
   but with one parameter. They are called InteractiveItem_Par1,
   InteractiveCommand_Par1, InteractiveMenu_Par1.

   The difference in comparison with interactive.hh is that there is
   an additional argument in operators()(..): const Par1& p1.

   E.g. to start menu:
   void InteractiveMenu_1Par::operator()
     (list<string>& args, istream* is, const Par1& p1);
   
   Inserting to the menu can be done with the same operator()(const string&)
   and operator[](const string&) as in the case without parameters.

   Finally to make an interactive command with one parameter, which can
   be inserted to the menu, use:
   template <class F, class Par1>
   inline MakeInteractiveCommand_1Par<F, Par1>* make_interactive_command_1par
   (F f, Par1*);
   it returns "new" MakeInteractiveCommand_1Par<F, Par1>(f); which is the 
   object derived from InteractiveCommand_1Par. 
   Your function object F should have an operator(), so that the expression
   f(is, p1); ("is" is of type istream*, p1 - Par1) is valid.
*/

   

// Author V.Balagura (balagura@mail.desy.de)   (23.04.98).

#ifndef _interactive_1par_hh
#define _interactive_1par_hh 1

#include <string>
#include <map>
#include <list>
#include "ptset/ptr_wrap.hh"
#include "ptset/global_error.hh"
#include "ptset/binder.hh"
#include "ptset/interactive.hh"

template <class Par1>
class InteractiveItem_1Par : public InteractiveItemBase {
 public:
  virtual void operator()(list<string>& args, istream* is, const Par1& p1) = 0;
  virtual InteractiveItem_1Par* clone() const = 0;
};
/* gcc version 2.7.2.3.f.1
   produces "Internal compiler error" if 
   class InteractiveItem_1Par<> is not derived from InteractiveItemBase 
   and all methods of InteractiveItemBase are defined directly in 
   InteractiveItem_1Par<>. This is due to the virtual destructor
   virtual ~InteractiveItem_1Par() {}
   It seems gcc version 2.7.2.3.f.1 does not like virtual destructors
   when they are defined in base TEMPLATED classes.
   gcc version egcs-2.90.27 980315 (egcs-1.0.2 release) works fine
*/
/* gcc 2.7.2.3.f.1:
   template member operator()(...) can not be defined not inlined 
   outside the class body
*/

template <class Par1>
class InteractiveCommand_1Par : public InteractiveItem_1Par<Par1> {
 public:
  virtual void operator()(list<string>& args, istream* is, const Par1& p1) {
    if ( !args.empty() ) {
      string s = "wrong command: " + *(args.begin());
      if ( args.size() > 1 ) s += " ...";
      global_error()(s);
    }
    else operator()(is, p1);
  }
  virtual void operator()(istream* is, const Par1& p1) = 0;
};

template <class Par1>
class InteractiveMenu_1Par 
: public InteractiveItem_1Par<Par1>, 
  public map<string, PtrWrapper<InteractiveItem_1Par<Par1> >, less<string> > {
  public:
  InteractiveMenu_1Par() { is_menu_ = true; }
  virtual InteractiveItem_1Par<Par1>* clone() const { 
    return new InteractiveMenu_1Par<Par1>(*this); 
  }
  virtual void operator()(list<string>& args, istream* is, const Par1& p1) {
    if ( args.empty() ) {
      global_error()("end of input where command was expected");
      return;
    }
    iterator iter_found;
    if ( (iter_found=find(args.front())) == end() ) {
      iterator iter_match;
      int n_matches = 0;
      for ( iter_match=begin(); iter_match!=end(); ++iter_match ) {
	if ( (*iter_match).first.compare(args.front(), 0,
					 args.front().size()) == 0 ) {
	  if ( n_matches > 0 ) {
	    global_error()("ambiguous command: " + args.front());
	    return;
	  }
	  iter_found = iter_match;
	  ++n_matches;
	}
      }
      if ( n_matches == 0 ) {
	global_error()("unrecognized command: " + args.front());
	return;
      }
    }
    InteractiveItem_1Par<Par1>& ii = *((*iter_found).second)();
    args.pop_front();
    ii(args, is, p1);
    if ( global_error() ) global_error() << (*iter_found).first;
  }
  // use it in the form menu("name") = new Smth
  // if element with "name" is found it is deleted and substituted by new Smth
  // otherwise new Smth is inserted to the map
  PtrWrapper<InteractiveItem_1Par<Par1> >& operator()(const string& s) {
    return map<string, PtrWrapper<InteractiveItem_1Par<Par1> >, 
      less<string> >::operator[](s);
  }
  // use it in the form menu["name1"]["name2"]("command_name") = new Smth
  // operator["name"] assumes the object marked as "name" is of 
  // InteractiveMenu_1Par type (if not this will produce run time error)
  InteractiveMenu_1Par<Par1>& operator[](const string& s) {
    PtrWrapper<InteractiveItem_1Par<Par1> >& pi = (*this)(s);
    if ( pi() == 0 ) pi = new InteractiveMenu_1Par<Par1>();
    else if ( ! pi()->is_menu() ) {
      global_error()
	("error in program: InteractiveItem_1Par corresponding to the key \""
	 + s + "\"\n is not an InteractiveMenu_1Par, but it is used in"
	 "\nexpression (some_interactive_menu_1par)[\"" + 
	 s + "\"]\n"
	 "which returns InteractiveMenu_1Par& of this item.\n"
	 "   Look at the codes in your program.");
      return *((InteractiveMenu_1Par<Par1>*)NULL);
    }
    return *((InteractiveMenu_1Par<Par1>*)pi());
  }
};

/* ----- makes InteractiveCommand_1Par<Par1> from function object ----- */
template <class F, class Par1>
class MakeInteractiveCommand_1Par : public InteractiveCommand_1Par<Par1> {
 public:
  F function;

  MakeInteractiveCommand_1Par(F f) : function(f) {}
  virtual void operator()(istream* is, const Par1& p1) { function(is, p1); }
  virtual InteractiveItem_1Par<Par1>* clone() const { 
    return new MakeInteractiveCommand_1Par(*this);
  }
  MakeInteractiveCommand_1Par* insert_to(InteractiveMenu_1Par<Par1>& menu, 
					 const string& command_name) {
    menu(command_name) = this;
    return this;
  }				       
};
template <class F, class Par1>
inline MakeInteractiveCommand_1Par<F, Par1>* make_interactive_command_1par
(F f, Par1*) { 
  return new MakeInteractiveCommand_1Par<F, Par1>(f); 
}

/* Does not work with  gcc 2.7.2.3.f.1:: */
/* F f; Par1 par1;
   expression f(istream*, par); should be valid */
/*
template <class Par1, class F>
void insert(InteractiveMenu_1Par<Par1>& menu, const string& name, F f) {
  menu(name) = make_interactive_command_1par(f, (Par1*)0);
} 
*/ 

#endif
