/* Collection of classes describing geometrical objects:
     Point   x,y: 2D; r(), r2();
                 functions: distance(p1,p2), abs(p), 
                            bool x_and_y_less(p1,p2),
			    bool x_and_y_less_or_equal(p1,p2)
			    minimum(p1,p2) -  Point(min(x1,x2),min(y1,y2))
			    maximum(p1,p2)

     Grid    x,y, Point base_offset
             Grid() {}
             Grid(double step_x, double step_y, const Point& any_node)
	     Grid(const Point& dp, const Point& any_node)
	     Point step()          const { return Point(x,y); }
	     Grid  shift_by_half() const
	           functions:  Point floor(p,grid), ceil(p,grid), 
	                             offset(p,grid) // p-floor(p,grid)

     Index   ix,iy
             Index() {}
	     Index(int iX, int iY)
	     Index(const Point& p, const Grid& grid)
	     Index(const Point& p, const Point& dp) //base_offset==0
	     functions: p1 = ind*p or p1 = p*ind //(index.ix*p.x,index.iy*p.y)
	                p1 = ind*grid // +base_offset
			Point grid_center(ind, grid)
    
     Point3D    x,y,z
                Point3D() {}
		Point3D(X, Y, Z)
		r(); Point xy();
		functions:  distance(p1,p2)
	                    bool x_y_and_z_less(p1,p2)
     Rectangle  Point a,b
                Rectangle() {}
		Rectangle(any_corner, opposite_corner);  
		area(); *=,*,/=,/ (double x)
		bool contains(p);  // including boundaries
		bool is_inside(p); // only inner region
		Point corners(int i); //low-left, lr, ul, ur
		Point center();

     template<class Operation>
     void for_each_node(const Rectangle& cover, const Grid& grid,
                        const Operation& op);


     Growing_Rectangle : public Rectangle
                         Growing_Rectangle()
			 Growing_Rectangle(const Rectangle& rectangle)
			 void add_point(p);
*/
// Author V.Balagura (balagura@mail.desy.de)   (09.04.98).

#ifndef _point_hh
#define _point_hh 1

#include <math.h>
#include <iostream.h>
#include <iomanip.h>  // for ostream& operator<< (ostream&, const Point&)
                      // and ostream& operator<< (ostream&, const Index&)
#include <string>
#include <algorithm> // for min, max
#include "ptset/io_binary.hh"
// #include "ptset/global_error.hh"

class Point {
public:
  double x, y;
  Point() { }
  Point(double X, double Y) : x(X), y(Y) { }
  double r2() {	return x*x + y*y;   }
  double r()  {	return hypot(x, y); }
  Point& operator += (const Point& p) { x += p.x; y += p.y; return *this; }
  Point& operator -= (const Point& p) { x -= p.x; y -= p.y; return *this; }
  Point& operator *= (double a) { x *= a; y *= a; return *this; }
  Point& operator /= (double a) { x /= a; y /= a; return *this; }

  friend ostream& operator<< (ostream& os, const Point& p);
  friend istream& operator>> (istream& is, Point& p);
};
inline bool operator==(const Point& p1, const Point& p2) {
  return ( p1.x==p2.x && p1.y==p2.y );
}
inline Point operator+ (const Point& p1, const Point& p2) {
  return Point(p1.x+p2.x, p1.y+p2.y);
}
inline Point operator- (const Point& p1, const Point& p2) {
  return Point(p1.x-p2.x, p1.y-p2.y);
}
inline
Point operator * (double a, const Point& p) { return Point(a*p.x, a*p.y); }
inline
Point operator * (const Point& p, double a) { return Point(a*p.x, a*p.y); }
inline
Point operator / (double a, const Point& p) { return Point(p.x/a, p.y/a); }
inline
Point operator / (const Point& p, double a) { return Point(p.x/a, p.y/a); }
inline Point operator - (const Point& p) { return Point(-p.x, -p.y); }
inline double distance(const Point& p1, const Point& p2) {
	return (p2 - p1).r();
}
inline Point abs(const Point& p) { return Point(fabs(p.x), fabs(p.y)); }

inline bool x_and_y_less(const Point& p1, const Point& p2) {
  return (p1.x < p2.x && p1.y < p2.y);
}
inline bool x_and_y_less_or_equal(const Point& p1, const Point& p2) {
  return (p1.x <= p2.x && p1.y <= p2.y);
}
inline Point minimum(const Point& p1, const Point& p2) {
  return Point(min(p1.x, p2.x), min(p1.y, p2.y));
}
inline Point maximum(const Point& p1, const Point& p2) {
  return Point(max(p1.x, p2.x), max(p1.y, p2.y));
}
inline ostream& operator<< (ostream& os, const Point& p) {
  // long flags     = os.setf(ios::showpoint);
  // os.setf(ios::fixed, ios::floatfield);
  // int  precision = os.precision(2);
  //  os << "(" << setw(6) << p.x << "," << setw(6) << p.y << ")";
  os << "(" << p.x << "," << p.y << ")";
  // os.precision(precision);
  // os.flags(flags);
  return os;
}
inline istream& operator>> (istream& is, Point& p) {
  char c;
  if ( !(is >> c) ) return is;
  if ( c != '(' ) { is.setstate(ios::failbit); return is; }
  if ( !(is >> p.x) ) return is;
  if ( !(is >> c) ) return is;
  if ( c != ',' ) { is.setstate(ios::failbit); return is; }
  if ( !(is >> p.y) ) return is;
  if ( !(is >> c) ) return is;
  if ( c != ')' ) { is.setstate(ios::failbit); return is; }
  return is;
}

class Grid {
public:
  double x, y;
  Point base_offset;
  Grid() { }
  Grid(double step_x, double step_y, const Point& any_node)
    : x(step_x), y(step_y) {
      base_offset = any_node -
	Point(floor(any_node.x/x)*x, floor(any_node.y/y)*y);
  }
  Grid(const Point& dp, const Point& any_node)
    : x(dp.x), y(dp.y) {
      base_offset = any_node -
	Point(floor(any_node.x/x)*x, floor(any_node.y/y)*y);
  }
  Point step()          const { return Point(x,y); }
  Grid  shift_by_half() const { return Grid(x,y, base_offset+Point(x/2,y/2)); }

  friend ostream& operator<< (ostream& os, const Grid& grid);
  friend istream& operator>> (istream& is, const Grid& grid);
};
inline ostream& operator<< (ostream& os, const Grid& grid) {
  os << "grid " << grid.x << "x" << grid.y << ", offset " << grid.base_offset;
  return os;
}
inline istream& operator>> (istream& is, Grid& grid) {
  string s;
  char c;
  if ( !(is >> s) ) return is;
  if ( s != "grid" ) { is.setstate(ios::failbit); return is; }
  if ( !(is >> grid.x) ) return is;
  if ( !(is >> c) ) return is;
  if ( c != 'x' ) { is.setstate(ios::failbit); return is; }
  if ( !(is >> grid.y) ) return is;
  if ( !(is >> c) ) return is;
  if ( c != ',' ) { is.setstate(ios::failbit); return is; }
  if ( !(is >> s) ) return is;
  if ( s != "offset" ) { is.setstate(ios::failbit); return is; }
  Point any_node;
  if ( !(is >> any_node) ) return is;
  grid.base_offset = any_node - Point(floor(any_node.x/grid.x)*grid.x, 
				      floor(any_node.y/grid.y)*grid.y);
  return is;
}
inline Point floor(const Point& p, const Grid& grid) {
  double dx = grid.x, dy = grid.y;
  Point p0 = grid.base_offset;
  Point dp = p - p0;
  return Point(floor(dp.x/dx)*dx, floor(dp.y/dy)*dy) + grid.base_offset;
}
inline Point ceil(const Point& p, const Grid& grid) {
  double dx = grid.x, dy = grid.y;
  Point p0 = grid.base_offset;
  Point dp = p - p0;
  return Point(ceil(dp.x/dx)*dx, ceil(dp.y/dy)*dy) + grid.base_offset;
}
inline Point offset(const Point& p, const Grid& grid) {
  return p - floor(p, grid);
}

class Index {
  void init(const Point& p, const Point& dp) {
    ix = int(floor(p.x / dp.x));
    iy = int(floor(p.y / dp.y));
  }
public:
  int ix, iy;
  Index() { }
  Index(int iX, int iY) : ix(iX), iy(iY) { }
  Index(const Point& p, const Grid& grid) { 
    init(p - grid.base_offset, Point(grid.x, grid.y));
  }
  Index(const Point& p, const Point& dp) { init(p, dp); }
  Index& operator += (const Index& ind) { 
    ix += ind.ix; iy += ind.iy; return *this; 
  }
  Index& operator -= (const Index& ind) { 
    ix -= ind.ix; iy -= ind.iy; return *this; 
  }
  bool operator< (const Index& ind) const {
    return ( iy<ind.iy || (iy==ind.iy && ix<ind.ix) );
  }
  bool operator==(const Index& ind) const {
    return ( ix==ind.ix && iy==ind.iy );
  }
};
inline Index operator+ (const Index& ind1, const Index& ind2) {
  return Index(ind1.ix+ind2.ix, ind1.iy+ind2.iy);
}
inline Index operator- (const Index& ind1, const Index& ind2) {
  return Index(ind1.ix-ind2.ix, ind1.iy-ind2.iy);
}
inline Index operator - (const Index& ind) { 
  return Index(-ind.ix, -ind.iy); 
}
inline ostream& operator<< (ostream& os, const Index& ind) {
  return os << '[' << setw(2) << ind.ix << ',' << setw(2) << ind.iy << ']';
}
inline istream& operator>> (istream& is, Index& ind) {
  char c;
  if ( !(is >> c) )      return is;
  if ( c != '[' ) { is.setstate(ios::failbit); return is; }
  if ( !(is >> ind.ix) ) return is;
  if ( !(is >> c) )      return is;
  if ( c != ',' ) { is.setstate(ios::failbit); return is; }
  if ( !(is >> ind.iy) ) return is;
  if ( !(is >> c) )      return is;
  if ( c != ']' ) { is.setstate(ios::failbit); return is; }
  return is;
}

inline Point operator*(const Index& index, const Point& p) {
  return Point(index.ix * p.x,
	       index.iy * p.y );
}
inline Point operator*(const Point& p, const Index& index) {
  return index*p;
}
inline Point operator*(const Index& ind, const Grid& grid) {
  return ind * grid.step() + grid.base_offset;
}
inline Point grid_center(const Index& ind, const Grid& grid) {
  return Point(
	       (ind.ix+0.5) * grid.x + grid.base_offset.x,
	       (ind.iy+0.5) * grid.y + grid.base_offset.y
	       );
}
    

class Point3D {
public:
  double x, y, z;
  Point3D() { }
  Point3D(double X, double Y, double Z) : x(X), y(Y), z(Z) { }
  double r()  {	return sqrt(x*x + y*y + z*z); }
  Point xy() { return Point(x, y); }
  Point3D& operator += (const Point3D& p) { 
    x += p.x; y += p.y; z += p.z; return *this; 
  }
  Point3D& operator -= (const Point3D& p) {
    x -= p.x; y -= p.y; z -= p.z; return *this; 
  }
  Point3D& operator *= (double a) { x *= a; y *= a; z *= a; return *this; }
  Point3D& operator /= (double a) { x /= a; y /= a; z /= a; return *this; }
};
inline bool operator==(const Point3D& p1, const Point3D& p2) {
  return ( p1.x==p2.x && p1.y==p2.y && p1.z==p2.z );
}
inline Point3D operator+ (const Point3D& p1, const Point3D& p2) {
  return Point3D(p1.x+p2.x, p1.y+p2.y, p1.z+p2.z);
}
inline Point3D operator- (const Point3D& p1, const Point3D& p2) {
  return Point3D(p1.x-p2.x, p1.y-p2.y, p1.z-p2.z);
}
inline Point3D operator * (double a, const Point3D& p) {
	return Point3D(a*p.x, a*p.y, a*p.z);
}
inline Point3D operator * (const Point3D& p, double a) {
	return Point3D(a*p.x, a*p.y, a*p.z);
}
inline Point3D operator - (const Point3D& p) {
	return Point3D(-p.x, -p.y, -p.z);
}
inline double distance(const Point3D& p1, const Point3D& p2) {
	return (p2 - p1).r();
}

inline bool x_y_and_z_less(const Point3D& p1, const Point3D& p2) {
  return (p1.x < p2.x && p1.y < p2.y && p1.z < p2.z);
}
inline ostream& operator<< (ostream& os, const Point3D& p) {
	long flags     = os.setf(ios::showpoint);
	os.setf(ios::fixed, ios::floatfield);
	int  precision = os.precision(2);
	os << "("
	   << setw(6) << p.x << ","
	   << setw(6) << p.y << ","
	   << setw(6) << p.z << ")";
	os.precision(precision);
	os.flags(flags);
	return os;
}
inline istream& operator>> (istream& is, Point3D& p) {
  char c;
  if ( !(is >> c) ) return is;
  if ( c != '(' ) { is.setstate(ios::failbit); return is; }
  if ( !(is >> p.x) ) return is;
  if ( !(is >> c) ) return is;
  if ( c != ',' ) { is.setstate(ios::failbit); return is; }
  if ( !(is >> p.y) ) return is;
  if ( !(is >> c) ) return is;
  if ( c != ',' ) { is.setstate(ios::failbit); return is; }
  if ( !(is >> p.z) ) return is;
  if ( !(is >> c) ) return is;
  if ( c != ')' ) { is.setstate(ios::failbit); return is; }
  return is;
}

class Rectangle {
public:
  Point a, b;
  Rectangle() { }
  Rectangle(const Point& any_corner, const Point& opposite_corner);  
  double area() const { Point delta = b - a; return delta.x * delta.y; }
  void operator *= (double x) { a *= x; b *= x; }
  void operator /= (double x) { a /= x; b /= x; }
  bool contains (const Point& p) const; // including boundaries
  bool is_inside(const Point& p) const; // only inner region
  Point corners(int i) const;
  Point center() const { return (a+b)/2; }

  friend ostream& operator<< (ostream& os, const Rectangle& rect);
  friend istream& operator>> (istream& is, Rectangle& rect);
};
inline Rectangle::Rectangle(const Point& any_corner, 
			    const Point& opposite_corner) {
  if (any_corner.x < opposite_corner.x) {
    a.x = any_corner.x;
    b.x = opposite_corner.x;
  }
  else {
    a.x = opposite_corner.x;
    b.x = any_corner.x;
  }
  if (any_corner.y < opposite_corner.y) {
    a.y = any_corner.y;
    b.y = opposite_corner.y;
  }
  else {
    a.y = opposite_corner.y;
    b.y = any_corner.y;
  }
}
inline bool Rectangle::contains(const Point& p) const {
  return (  x_and_y_less_or_equal(a, p) &&  x_and_y_less_or_equal(p, b));
}
inline bool Rectangle::is_inside(const Point& p) const {
  return (  x_and_y_less(a, p) &&  x_and_y_less(p, b));
}

inline Point Rectangle::corners(int i) const { 
  switch (i) {
  case 0: return a;
  case 1: return Point(b.x, a.y);
  case 2: return Point(a.x, b.y);
  case 3: return b;
  default: return b;
  }
}

inline ostream& operator<< (ostream& os, const Rectangle& rect) {
  return os << "[" << rect.a << "-" << rect.b << "]";
}
inline istream& operator>> (istream& is, Rectangle& rect) {
  char c;
  if ( !(is >> c) ) return is;
  if ( c != '[' ) { is.setstate(ios::failbit); return is; }
  if ( !(is >> rect.a) ) return is;
  if ( !(is >> c) ) return is;
  if ( c != '-' ) { is.setstate(ios::failbit); return is; }
  if ( !(is >> rect.b) ) return is;
  if ( !(is >> c) ) return is;
  if ( c != ']' ) { is.setstate(ios::failbit); return is; }
  return is;
}

inline Rectangle operator * (double a, const Rectangle& r) { 
  return Rectangle(a*r.a, a*r.b);
}
inline Rectangle operator * (const Rectangle& r, double a) { return a*r; }
inline
Rectangle operator / (const Rectangle& r, double a) {
  return Rectangle(r.a/a, r.b/a);
}

template<class Operation>
void for_each_node(const Rectangle& cover, const Grid& grid,
		   const Operation& op) {
  Point pmin = ceil(cover.a, grid), p;
  for (p.y = pmin.y; p.y < cover.b.y; p.y += grid.y) 
    for (p.x = pmin.x; p.x < cover.b.x; p.x += grid.x) op(p);
}

class Growing_Rectangle : public Rectangle {
  bool empty;
public:
  Growing_Rectangle() : empty(true) { }
  Growing_Rectangle(const Rectangle& rectangle) : Rectangle(rectangle), 
    empty(true) { }
  void add_point(const Point& p);
};
inline void Growing_Rectangle::add_point(const Point& p) {
  if (empty) {
    a = b = p;
    empty = false;
  }
  else {
    if (      p.x < a.x ) a.x = p.x;
    else if ( b.x < p.x ) b.x = p.x;
    if (      p.y < a.y ) a.y = p.y;
    else if ( b.y < p.y ) b.y = p.y;
  }
}
    
#endif
