/* Defines SetupFigureBase, SetupFigure<..> classes and 
   setup_figure_maker(...) functions.

   The prefix "Setup" in SetupFigure... was chosen to avoid conflict
   with other "Figure"s which can be defined in other programs.
   (Figure is quite frequently used name).

   SetupFigureBase is a pure virtual base class for all figures. 
   It has 
     virtual bool SetupFigureBase::contains(Point p) = 0;
   function defining figure.
   Also it has 
     virtual bool contains(Point p) = 0;
   so that PtrWrapper<SetupFigureBase> can be constructed.

   SetupFigure<F> is derived from SetupFigureBase. 
   F is a function object so that experssion f(p, ptr_bool) is valid.
   Here f is of type F, p - Point, ptr_bool - bool*.
   It returns yes/no in *ptr_bool depending whether the point is inside
   the figure to be defined.

   Usage: make_setup_figure(f); returns SetupFigure<F> (types deduced).

   Another tool is setup_figure_maker(a1, a2, ..., a<N>, &fig); (N<=5).
   It returns function object MakeSetupFigure_<N>Arg.
   It can be used as:
     SetupFigureBase** fig;
     setup_figure_maker(f_contains, (Arg1*)0, ..., (Arg<N>*)0)
                           (a1, ..., a<N>, fig);
   last command sets *fig to "new" SetupFigureBase. SetupFigure is
   defined by the f_contains function object.
   See below for more details.
   */
   

// Author V.Balagura (balagura@mail.desy.de)   (09.04.98).

#ifndef _setup_figure_hh
#define _setup_figure_hh 1

#include "ptset/point.hh"
#include "ptset/binder.hh"

class SetupFigureBase {
 public:
  virtual bool contains(Point p) = 0;
  virtual SetupFigureBase* clone() const = 0;
  virtual ~SetupFigureBase() {}
};

/* ----- makes SetupFigureBase from function object ----- */
/* ----- F f; Point p; bool b; f(p, &b) should be valid ----- */
template <class F>
class SetupFigure : public SetupFigureBase {
 public:
  F function;

  SetupFigure(F f) : function(f) {}
  virtual bool contains(Point p) {
    bool tmp;
    function(p, &tmp);
    return tmp;
  }
  virtual SetupFigureBase* clone() const {
    return new SetupFigure<F>(*this);
  }
};
template <class F>
inline SetupFigureBase* make_setup_figure(F f) {
  return new SetupFigure<F>(f);
}

/* -------------------- MakeSetupFigure_<N>Arg --------------------
   Can be created from any function object which has the arguments:
   (n <= 5 arbitrary, Point, bool*).

   Has void operator()(Arg1 a1, ..., Arg<N> a<N>, SetupFigureBase** fig);
   it binds all args to the given values, thus gets function object
   with arguments (Point, bool*), makes "new" figure from it and
   sets *fig to the address of this new figure.

   Usage: setup_figure_maker(f, (Arg1*)0, ..., (Arg<N>*)0) returns
   corrsponding MakeSetupFigure_<N>Arg.

   E.g. suppose you have f_contains(A1 a1, ..., A<N> a<N>, 
                                    Point p, bool* is_inside);
   then:
     SetupFigureBase** fig;
     setup_figure_maker(f_contains, (Arg1*)0, ..., (Arg<N>*)0)
                         (a1, ..., a<N>, fig);
   sets *fig to "new" SetupFigureBase. This SetupFigureBase represents
   the figure which is made of all the points which satisfy the condition:
   if one puts it in: f_contains(a1, ..., a<N>, this_point, 
                                 bool* is_inside),
   *is_inside will be set to true. 
 */
/* -------------------- 0 args -------------------- */
template <class F>
struct MakeSetupFigure_0Arg {
  F function;
  MakeSetupFigure_0Arg(F f) : function(f) {}
  void operator()(SetupFigureBase** fig) {
    *fig = make_SetupFigure(function);
  }
};
template <class F>
inline MakeSetupFigure_0Arg<F> setup_figure_maker(F f) {
  return MakeSetupFigure_0Arg<F>(f);
}
/* -------------------- 1 arg -------------------- */
template <class F, class A1>
struct MakeSetupFigure_1Arg {
  F function;
  MakeSetupFigure_1Arg(F f) : function(f) {}
  void operator()(A1 a1, SetupFigureBase** fig) {
    *fig = make_setup_figure(bind1(function, a1, (Point*)0, (bool**)0));
  }
};
template <class F, class A1>
inline MakeSetupFigure_1Arg<F, A1> setup_figure_maker(F f, A1*) {
  return MakeSetupFigure_1Arg<F, A1>(f);
}
/* -------------------- 2 args -------------------- */
template <class F, class A1, class A2>
struct MakeSetupFigure_2Arg {
  F function;
  MakeSetupFigure_2Arg(F f) : function(f) {}
  void operator()(A1 a1, A2 a2, SetupFigureBase** fig) {
    *fig = make_setup_figure
      (bind1(bind1(function, a1, (A2*)0, (Point*)0, (bool**)0),
	     a2, (Point*)0, (bool**)0));
  }
};
template <class F, class A1, class A2>
inline MakeSetupFigure_2Arg<F, A1, A2> setup_figure_maker(F f, A1*, A2*) {
   return MakeSetupFigure_2Arg<F, A1, A2>(f);
}
/* -------------------- 3 args -------------------- */
template <class F, class A1, class A2, class A3>
struct MakeSetupFigure_3Arg {
  F function;
  MakeSetupFigure_3Arg(F f) : function(f) {}
  void operator()(A1 a1, A2 a2, A3 a3, SetupFigureBase** fig) {
    *fig = make_setup_figure
      (bind1(bind1(bind1(function, a1, (A2*)0, (A3*)0, (Point*)0, (bool**)0), 
		   a2, (A3*)0, (Point*)0, (bool**)0),
	     a3, (Point*)0, (bool**)0));
  }
};
template <class F, class A1, class A2, class A3>
inline MakeSetupFigure_3Arg<F, A1, A2, A3> 
setup_figure_maker(F f, A1*, A2*, A3*) {
  return MakeSetupFigure_3Arg<F, A1, A2, A3>(f);
}
/* -------------------- 4 args -------------------- */
template <class F, class A1, class A2, class A3, class A4>
struct MakeSetupFigure_4Arg {
  F function;
  MakeSetupFigure_4Arg(F f) : function(f) {}
  void operator()(A1 a1, A2 a2, A3 a3, A4 a4, SetupFigureBase** fig) {
    *fig = make_setup_figure
      (bind1(bind1(bind1(bind1(function, a1, (A2*)0, (A3*)0, (A4*)0, 
			       (Point*)0, (bool**)0),
			 a2, (A3*)0, (A4*)0, (Point*)0, (bool**)0),
		   a3, (A4*)0, (Point*)0, (bool**)0), 
	     a4, (Point*)0, (bool**)0));
  }
};
template <class F, class A1, class A2, class A3, class A4>
inline MakeSetupFigure_4Arg<F, A1, A2, A3, A4> 
setup_figure_maker(F f, A1*, A2*, A3*, A4*) {
  return MakeSetupFigure_4Arg<F, A1, A2, A3, A4>(f);
}
/* -------------------- 5 args -------------------- */
template <class F, class A1, class A2, class A3, class A4, class A5>
struct MakeSetupFigure_5Arg {
  F function;
  MakeSetupFigure_5Arg(F f) : function(f) {}
  void operator()(A1 a1, A2 a2, A3 a3, A4 a4, A5 a5, 
		  SetupFigureBase** fig) {
    *fig = make_setup_figure
      (bind1(bind1(bind1(bind1(bind1(function, a1, (A2*)0, (A3*)0, (A4*)0,
				     (A5*)0, (Point*)0, (bool**)0), 
			       a2, (A3*)0, (A4*)0, (A5*)0, (Point*)0,
			       (bool**)0),
			 a3, (A4*)0, (A5*)0, (Point*)0, (bool**)0), 
		   a4, (A5*)0, (Point*)0, (bool**)0), 
	     a5, (Point*)0, (bool**)0));
  }
};
template <class F, class A1, class A2, class A3, class A4, class A5>
inline MakeSetupFigure_5Arg<F, A1, A2, A3, A4, A5> 
setup_figure_maker(F f, A1*, A2*, A3*, A4*, A5*) {
  return MakeSetupFigure_5Arg<F, A1, A2, A3, A4, A5>(f);
}

#endif
