// 
//  Author:  Rainer Mankel                     Created:  29-Sep-98
//           Humboldt University Berlin        Modified: 29-Mar-99
//  Email:   mankel@ifh.de                     HERA-B Collaboration
//
// Modifications:
// - Introduced list of active walls   14-Aug-00 Wouter Hulsbergen 
//   use of all gwals in the vds                 17/05/01 A. Spiridonov
//

#include <iostream>
#include <stdio.h>
#include <algo.h>
#include "arte/arte.hh"
#include "arte/ArteMessage.hh"
#include <arte/artefunc.h>
#include "rang/Geom.hh"
#include "rang/GeWa.hh"
#include "rang/Gede.hh"
#include "rang/Gwal.hh"
#include "rang/sortzv.hh"

#define GWALL_SELECT
#undef GWALL_NOWALLS  

vector<rangerVolume*> rangerGeom::volumes;
vector<vector<rangerVolume*>::iterator> rangerGeom::vGede;
vector<vector<rangerVolume*>::iterator> rangerGeom::vGwal;

vector<rangerVolume*> rangerGeom::activeVolumes;
int rangerGeom::firstPlane=0;
vector<vector<rangerVolume*>::const_iterator> rangerGeom::planeToUpstrActive;

void rangerGeom::update() {
  reset();
  build();
  gsort();
  arteMsg << "rangerGeom::update: geometry table (re)built" 
	  << " with " << volumes.size() <<  " volumes" << endl
	  << " and "  << activeVolumes.size() << " active volumes." << endl ;
  arteMsg.warn();
}

void rangerGeom::condUpdate() {
  unsigned int nVol = ArteTable<GEDE>::nrow() + ArteTable<GWAL>::nrow();
  if (nVol != volumes.size()) update();
}

void rangerGeom::reset() {
  for (vector<rangerVolume*>::iterator vIt=volumes.begin();
       vIt != volumes.end(); ++vIt) delete (*vIt);
  volumes.erase(volumes.begin(),volumes.end());
  vGede.erase(vGede.begin(),vGede.end());
  vGwal.erase(vGwal.begin(),vGwal.end());
  activeVolumes.clear() ;
  planeToUpstrActive.clear() ;
}

void rangerGeom::build() {
  volumes.clear() ;
  volumes.reserve(ArteTable<GEDE>::nrow()+ArteTable<GWAL>::nrow());
  
  for ( ArteTable<GEDE>::iterator gede = ArteTable<GEDE>::begin();
	gede != ArteTable<GEDE>::end(); ++gede ) {
    rangerGede* Gede = new rangerGede( gede ); // allocate new Gede
    rangerGeWa* GeWa = Gede;
    rangerVolume* vol = rangerVolume::makeVolume( GeWa );
    volumes.push_back(vol);

  }

  for ( ArteTable<GWAL>::iterator gwal = ArteTable<GWAL>::begin();
	gwal != ArteTable<GWAL>::end(); ++gwal ) {
    rangerGwal* Gwal = new rangerGwal( gwal ); // allocate new Gwal
    rangerGeWa* GeWa = Gwal;
    rangerVolume* vol = rangerVolume::makeVolume( GeWa );
    volumes.push_back(vol);
  }
}

void rangerGeom::gsort() {

  sort(volumes.begin(),volumes.end(), compareCall( &rangerVolume::zMean ) );

  // initialize the vGede and vGwal pointers
  vGede.reserve(ArtePointer<GEDE>::nrow());
  for ( ArteTable<GEDE>::iterator gede = ArteTable<GEDE>::begin();
	gede != ArteTable<GEDE>::end(); ++gede ) {
      vGede.push_back(volumes.end());
  }
  vGwal.reserve(ArtePointer<GWAL>::nrow());
  for ( ArteTable<GWAL>::iterator gwal = ArteTable<GWAL>::begin();
	gwal != ArteTable<GWAL>::end(); ++gwal ) {
      vGwal.push_back(volumes.end());
  }
  for (vector<rangerVolume*>::iterator vi=volumes.begin();
       vi != volumes.end(); ++vi) {
    if ((*vi)->plNumber() > 0) {
      if ((*vi)->plNumber() > (int) ArtePointer<GEDE>::nrow()) {
	cout << "warning: gede::nrow=" << ArtePointer<GEDE>::nrow()
	     << " pln=" << (*vi)->plNumber() << endl;
      }
      vGede[(*vi)->plNumber()-1] = vi;
    }
    else {
      if ((*vi)->plNumber() < 0) {
	if (-(*vi)->plNumber() > (int) ArtePointer<GWAL>::nrow()) {
	  cout << "warning: gwal::nrow=" << ArtePointer<GWAL>::nrow()
	       << " pln=" << (*vi)->plNumber() << endl;
	}
	vGwal[-(*vi)->plNumber()-1] = vi;
      }
      else {
	cout << "plNumber zero!" << endl;
      }
    }
  }

  // create vector of active volumes
  activeVolumes.clear() ;
  // this reserve is necessary, otherwise pointers are not correctly
  // set. it is a bit a waste of space though ...
  activeVolumes.reserve(volumes.size()) ;
  // these are pointers to the first active volume with a lower z value
  planeToUpstrActive.clear() ;
  firstPlane = -ArteTable<GWAL>::size() ;
  int nplanes = ArteTable<GWAL>::size()+ArteTable<GEDE>::size()+1 ;
  planeToUpstrActive.resize(nplanes,activeVolumes.begin()) ;
  
  for(vector<rangerVolume*>::iterator vi=volumes.begin();
      vi != volumes.end(); ++vi) {
    int plane = (*vi)->plNumber() ;

    // set pointer of this plane to first active volume with lower z
    if(!activeVolumes.empty())
      planeToUpstrActive[plane-firstPlane] = activeVolumes.end() - 1 ;
    
    // here the code to de/activate volume
    bool active = true;
    
#ifdef GWALL_SELECT
    rangerGeWa* gewa = const_cast<rangerGeWa*> ((*vi)->getGeWa()) ;
    
    // remove all objects with very low ms:
    // (there are no gedes below 1.e-7)
    active = active && (*vi)->ms() > 1.e-8 ;

    // remove rich plane that exists twice 
    active = active && plane != 2230 ;
    
    // remove very small planes
    active = active && !( gewa->shp()==1 &&
     			  (gewa->xmax()-gewa->xmin())*
			  (gewa->ymax()-gewa->ymin())<1.) ;
    
    // remove all otr/itr gwals after vds and put fudge factor for gedes
    const float fudge = 4;
    bool istracker = 
      (gewa->zmin()>220  && gewa->zmax()<860) ||
      (gewa->zmin()>1175 && gewa->zmax()<1330) ;
    if(active && istracker) {
      if(plane>0) (*vi)->setMs((*vi)->ms()*fudge) ;
      else        active = false ;
    }
    
    // for the moment: remove all gwals in the vds (a lot of rotations)
    //bool isvds = gewa->zmin()>6 && gewa->zmax()<202 ;
    //if(isvds && gewa->shp()==1 && plane<0) active = false ;
#endif 
#ifdef GWALL_NOWALLS
    active=false ;
#endif
    // add volume to active list
    if(active) activeVolumes.push_back(*vi) ;
  }
}
