// 
//  Author:  Rainer Mankel                     Created:  29-Sep-98
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
// 
// Modifications:
// - In ::insertWalls: loop `active' volumes only
//                new  rangerTrack::fitGlobal() 
//                     trick for path in magnet
//                                      14-Aug-00 Wouter Hulsbergen 
//  check parameter fitmagn  in rangerTrack(rtra..        27/11/00 A. Spiridonov
//  upper limit for magmet < 660.                         04/12/00 A. Spiridonov
//  upper limit for magnet > 660. in rangerTrack::genCmp()21/01/01 A. Spiridonov 
// - added filling of rtra->ndof, 
//   removed chiSquare data member                        11/04/01 Wouter
//
//   float dz = (*insertBaseNode)->getState().z() - z ;   16/05/01 Wouter
//   Fixed problem which occurs only when there are two 
//   hits at exactly the same z-position.
//   Problem is solved by a simple hack: Before the planes
//   are inserted an (stl) `set' is filled with the planes 
//   that are already in the track. Then every new plane is 
//   checked against the already existing plane.
//   discard muon hits in rangerTrack(ArtePointer<RTRA>.. 20/05/01 A. Spiridonov
//

#include <iostream>
#include <list>
#include <vector>
#include <algo.h>
#include <arte/arte.hh>
#include <arte/RSEG_cmp.hh>
#include <arte/artefunc.h>
#include <arte/ArteMessage.hh>
#include "rang/Node.hh"
#include "rang/Hit.hh"
#include "rang/HitWire.hh"
#include "rang/Geom.hh"
#include "rang/Volume.hh"
#include "rang/Track.hh"
#include "rang/hfill.hh"
#include "rang/zfpar.hh"
#include "rang/RMap.hh"
#include "rang/HitMap.hh"
#include "rang/FLAGHist.hh"

#undef FULL
#undef SAVEWALLS

//-- Default constructor
rangerTrack::rangerTrack() {
  qpEstimate = 0;
  electron = 0;
  nIteration = 0;
  smoothed = 0;
  tFitMode = ModeP;
  tPartType = MUON;
  status = 0;
};

//-- constructor from RSEG
rangerTrack::rangerTrack(ArtePointer<RSEG> rseg, rangerFitMode fm,
			 rangerParticleType pType,
			 rangerRMap* rmap) {
  qpEstimate = rseg->pf;
  electron = 0;        // set according to electron hypothesis
  nIteration = 0;
  smoothed = 0;
  tFitMode = fm;
  tPartType = pType;
  status = 0;

  // the RHITs are assumed to be retrieved in the order of increasing z.
  // the nodes are stored such that the first node in the list 
  // points to highest z
  
  for ( ArteRelationIterator<RHIT> rhitIt = rseg->rhit.begin();
	rhitIt != rseg->rhit.end(); ++rhitIt ) {
    ArtePointer<RHIT> rhit(*rhitIt);
#ifdef FULL
    cout << "Current rhit=" << rhit << 
      " hitb=" << rhit->hitb << " z=" << rhit->hitb->z << endl;
#endif

    rangerNode* w = rangerNode::makeNode( rhit, fitMode() );
    if (w) {
      prepend(*w);
      if (rmap) {
	rangerHitMap hm( w,rhit );
	rmap->addHitMap( hm );
      }
    }
  }
  orderIfNeeded();
};

//-- constructor from RSEG vector
rangerTrack::rangerTrack(const vector<ArtePointer<RSEG> >& rsegVec, 
			 rangerFitMode fm,
			 rangerParticleType pType,
			 vector<rangerRMap>* rmapVec) {
  qpEstimate = 0;
  electron = 0;        // set according to electron hypothesis
  nIteration = 0;
  smoothed = 0;
  tFitMode = fm;
  tPartType = pType;
  status = 0;

  // the RHITs are assumed to be retrieved in the order of increasing z.
  // the nodes are stored such that the first node in the list 
  // points to highest z

  for (vector<ArtePointer<RSEG> >::const_iterator sIt=rsegVec.begin();
       sIt != rsegVec.end(); sIt++) {
    ArtePointer<RSEG> rseg = *sIt;
    if (rseg->pf != 0) qpEstimate = rseg->pf;
    rangerRMap rMap(rseg);
  
    vector<ArtePointer<RHIT> > rhits;
    rhits.reserve(256);
    for ( ArteRelationIterator<RHIT> rhitIt = rseg->rhit.begin();
	  rhitIt != rseg->rhit.end(); ++rhitIt ) {
      ArtePointer<RHIT> rhit(rhitIt);
      rhits.push_back(rhit);
    }

    int correctOrder = 1;
    if (rhits.size() > 1) {  // determine correct order
      if ((*rhits.begin())->hitb->z > (*rhits.rbegin())->hitb->z) {
	correctOrder = 0;
      }
    }

    if (correctOrder) {
      for (vector<ArtePointer<RHIT> >::iterator rIt=rhits.begin();
	   rIt != rhits.end(); ++rIt) {
#ifdef FULL
	cout << "Track(Vec...): current rhit=" << *rIt << 
	  " hitb=" << (*rIt)->hitb << " z=" << (*rIt)->hitb->z << endl;
#endif

	rangerNode* w = rangerNode::makeNode( *rIt, fitMode() );
	if (w) {
	  prepend(*w);
	  rangerHitMap hm( w,*rIt );
	  rMap.addHitMap( hm );
	}
      }
      if (rmapVec) rmapVec->push_back(rMap);
    }
    else {
      for (vector<ArtePointer<RHIT> >::reverse_iterator rIt=rhits.rbegin();
	   rIt != rhits.rend(); ++rIt) {
#ifdef FULL
	cout << "Track(Vec...): current rhit=" << *rIt << 
	  " hitb=" << (*rIt)->hitb << " z=" << (*rIt)->hitb->z << endl;
#endif

	rangerNode* w = rangerNode::makeNode( *rIt, fitMode() );
	if (w) {
	  prepend(*w);
	  rangerHitMap hm( w,*rIt );
	  rMap.addHitMap( hm );
	}
      }
      if (rmapVec) rmapVec->push_back(rMap);
    }
  }
  orderIfNeeded();
};


//-- constructor from RSEG vector + RCCL
rangerTrack::rangerTrack(const vector<ArtePointer<RSEG> >& rsegVec, 
			 ArtePointer<RCCL> rccl,int charge,
			 rangerFitMode fm,
			 rangerParticleType pType,
			 vector<rangerRMap>* rmapVec) {
  qpEstimate = 0;
  electron = 0;        // set according to electron hypothesis
  nIteration = 0;
  smoothed = 0;
  tFitMode = fm;
  tPartType = pType;
  status = 0;

  // the RHITs are assumed to be retrieved in the order of increasing z.
  // the nodes are stored such that the first node in the list 
  // points to highest z

  for (vector<ArtePointer<RSEG> >::const_iterator sIt=rsegVec.begin();
       sIt != rsegVec.end(); sIt++) {
    ArtePointer<RSEG> rseg = *sIt;
    if (rseg->pf != 0) qpEstimate = rseg->pf;
    rangerRMap rMap(rseg);
  
    vector<ArtePointer<RHIT> > rhits;
    rhits.reserve(256);
    for ( ArteRelationIterator<RHIT> rhitIt = rseg->rhit.begin();
	  rhitIt != rseg->rhit.end(); ++rhitIt ) {
      ArtePointer<RHIT> rhit(rhitIt);
      rhits.push_back(rhit);
    }

    int correctOrder = 1;
    if (rhits.size() > 1) {  // determine correct order
      if ((*rhits.begin())->hitb->z > (*rhits.rbegin())->hitb->z) {
	correctOrder = 0;
      }
    }

    if (correctOrder) {
      for (vector<ArtePointer<RHIT> >::iterator rIt=rhits.begin();
	   rIt != rhits.end(); ++rIt) {
#ifdef FULL
	cout << "Track(Vec...): current rhit=" << *rIt << 
	  " hitb=" << (*rIt)->hitb << " z=" << (*rIt)->hitb->z << endl;
#endif

	rangerNode* w = rangerNode::makeNode( *rIt, fitMode() );
	if (w) {
	  prepend(*w);
	  rangerHitMap hm( w,*rIt );
	  rMap.addHitMap( hm );
	}
      }
      if (rmapVec) rmapVec->push_back(rMap);
    }
    else {
      for (vector<ArtePointer<RHIT> >::reverse_iterator rIt=rhits.rbegin();
	   rIt != rhits.rend(); ++rIt) {
#ifdef FULL
	cout << "Track(Vec...): current rhit=" << *rIt << 
	  " hitb=" << (*rIt)->hitb << " z=" << (*rIt)->hitb->z << endl;
#endif

	rangerNode* w = rangerNode::makeNode( *rIt, fitMode() );
	if (w) {
	  prepend(*w);
	  rangerHitMap hm( w,*rIt );
	  rMap.addHitMap( hm );
	}
      }
      if (rmapVec) rmapVec->push_back(rMap);
    }
  }
  rangerNode* w = rangerNode::makeNode( rccl,charge,fitMode());
  if (w) prepend(*w);
  orderIfNeeded();
};


//-- constructor from RTRA
rangerTrack::rangerTrack(ArtePointer<RTRA> rtra, rangerFitMode fm,
			 rangerParticleType pType,
			 rangerRMap* rmap) {
  qpEstimate = rtra->pf;
  electron = 0;        // set according to electron hypothesis
  nIteration = 0;
  smoothed = 0;
  tFitMode = fm;
  tPartType = pType;
  status = 0;

  // the RHITs are assumed to be retrieved in the order of increasing z.
  // the nodes are stored such that the first node in the list 
  // points to highest z
  
  for ( ArteRelationIterator<HITB> hitbIt = rtra->hitb.begin();
	hitbIt != rtra->hitb.end(); ++hitbIt ) {
    ArtePointer<HITB> hitb(*hitbIt);

    // check hit in magnet
    if( !zfpar_.fitmagn )
      {
	if(hitb->z > 206. && hitb->z < 660.)
	  {
	    //cout << " Track.C rtra " << rtra << " hit " << hitb.index()
	    //	 << " z " << hitb->z << " discarded " << endl;

	    continue;
	  }
      }

    // discard muon hits, gede->cmp == 8  
    if(hitb->gede->cmp == 8 )
      {
	//cout << " Track.C hitb " << hitb << " cmp " << hitb->gede->cmp<< endl;
	continue;
      }
    for ( ArteRelationIterator<RHIT> rhitIt = hitb->rhit.begin();
	rhitIt != hitb->rhit.end(); ++rhitIt ) {
      ArtePointer<RHIT> rhit(*rhitIt);
      if (rhit->rtra == rtra) {
	rangerNode* w = rangerNode::makeNode( rhit, fitMode() );
	if (w) {
	  prepend(*w);
	  if (rmap) {
	    rangerHitMap hm( w,rhit );
	    rmap->addHitMap( hm );
	  }
	}
      }
    }
  }
  orderIfNeeded();
}


// Destructor
rangerTrack::~rangerTrack() {
  for (rangerNodeIterator nodeIt = begin();
       nodeIt != end();
       nodeIt++) {
    delete *nodeIt;
  }
}

ostream& operator<<(ostream &os,const rangerTrack &t) {
  os << " #nodes=" << (t.nodes).size() << " chiSquare=" << t.totalChiSq()
     << " qpEst=" << t.qpEstimate
     << " FMode=" << int(t.tFitMode)
     << " ParType=" << int(t.tPartType) << endl;
  for (list<rangerNode*>::const_iterator n = (t.nodes).begin(); 
       n != (t.nodes).end(); n++) {
    os << *n << '\t' << *(*n) << endl;
  }
  return os;
}
  


void rangerTrack::covInit() {
  (*begin())->covInit();
}



int rangerTrack::fit(const int firstPass,const int thirdPass,
		     const int insertMaterial,const int updateInternal) {
  //cout << "In rangerTrack::fit: " << firstPass << endl  ;
  // PASS 1
  if (firstPass) {
    status = filter();
    if (status < 0) {
      return -10;
    }
    status = smooth();
    if (status < 0) {
      cout << "rangerTrack::fit: smoother failed !" << endl  ;
      return -11;
    }
    if (updateInternal) updInternal();
  }

  int nInsert = 0;
  if (insertMaterial) {
    nInsert=insertWalls();
  }

  // PASS 2
  if (nInsert || (! firstPass)) {
    status = filter();
    if (status < 0) {
      //cout << "rangerTrack::fit: filter (pass 2) failed ! " << status <<endl  ;
      static int id4;
      arteMsg << "rangerTrack::fit: filter (pass 2) failed ! " << endl;
      arteMsg.allp(id4,10);
      return -10;
    }
    status = smooth();
    if (status < 0) {
      //cout << "rangerTrack::fit: smoother (pass 2) failed !" << endl  ;
      static int id5; 
      arteMsg << "rangerTrack::fit: smoother (pass 2) failed !" << endl; 
      arteMsg.allp(id5,10); 
      return -11;
    }
    if (updateInternal) updInternal();
    
    // PASS 3
    if (thirdPass) {
      if (fabs((*rbegin())->getState().qP()) > 0.19) {
	status = filter();
	if (status < 0) {
	  return -12;
	}
	status = smooth();
	if (updateInternal) updInternal();
	if (status < 0) {
	  //cout << "rangerTrack::fit: smoother (pass 3) failed !" << endl  ;
	  static int id6; 
	  arteMsg << "rangerTrack::fit: smoother (pass 3) failed !" << endl; 
	  arteMsg.allp(id6,10); 
	  return -11;
	}
      }
    }
  }
  return 0;
}


int rangerTrack::fitFast() {

  status = filterDownStr();
  if (status < 0) {
    return -10;
  }
  status = filter();
  if (status < 0) {
    return -10;
  }
  return 0;
}




int rangerTrack::filter() { 
#ifdef FULL
  int counter=0;
#endif
  float chiSquare ;
  if (! empty()) {
    int stat = (*begin())->filter();
    chiSquare = (*begin())->getChiSq(); // chiSq contribution of first node should be 0 ??
    
#ifdef FULL
    cout << "first filter status stat=" << stat << endl;

    cout << "rangerTrack::filter: after filter step " << ++counter << endl;
    //char cc;
    //cc = cin.get();
#endif

    if (stat >= 0) {
      int last_stat = 0;
      rangerNodeIterator baseNode = begin();
      rangerNodeIterator nextNode = begin();
      nextNode++;
      for (rangerNodeIterator n = nextNode; 
	   n != end() && (last_stat >=0); 
	   n++) {
	last_stat = (*n)->filter(*baseNode,tPartType);
	chiSquare += (*n)->getChiSq() ;
	baseNode = n;
#ifdef FULL  
	if(last_stat < 0) 
	  {
	    cout << "ranger::Track.filter(): last_stat " 
		 << last_stat << " z " << (*n)->zNode()
		 << " hit " << (*n)->getHitb() << endl; 
	  }
#endif 
      }
      return last_stat;
    } 
    else {
      cout << "first filter status stat=" << stat << endl;
      return stat;
    }
  }
  //#ifdef FULL 
  cout << "ranger::Track.filter(): no first node" << endl;
  //#endif
  static int id1;
  arteMsg << "Track.C : ranger::Track.filter(): no first node" << endl;
  arteMsg.allp(id1,1);

  return -10;
}

int rangerTrack::smooth() { 

  if ( empty() ) {
    cout << "rangerTrack::smooth(): empty track" << endl;
    return -10;
  }

  int last_stat = 0;
  for (rangerNodeReverseIterator n = rbegin(); 
       (n != rend()) && (last_stat >= 0); n++) {
    rangerNodeReverseIterator nk = n;
    nk++;
    if (nk != rend()) last_stat = (*nk)->smooth(*n);
  }
  return last_stat;
}

int rangerTrack::filterDownStr() { 

#ifdef FULL
  int counter=0;
#endif

  if (! empty()) {
    int stat = (*rbegin())->filter();

#ifdef FULL
    cout << "first filterDownStr status stat=" << stat << endl;

    cout << "rangerTrack::filterDownStr: after filter step " 
	 << ++counter << endl;
#endif

    if (stat >= 0) {
      int last_stat = 0;
      rangerNodeReverseIterator baseNode = rbegin();
      rangerNodeReverseIterator nextNode = rbegin();
      nextNode++;
      for (rangerNodeReverseIterator n = nextNode; 
	   n != rend() && (last_stat >=0); 
	   n++) {
	last_stat = (*n)->filter(*baseNode,tPartType);
	baseNode = n;
#ifdef FULL
	cout << "filter status last_stat=" << last_stat << endl;
	cout << "rangerTrack::filterDownStr: after filter step " 
	     << ++counter << endl;
	//cc = cin.get();
#endif
      }
      if (last_stat != 0) 
	cout << "rangerTrack::filterDownStr: filtering returns last_stat=" 
	     << last_stat << endl;
      return last_stat;
    } 
    else {
      cout << "rangerTrack::filterDownStr: first filtering returns stat=" 
	   << stat << endl;
      return stat;
    }
  }
#ifdef FULL 
  cout << "ranger::Track.filterDownStr(): no first node" << endl;
#endif
  static int id2; 
  arteMsg << "Track.C : ranger::Track.filterDownStr(): no first node" << endl; 
  arteMsg.allp(id2,1); 
  return -10;
}
      
#include <set>
int rangerTrack::insertWalls(){
  int nInserts = 0;
  if ( empty() ) return 0;
  // test if first node corresponds to highest z
  if ((*begin())->zNode() < (*last())->zNode()) {
    cout << "rangerTrack::insertWalls(): firstNode->zNode()="
	 << (*begin())->zNode() << " last->zNode()="
	 << (*last())->zNode() << endl;
    return 0;
  }
  
  // we keep track of which planes are already inserted.
  set<int> usedplanes ;
  for (rangerNodeIterator nodeIt=begin(); nodeIt != end(); ++nodeIt) 
    if ( (*nodeIt)->plane() != 0 ) usedplanes.insert((*nodeIt)->plane()) ;

  for (rangerNodeIterator nodeIt=begin(); nodeIt != end(); ++nodeIt) {
    if ((*nodeIt)->plane() == 0) continue;
    rangerNodeIterator insertBaseNode = next(nodeIt);
    if (insertBaseNode == end()) break;
    if ((*insertBaseNode)->plane() == 0) continue;
    if (distance(rangerGeom::volN((*nodeIt)->plane()),
		 rangerGeom::volN((*insertBaseNode)->plane())) >= 0) {
      //       cout << "rangerTrack::insertWalls(): distance=" 
      // 	   << distance(rangerGeom::volN((*nodeIt)->plane()),
      // 		       rangerGeom::volN((*insertBaseNode)->plane()))
      // 	   << endl;
      continue;
    }
    
    // Well, there were some problems with linear extrapolation in the
    // magnet. This is a fast solution: we average between the two states.
    float z  = (*nodeIt)->getState().z() ;
    float x  = (*nodeIt)->getState().x() ;
    float y  = (*nodeIt)->getState().y() ;
    float tx = (*nodeIt)->getState().tx() ;
    float ty = (*nodeIt)->getState().ty() ;

    // this is to solve problem of not straight tracks in magnet, where
    // linear extrapolation used in (*volIt)->sLength fails
    float dz = (*insertBaseNode)->getState().z() - z ;
    if(fabs(dz)>0.1) {
      tx = ((*insertBaseNode)->getState().x() - x)/dz;
      ty = ((*insertBaseNode)->getState().y() - y)/dz;
    }

    // locate in-between nodes 
    vector<rangerVolume*>::const_iterator volBeginIt, volEndIt ;
    volBeginIt = rangerGeom::firstUpstreamActiveVolN((*nodeIt)->plane()) ;
    volEndIt   = rangerGeom::firstUpstreamActiveVolN((*insertBaseNode)->plane()) ;

    float sLeng ;
    for (vector<rangerVolume*>::const_iterator volIt=volBeginIt ;
	 volIt > volEndIt ; --volIt) 
      //if (float sLeng = (*volIt)->sLengNode(*nodeIt)) { //<-- wrong in magnet
      if ( (sLeng = (*volIt)->sLength(z,x,y,tx,ty)) > 0 &&  //<-- slightly better
	   usedplanes.find( (*volIt)->plNumber() ) == usedplanes.end() ) {
	usedplanes.insert( (*volIt)->plNumber() ) ;
	
	rangerNode* newNode = 0;
	if ((*volIt)->getGede() != 0)
	  newNode = rangerNode::makeNode((*volIt)->getGede(),
					 *nodeIt,sLeng);
        else if ((*volIt)->getGwal() != 0)
	  newNode = rangerNode::makeNode((*volIt)->getGwal(),
					 *nodeIt,sLeng);
	
	// insert node into track
	if (newNode) {
	  nodeIt = insertBefore(insertBaseNode,newNode);
	  nInserts++;
	}
      }
  }
  return nInserts;
}


int rangerTrack::insertWallsV(){
  //                             special version of insertWalls() for VDS
  int nInserts = 0;
  if ( empty() ) return 0;
  // test if first node corresponds to highest z
  if ((*begin())->zNode() < (*last())->zNode()) {
    cout << "rangerTrack::insertWalls(): firstNode->zNode()="
	 << (*begin())->zNode() << " last->zNode()="
	 << (*last())->zNode() << endl;
    return 0;
  }

  for (rangerNodeIterator nodeIt=begin(); nodeIt != end(); ++nodeIt) {
    if ((*nodeIt)->plane() == 0) continue;
    rangerNodeIterator insertBaseNode = next(nodeIt);
    if (insertBaseNode == end()) break;
    if ((*insertBaseNode)->plane() == 0) continue;
    // check pointers
    if (distance(rangerGeom::volN((*nodeIt)->plane()),
		 rangerGeom::volN((*insertBaseNode)->plane())) >= 0) {
//       cout << "rangerTrack::insertWalls(): distance=" 
// 	   << distance(rangerGeom::volN((*nodeIt)->plane()),
// 		       rangerGeom::volN((*insertBaseNode)->plane()))
// 	   << endl;
      continue;
    }

    // locate in-between nodes 
    vector<rangerVolume*>::const_iterator volBeginIt, volEndIt ;
    volBeginIt = rangerGeom::firstUpstreamActiveVolN((*nodeIt)->plane()) ;
    volEndIt = rangerGeom::firstUpstreamActiveVolN((*insertBaseNode)->plane()) ;

    for (vector<rangerVolume*>::const_iterator volIt=volBeginIt ;
	 volIt != volEndIt ; --volIt) {
      if (float sLeng = (*volIt)->sLengNode(*nodeIt)) {
	rangerNode* newNode = 0;
	if ((*volIt)->getGede() != 0) {
	  newNode = rangerNode::makeNode((*volIt)->getGede(),
					       *nodeIt,sLeng);
	} else if ((*volIt)->getGwal() != 0) {
	  newNode = rangerNode::makeNode((*volIt)->getGwal(),
					 *nodeIt,sLeng);
	}
     	// insert node into track
	if (newNode) {
	  nodeIt = insertBefore(insertBaseNode,newNode);
	  nInserts++;
	}
      }
    }
  }
  return nInserts;
}



void rangerTrack::updInternal(){
  for (rangerNodeIterator nodeIt = begin();
       nodeIt != end();
       nodeIt++) {
    (*nodeIt)->updInternal();
  }
}



void rangerTrack::append(rangerNode& node) {
  nodes.push_back( &node );
}

void rangerTrack::prepend(rangerNode& node) {
  nodes.push_front( &node );
}


list<rangerNode*>::iterator rangerTrack::insertAfter
                                 (rangerNodeIterator baseNode, 
                                            rangerNode* newNode) { 
  if (baseNode != end()) {
    rangerNodeIterator insertPosition = baseNode;
    insertPosition++;
    rangerNodeIterator tmp = nodes.insert( insertPosition,newNode);
    return tmp;
  }
  else {
    cout << "rangerTrack::insertAfter: invalid base node " << endl;
    return baseNode;
  }
} 


list<rangerNode*>::iterator rangerTrack::insertBefore
(rangerNodeIterator baseNode,rangerNode* newNode) { 
  return nodes.insert(baseNode,newNode);

} 



void rangerTrack::trackList() {
  cout << "tFitMode=" << int(tFitMode) << endl;
  for (rangerNodeIterator n = begin(); n != end(); n++) {
    (*n)->list();
  }
}


void rangerTrack::dump() {
  cout << "D u m p           tFitMode=" << int(tFitMode) << endl;
  for (rangerNodeIterator n = begin(); n != end(); n++) {
    (*n)->dump();
  }
}


void rangerTrack::storeback(rangerRMap* rmap){

  if (! rmap) return;
  if (empty()) return;

  // Pass 0: prepare RPNTs for particle id
  rmap->prepareIdPnt();

  // for RTRA: prepare RPNT for high z point
  if ((rmap->getRtra()) != 0) {
    rmap->addPnt(*begin());
  }

  // Pass 1: store the RHITs
  for (rangerNodeIterator n = begin(); n != end(); n++) {
    if ((*n)->hasHitb()) {
      ArtePointer<RHIT> rhit = rmap->getRhit(*n);
      if (rhit != 0) (*n)->storeRhit(rhit);
    }
#ifdef SAVEWALLS
    else {              // store wall RPNTs (for diagnosis only)
      if (rmap->getRpnt(*n) == 0) {
	ArtePointer<RPNT> rp = (*n)->storeRpnt(0);
	if (rp != 0) {  // new RPNT, set relation
	  if ((rmap->getRseg()) != 0) {
	    createRelation( rmap->getRseg(), rp, &RSEG::rpnt );
	  }
	  else if ((rmap->getRtra()) != 0) {
	    createRelation( rmap->getRtra(), rp, &RTRA::rpnt );
	  }
	}
      }
    }
#endif
  }

  // Pass 2A: store the RSEG
  if ((rmap->getRseg()) != 0) {
    rangerNode* nLow = rmap->getNodeZLow();
    if (nLow) nLow->storeRseg(rmap->getRseg(),1);
    rangerNode* nHigh = rmap->getNodeZHigh();
    if (nHigh) nHigh->storeRseg(rmap->getRseg(),2);

    (rmap->getRseg())->chi = totalChiSq() ;
    (rmap->getRseg())->fit = (status<0)? -10: 0;

    if ((rmap->getRseg())->cmp == 0) (rmap->getRseg())->cmp = genCmp();
  }
  // Pass 2B: store the RTRA
  else if ((rmap->getRtra()) != 0) {
    rangerNode* nLow = rmap->getNodeZLow();
    if (nLow) nLow->storeRtra(rmap->getRtra(),1);

    (rmap->getRtra())->fit  = (status<0)? -10: 0;
    (rmap->getRtra())->chi  = totalChiSq() ;
    (rmap->getRtra())->ndof = totalNdof() ;
  }

  // pass 3: store the RPNT
  for ( list< rangerHitMap >::const_iterator 
	  HMIt=(rmap->getHMList()).begin();
	HMIt !=(rmap->getHMList()).end(); 
	HMIt++){
    if ((*HMIt).getRpnt() != 0) {   // store into assigned RPNT
      (*HMIt).getNode()->storeRpnt((*HMIt).getRpnt());

      if ((rmap->getRseg()) != 0) {
	int related = 0;
	related = (((*HMIt).getRpnt())->rseg == rmap->getRseg());

	if (! related) {  // set relation
	  createRelation( rmap->getRseg(), (*HMIt).getRpnt(), &RSEG::rpnt );
	}
      }
      if ((rmap->getRtra()) != 0) {
	if( find( (*HMIt).getRpnt()->rtra.begin(),
		  (*HMIt).getRpnt()->rtra.end(),
		  rmap->getRtra() ) == (*HMIt).getRpnt()->rtra.end() ) {
	  createRelation( rmap->getRtra(), (*HMIt).getRpnt(), &RTRA::rpnt );
	}
      }
    }
  }
};


void rangerTrack::storeback(vector<rangerRMap>* rmapVec) {
  if (! rmapVec) return;
  for (vector<rangerRMap>::iterator rIt=rmapVec->begin();
       rIt != rmapVec->end(); rIt++) {
    storeback( &(*rIt) );
  }
}


ArtePointer<RTRA> rangerTrack::storeRtra() const{

  if (empty()) return 0;

  // Pass 1: allocate and store RTRA entry
  mmbook_(ArtePointer<RTRA>::getLink().getRef(),1);
  ArtePointer<RTRA> rt( ArtePointer<RTRA>::nrow() );
  (*rbegin())->storeRtra(rt,1);   // store point of lowest z

  rt->fit  = (status<0)? -10: 0;
  rt->chi  = totalChiSq(); 
  rt->ndof = totalNdof() ;
  rt->hsi  = 0;
  rt->hin  = 0;
  rt->hou  = 0;
  rt->hri  = 0;
  rt->hca  = 0;
  rt->htr  = 0;
  rt->hmu  = 0;

  rangerRMap rmap(rt);  // create hit map to aid further operations

  // Pass 2: store the RHITs
  for (rangerNodeConstIterator n = begin(); n != end(); n++) {
    if ((*n)->hasHitb()) {
      mmbook_(ArtePointer<RHIT>::getLink().getRef(),1);
      ArtePointer<RHIT> rhit( ArtePointer<RHIT>::nrow() );
      if (rhit != 0) {
	createRelation( rt, (*n)->getHitb(), &RTRA::hitb );
	createRelation( rhit, (*n)->getHitb(), &HITB::rhit );
	(*n)->storeRhit(rhit);
	rhit->rtra = rt.index();
	rangerHitMap hm((*n),rhit);
	rmap.addHitMap( hm );
      }
      ArtePointer<GEDE> ge(((*n)->getHitb())->gede);  // count hits
      switch (ge->cmp) {
      case 1:
	rt->hsi++;
	break;
      case 2:
	rt->hou++;
	break;
      case 3:
	rt->hin++;
	break;
      case 4:
	rt->hri++;
	break;
      case 5:
	rt->hca++;
	break;
      case 6:
	rt->htr++;
	break;
      case 7:
	rt->hmu++;
	break;
      default:
	cout << " rangerTrack::storeRtra: bad cmp=" << ge->cmp << endl;
      }
    }
#ifdef SAVEWALLS
    else {              // store wall RPNTs (for diagnosis only)
      ArtePointer<RPNT> rp = (*n)->storeRpnt(0);
	if (rp != 0) {  // new RPNT, set relation
	  createRelation( rt, rp, &RTRA::rpnt );
	}
    }
#endif
  }
  // Pass 3: prepare RPNTs for particle id
  rmap.prepareIdPnt();

  // prepare RPNT for high z point
  rmap.addPnt(*begin());

  // pass 4: store the RPNT
  for ( list< rangerHitMap >::const_iterator 
	  HMIt=(rmap.getHMList()).begin();
	HMIt !=(rmap.getHMList()).end(); 
	HMIt++){
    if ((*HMIt).getRpnt() != 0) {   // store into assigned RPNT
      (*HMIt).getNode()->storeRpnt((*HMIt).getRpnt());

      // set relation
      createRelation( rt, (*HMIt).getRpnt(), &RTRA::rpnt );
    }
  }
  return rt;
}


ArtePointer<RSEG> rangerTrack::storeRseg() const{

  if (empty()) return 0;

  // Pass 1: allocate and store RSEG entry
  mmbook_(ArtePointer<RSEG>::getLink().getRef(),1);
  ArtePointer<RSEG> rs( ArtePointer<RSEG>::nrow() );
  (*rbegin())->storeRseg(rs,1);   // store point of lowest z
  (*begin())->storeRseg(rs,2);    // store point of high z

  rs->fit = (status<0)? -10: 0;
  rs->chi = totalChiSq();
  rs->cmp = genCmp();

  rangerRMap rmap(rs);  // create hit map to aid further operations

  // Pass 2: store the RHITs
  for (rangerNodeConstIterator n = begin(); n != end(); n++) {
    if ((*n)->hasHitb()) {
      mmbook_(ArtePointer<RHIT>::getLink().getRef(),1);
      ArtePointer<RHIT> rhit( ArtePointer<RHIT>::nrow() );
      if (rhit != 0) {
	createRelation( rs,   (*n)->getHitb(), &RSEG::hitb );
	createRelation( rhit, (*n)->getHitb(), &HITB::rhit );
	createRelation( rs,   rhit,            &RSEG::rhit );
	(*n)->storeRhit(rhit);
	rhit->rtra = 0;
	rangerHitMap hm((*n),rhit);
	rmap.addHitMap( hm );
      }
    }
#ifdef SAVEWALLS
    else {              // store wall RPNTs (for diagnosis only)
      ArtePointer<RPNT> rp = (*n)->storeRpnt(0);
	if (rp != 0) {  // new RPNT, set relation
	  createRelation( rs, rp, &RSEG::rpnt );
	}
    }
#endif
  }

  // Pass 3: prepare RPNTs for particle id
  rmap.prepareIdPnt();

  // pass 4: store the RPNT
  for ( list< rangerHitMap >::const_iterator 
	  HMIt=(rmap.getHMList()).begin();
	HMIt !=(rmap.getHMList()).end(); 
	HMIt++){
    if ((*HMIt).getRpnt() != 0) {   // store into assigned RPNT
      (*HMIt).getNode()->storeRpnt((*HMIt).getRpnt());

      // set relation
      createRelation( rs, (*HMIt).getRpnt(), &RSEG::rpnt );
    }
  }
  return rs;
}


double rangerTrack::getLowZ() const {
  if (empty()) return 0;
  else {
    return (*rbegin())->zNode();
  }
}

const HepVector* rangerTrack::getLowZVec() const {
  if (empty()) return 0;
  else {
    return &(((*rbegin())->getState()).vec());
  }
}

const HepSymMatrix* rangerTrack::getLowZCov() const {
  if (empty()) return 0;
  else {
    return &(((*rbegin())->getState()).cov());
  }
}

float rangerTrack::totalChiSq() const 
{
  // returns the total chi-square of the fit
  double chiSquare = 0;
  for(rangerNodeConstIterator n = begin(); n != end(); ++n) 
    chiSquare += (*n)->getChiSq() ;
  return float(chiSquare) ;
}

int rangerTrack::totalNdof() const
{
  // returns the number of degrees of freedom in the fit
  int ndof = - (*begin())->getState().dim() ;
  for(rangerNodeConstIterator n = begin(); n != end(); ++n)
    ndof += (*n)->dim() ;
  return ndof ;
}

extern "C" float prob_(const float&, const int&) ;
float rangerTrack::chiSqProb() const
{
  int ndof    = totalNdof() ;
  float chiSq = totalChiSq() ;
  return ndof>0 ? prob_(chiSq,ndof) : -1 ;
}

double rangerTrack::totalRadThickness() const {
  
  double sum = 0;
  for (rangerNodeConstIterator n = begin(); n != end() ; n++) {
    sum += (*n)->radiationThickness();
  }
  return sum;
}

int rangerTrack::genCmp() const {
  int CMP = 0;
  for (rangerNodeConstIterator n = begin(); n != end() ; n++) {
    if ((*n)->hasHitb()) {
      if ((*n)->plane()>0) {
	ArtePointer<GEDE> ge((*n)->plane());
	float zsl = 0.5 * (((ArtePointer<GESL>) ge->gesl)->z1 
			   +((ArtePointer<GESL>) ge->gesl)->z2);

	if (               zsl <  190) CMP = CMP | Rsegc::vxd;  // set functional bits
	if (zsl >=  190 && zsl <  660) CMP = CMP | Rsegc::magt;
	if (zsl >=  660 && zsl < 1000) CMP = CMP | Rsegc::patt;
	if (zsl >= 1000 && zsl < 1330) CMP = CMP | Rsegc::trit;
	if (zsl >= 1330 && zsl < 1400) CMP = CMP | Rsegc::ecal;
	if (zsl >= 1400              ) CMP = CMP | Rsegc::muon;

	switch (ge->cmp) {
	case 1:
	  CMP = CMP | Rsegc::sil;
	  break;
	case 2:
	  CMP = CMP | Rsegc::wch;
	  break;
	case 3:
	  CMP = CMP | Rsegc::msg;
	  break;
	case 4:
	  CMP = CMP | Rsegc::ptc;
	  break;
	case 5:
	  CMP = CMP | Rsegc::rich;
	  break;
	case 6:
	  CMP = CMP | Rsegc::trd;
	  break;
	case 7:
	  CMP = CMP | Rsegc::bit23;
	  break;
	case 8:
	  CMP = CMP | Rsegc::bit22;
	  break;
	default:
	  break;
	}
      }
    }
  }
  return CMP;
}


void rangerTrack::lastNoise() {
  (*last())->processNoise(tPartType);
}

void rangerTrack::zOrder() {
  nodes.sort( not2( compareCall( &rangerNode::zNode ) ) );
}

void rangerTrack::orderIfNeeded() {
  float zHere = 999999;
  int nOrderFault = 0;
  float zFault = 0;
  for (rangerNodeIterator nodeIt = begin();
       nodeIt != end();
       nodeIt++) {
    if ( (*nodeIt)->zNode() > zHere || (*nodeIt)->zNode() == zFault ) {
      ++nOrderFault;
      zFault = (*nodeIt)->zNode();
    }
    zHere = (*nodeIt)->zNode();
  }

  if (nOrderFault > 0) {
    static int id;
#ifdef PRINT_ORDER_FAULTS
    cout << "rangerTrack::orderIfNeeded(): number of order faults on input "
	 << nOrderFault << " last at z=" << zFault << endl;
#endif
    arteMsg << "rangerTrack::orderIfNeed(): track required sort,"
	    << " zb=" << (*begin())->zNode() 
	    << " ze=" << (*last())->zNode() << endl;
    arteMsg.allp(id,10);
    zOrder();
  }
}

int rangerTrack::fitV() {

  int nInsert = 0;
  nInsert=insertWalls();
  status = filter();
  if (status < 0) { 
    return -10; 
  } 
  return 0; 
}

int rangerTrack::fitVf() {

  status = filter();
  if (status < 0) {  
    return -10;  
  }  
  return 0;  
}


int rangerTrack::fitVs() {

  int nInsert = 0;
  nInsert=insertWallsV();
  status = filter();
  if (status < 0) {  
    return -10;  
  }  
  return 0;  
}

int rangerTrack::fitGlobal()
{
  const int maxnstep    = 5 ;
  const float dchi2conv = 0.1 ; 
  int stat = -1 ;
  if (! empty()) {
    rangerNode* baseNode = (*begin());
    stat = baseNode->filter();
    if (stat >= 0) {
      rangerState* globstate = &(baseNode->getState()) ;
      int dim = globstate->dim() ;
      int ndof = -dim;
      float chiSquare=0 ;
      for (rangerNodeIterator n = begin() ; n != end() && stat>=0 ; n++) {
	stat = (*n)->projectGlobal(globstate) ;
	ndof += (*n)->dim() ;
	chiSquare += (*n)->getChiSq() ;
      }
      if(stat>=0) {
	// check ndof
	if(ndof<=0) stat = -2;
	else {
	  HepVector hdChi2dQ(dim) ;
	  HepSymMatrix hd2Chi2dQ2(dim) ;
	  bool converged=false ;
	  for(int nstep=0; nstep<maxnstep && !converged && stat>=0; nstep++) {
	    //cout << "Step: " << nstep << endl ;
	    // get chi2 contribution
	    for(int row=1; row<=dim; row++) {
	      hdChi2dQ(row) = 0;
	      for(int col=1; col<=row; col++) 
		hd2Chi2dQ2(row,col) = 0;
	    }	  

	    for (rangerNodeIterator n = begin() ; n != end() && stat>=0 ; n++)
	      stat = (*n)->absorbGlobal(&hdChi2dQ,&hd2Chi2dQ2) ;

	    if(stat >=0) {
	      // invert covariance matrix
	      int ierr ;
	      hd2Chi2dQ2.invert(ierr) ;
	      if(ierr) stat = -3 ; 
	      else {
		// calc change in chi2
		float dchi2 = hd2Chi2dQ2.similarity(hdChi2dQ) ;
		if(dchi2 < dchi2conv/ndof) converged = true ;
		//cout << "dchi2: " << dchi2 << endl ;
		// update state
		globstate->cov() = hd2Chi2dQ2;
		globstate->vec() -= hd2Chi2dQ2*hdChi2dQ ;
		
		chiSquare = 0;
		for (rangerNodeIterator n = begin() ; n != end() ; n++) {
		  (*n)->projectGlobal(globstate) ;
		  chiSquare += (*n)->getChiSq() ;
		}
	      }
	    }
	  }
	}
      }
    }
  }
  return stat ;
}
