// 
//  Author:  Rainer Mankel                     Created:  29-Sep-98
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
// 
#include <iostream.h>
#include <math.h>
#include "rang/VolTube.hh"
#include "rang/FLAGRotFix.hh"

// Before application of the rotation matrix, the tube/s axis is assumed
// to be identical with the z axis
// 
// The following information from GEDE/GWAL is used:
//
//   gwal->xmin   inner radius of tube
//       ->xmax   outer radius of tube
//       ->ymin   x offset of CM
//       ->ymax   y offset of CM
//       ->zmin   lower z bound
//       ->zmax   upper z bound
//       ->rot    rotation matrix
//
// See also gean/crvtrm.F


//-- Constructor
rangerVolTube::rangerVolTube(){};

//-- Constructor
rangerVolTube::rangerVolTube( rangerGeWa* gewa ) : rangerVolume( gewa ) {
};

//-- Destructor
rangerVolTube::~rangerVolTube(){};


double rangerVolTube::sLength(double zz,   // give traversed length
			      double x,    // in units of
			      double y,    // zmax - zmin
			      double tx,
			      double ty){
  // rotation into PCS
  // see Hera-B 97-090 p 67

#ifdef ROTFIX
  double xP = Gewa->rot(1) * x + Gewa->rot(4) * y + Gewa->rot(7) * (zz-1080);
  double yP = Gewa->rot(2) * x + Gewa->rot(5) * y + Gewa->rot(8) * (zz-1080);
  double zP = Gewa->rot(3) * x + Gewa->rot(6) * y + Gewa->rot(9) * (zz-1080)
    + 1080;
#else
  double xP = Gewa->rot(1) * x + Gewa->rot(4) * y + Gewa->rot(7) * zz;
  double yP = Gewa->rot(2) * x + Gewa->rot(5) * y + Gewa->rot(8) * zz;
  double zP = Gewa->rot(3) * x + Gewa->rot(6) * y + Gewa->rot(9) * zz;
#endif

  double tzn = Gewa->rot(3) * tx + Gewa->rot(6) * ty + Gewa->rot(9);
  if (tzn == 0) {
    cout << "rangerVolTube::sLength: tzn=" << tzn << endl;
    tzn = .00001;
  }
  double txP = (Gewa->rot(1) * tx + Gewa->rot(4) * ty + Gewa->rot(7)) / tzn;
  double tyP = (Gewa->rot(2) * tx + Gewa->rot(5) * ty + Gewa->rot(8)) / tzn;

  // z of tube center
  double zC = 0.5 * (Gewa->zmin() + Gewa->zmax());

  // the following coordinates are in the normalized system of the tube
  // the center of the tube is at (0,0,0)
  double xL = xP + (zC - zP) * txP - Gewa->ymin();  
  double yL = yP + (zC - zP) * tyP - Gewa->ymax();
  double rLsq = xL * xL + yL * yL;

  // use mean radius to calculate intersection
  double rMean = 0.5 * (Gewa->xmin() + Gewa->xmax());

  // coefficients in quadratic equation
  double tTot2 = txP * txP + tyP * tyP;
  double PPP = 2 * (xL * txP + yL * tyP) / tTot2;
  double QQQ = (rLsq - rMean * rMean) / tTot2;
  double discr = 0.25 * PPP * PPP - QQQ;

  double length = 0;  // initialize traversed length

  // check for passage through cylinder wall
  if (discr > 0) {
    for (int sig=-1; sig<=1; sig += 2) {    // loop over solutions
      double zInterSect = -0.5 * PPP + sig * sqrt(discr);
      if (fabs(zInterSect) < 0.5*fabs(Gewa->zmax() - Gewa->zmin())) {
	double xIntersect = xL + zInterSect * txP;
	double yIntersect = yL + zInterSect * tyP;

	double exN = xIntersect / rMean;    // unit vector normal to surface
	double eyN = yIntersect / rMean;    // ezN = 0

	double pro = txP * exN + tyP * eyN; // projection normal to surface

	if (pro != 0) {
	  length += fabs(Gewa->xmax() - Gewa->xmin()) 
	    * sqrt(1 + txP*txP + tyP*tyP) / fabs(pro);
	}
	else {
	  length += fabs(Gewa->zmax() - Gewa->zmin());
	}
      }
    }
  }

  // check for passage through ring surface
  if ( (rLsq > Gewa->xmin()*Gewa->xmin()) &&
       (rLsq < Gewa->xmax()*Gewa->xmax()) ) {
    length += fabs(Gewa->zmax() - Gewa->zmin()) * sqrt(1 + txP*txP + tyP*tyP);
  }

  double scaledLength = length / fabs(Gewa->zmax() - Gewa->zmin());

  return scaledLength;

}


const float rangerVolTube::zMean(){
#ifdef ROTFIX
  // This is a fix for a HBGEAN bug in ARTE-01-08-r1, where the rotation
  // was made around the GEANT origin, not the ARTE origin 
  float zM =  Gewa->rot(7) * Gewa->ymin() + Gewa->rot(8) * Gewa->ymax() 
    + Gewa->rot(9) * (0.5 * (Gewa->zmin()+Gewa->zmax()) - 1080) +1080;
#else
  // This is the standard version for the case that the bug is fixed
  float zM =  Gewa->rot(7) * Gewa->ymin() + Gewa->rot(8) * Gewa->ymax() 
    + Gewa->rot(9) * 0.5 * (Gewa->zmin()+Gewa->zmax());
#endif
//   cout << "rangerVolTube::zMean(): return zMean=" << zM
//        << " for Gewa=" << Gewa->plNumber() << endl;
  return zM;
}
