// *****************************************************************************
//
// source:      Cloneremove.C
//
// version:     V1.0
//
// created by:  Marc-Andre Pleier (02-05-01)
//
// purpose:     Select a clone-free subset out of the RTRA table  
//
// changes:
//
// Remarks: This code selects tracks out of the RTRA table ensuring no track-
//          segment of VDS/MAIN/RICH/ECAL is used twice. 
//          There are NO quality cuts for those tracks, so that following 
//          analysis should not be biased by this code.  
//          The only "cuts" this code is using are:
//             a) demanding a fit-flag == 0 (reconstructed charged track)
//             b) demanding that a VDS-segment exists.
//
//          Afterwards, for every track a FoM is calculated taking into account
//          the number of hits in VDS, Maintracker and RICH, the existence of ECAL
//          information for this track and the chi2 of the track.
//          All tracks are filled into a map sorted with respect to their FoM.
//          Finally the tracks are accepted by the code starting from the best 
//          ones skipping any track whose track-segments have been used by a 
//          track already accepted before.
//
//          The tracks accepted by this code thus can be any combination of a 
//          VDS-segment with MAIN tracker/ECAL/RICH or VDS-only tracks.          
//          For all accepted tracks, the bit Rsegc::bit20 is set and the number 
//          of accepted tracks is returned.
//          SEE THE CORRESPONDING HERA-B NOTE (01-062) FOR MORE DETAILS.
// *****************************************************************************

#include <iostream.h>
#include <vector.h>
#include <map.h>

// ARTE include
#include "arte/arte.hh"
#include "arte/RSEG_cmp.hh"

int Cloneremove()
{ 
  const ArteTable<RTRA>::iterator rtend = ArteTable<RTRA>::end(); 

  // map for the sorting of tracks
  map< float, ArtePointer<RTRA> > mrtra;
  // vector for the accepted tracks
  vector< ArtePointer<RTRA> > rtra; 
  
  // put all VDS(/OTR/RICH/ECAL) tracks into a sorted map
  for(ArteTable<RTRA>::iterator itrtra = ArteTable<RTRA>::begin();
      itrtra != rtend; ++itrtra) 
    {
      const ArtePointer<RTRA> rtrap = (*itrtra);

      // First unselect all RTRAs
      if( rtrap->cmp & Rsegc::bit20 ) rtrap->cmp ^= Rsegc::bit20;

      // we want reconstructed charged tracks only...
      if( rtrap->fit !=  0 ) continue; 
                                      
      //  ... and a VDS-segment should exist.
      const int nvds  = rtrap->hsi;
      if( nvds  < 5 ) continue; 

      // check for the MAINtracker...
      int nmain = rtrap->hou + rtrap->hin;
      if ( nmain <= 0 ) nmain = 0; // this looks paranoiac, I know...

      // check for the RICH...
      int nrich = rtrap->hri;
      float richh = 0.;
      if( nrich > 0 ) 
	{
	  richh = nrich /4.;
	  // for the ndof
	  nrich = 1;
	}
      else
	nrich = 0;

      // check for the ECAL...
      const int necal = ( (rtrap->hca > 0) ? 1 : 0 );

      // to avoid division by 0 for VDS-only tracks...
      const int vdsonly = ( ( !nmain && !nrich && !necal ) ? 1 : 0 ); 

      // calculate FoM...
      const float hitsum = nvds + nmain + necal + richh;
      const float ndof = vdsonly + nvds + nmain + (2*nrich) + (2*necal) - 5.;
      const float fom = hitsum - ( ( rtrap->chi ) / ndof );
      mrtra.insert(pair< float, ArtePointer<RTRA> >(-fom,rtrap));
    }
  
  // copy the accepted tracks into the rtra vector, omitting all clones
  const map< float, ArtePointer<RTRA> >::const_iterator end=mrtra.end();
  for(map< float, ArtePointer<RTRA> >::iterator it=mrtra.begin();
      it != end; ++it) 
    {
      int OK = 1;
      const ArtePointer<RTRA> artra = it->second;
      const unsigned int size = rtra.size();
      for(unsigned int i=0; i!=size; ++i) 
	{
	  const ArtePointer<RTRA> grtra = rtra[i];
	  if(grtra->csvd  == artra->csvd ) { OK=0; break; }
	  if( ((artra->hou + artra->hin) > 0 ) && (grtra->cmain == artra->cmain)) { OK=0; break; }
	  if( ((artra->hri > 0) && (artra->crich != -10 )) && (grtra->crich == artra->crich)) { OK=0; break; }
	  if( (artra->hca > 0) && (grtra->cecal == artra->cecal)) { OK=0; break; }
	}
      
      if(OK)  
	{
	  rtra.push_back(artra);
	  artra->cmp |= Rsegc::bit20;
	}
    }
  
  return rtra.size();
}                              




