////////////////////////////////////////////////////////////
// 
// Save RICH likelihood information from RTRA to RTRI,
// copy saved information back to RTRI:
//     algorithm 0 = RISE
//     algorithm 1 = RITER/RIRE
//
// copyRTRAtoRTRI(int algorithm)
// copyRTRItoRTRA(int algorithm)
//
//  returns 0 if OK
//
//  Created: 15 July, 2000 D.Dujmic
//
//  modified:
//  06 Aug 2000  (DD) maximizeRTRA - chooses higher likelihood 
//                    of two two methods
//  06 Aug 2000  (DD) favorRTRA - choses one method, if not filled
//                    takes the other method
//  28 Nov 2000  (Rok Pestotnik) copy the lrme field to/from RTRI table for RITER/RIRE
//  28 Nov 2000  (Rok Pestotnik) reset values in RTRA to -1 after being copied to RTRI
//  12 Dec 2000  (AG) rtri->lsme=int( rtra->lrme ) (int lsme; float lrme)
//  10 Jan 2001  (AG) re-changed lsme
//  11 Jan 2001  (DD) rcResetRichRtra: set RICH likelihoods to -1
//                    remove -1 from copyRTRAtoRTRI
//  13 Apr 2001  (RP,MS) in fields rtri->flt, rtri->fls both, hri and flr are stored:
//                    bits 0:15 = hri, bits 16:30 = flr, if bit 31 = 1
//                    when copiing back, all fields are restored (if bit 31 = 1)
////////////////////////////////////////////////////////////

#include "arte/arte.hh"
#define FLR_NEW 0x80000000

unsigned int nrtra;

int copyRTRAtoRTRI(int algorithm) {

  // -- setup RTRI for a new event, create a parallel table to RTRA
  static int old_event=0;
  if (old_event!=ArtePointer<EVHD>(1)->evt) {
    ArteTable<RTRI>::clear();
    nrtra=ArteTable<RTRA>::size();
    RTRI rtri;
    memset(&rtri,0,sizeof(RTRI));
    for (unsigned int i=1; i<=nrtra; i++) ArteTable<RTRI>::push_back(rtri);
    old_event=ArtePointer<EVHD>(1)->evt;
  }
  if (!nrtra) return 0;
  if (nrtra!=ArteTable<RTRI>::size()) { 
    cout<<"copyRTRAtoRTRI: nrtra!=ArteTable<RTRI>::size() !!"<<endl;
    return -1;
  }

  // -- copy RTRA for an algorithm
  // -- RISE=0, RITER/RIRE=1
  for (unsigned int i=1; i<=nrtra; i++) {
    ArtePointer<RTRI> rtri(i);
    ArtePointer<RTRA> rtra(i);
    switch (algorithm) {
    case 0:
      rtri->lte =rtra->lre;
      rtri->ltmu=rtra->lrmu;
      rtri->ltpi=rtra->lrpi;
      rtri->ltk =rtra->lrk;
      rtri->ltp =rtra->lrp;
      rtri->flt= FLR_NEW | ( ( rtra->flr & 0x7FFF) << 16) | ( rtra->hri & 0xFFFF);  
      break;
    case 1:
      rtri->lse =rtra->lre;
      rtri->lsmu=rtra->lrmu;
      rtri->lspi=rtra->lrpi;
      rtri->lsk =rtra->lrk;
      rtri->lsp =rtra->lrp;
      rtri->lsme=rtra->lrme;
      rtri->fls= FLR_NEW | ( ( rtra->flr & 0x7FFF) << 16) | ( rtra->hri & 0xFFFF);  
      break;
    default :
      cout<<"copyRTRAtoRTRI: wrong algorithm specified: "<<algorithm
          <<"  must be 0 or 1"<<endl;
      return -1;
      break;
    }
  }

  return 0;
}

int copyRTRItoRTRA(int algorithm) {

  nrtra=ArteTable<RTRI>::size();
  if (!nrtra) return 0;

  // -- copy RTRI to RTRA for an algorithm
  // -- RISE=0, RITER/RIRE=1
  for (unsigned int i=1; i<=nrtra; i++) {
    ArtePointer<RTRI> rtri(i);
    ArtePointer<RTRA> rtra(i);
    switch (algorithm) {
    case 0:
      rtra->lre =rtri->lte;
      rtra->lrmu=rtri->ltmu;
      rtra->lrpi=rtri->ltpi;
      rtra->lrk =rtri->ltk;
      rtra->lrp =rtri->ltp;
      rtra->hri =( rtri->flt & FLR_NEW ) ? ( rtri->flt & 0xFFFF ) : rtri->flt ;
      if ( rtri->flt & FLR_NEW ) { rtra->flr = ( rtri->flt) >> 16 & 0x7FFF; }
      break;
    case 1:
      rtra->lre =rtri->lse;
      rtra->lrmu=rtri->lsmu;
      rtra->lrpi=rtri->lspi;
      rtra->lrk =rtri->lsk;
      rtra->lrp =rtri->lsp;
      rtra->lrme=rtri->lsme;
      rtra->hri =( rtri->fls & FLR_NEW ) ? ( rtri->fls & 0xFFFF ) : rtri->fls ;
      if ( rtri->fls & FLR_NEW ) { rtra->flr = ( rtri->fls) >> 16 & 0x7FFF; }
      break;
    default :
      cout<<"copyRTRItoRTRA: wrong algorithm specified: "<<algorithm
	  <<"  must be 0 or 1"<<endl;
      return -1;
      break;
    }
  }

  return 0;
}



int maximizeRTRA() {
  nrtra=ArteTable<RTRI>::size();
  if (!nrtra) return 0;
  // -- take higher likelihood of two algorithms
  for (unsigned int i=1; i<=nrtra; i++) {
    ArtePointer<RTRI> rtri(i);
    ArtePointer<RTRA> rtra(i);
    rtra->lre  = (rtri->lte >rtri->lse ) ? rtri->lte  : rtri->lse;
    rtra->lrmu = (rtri->ltmu>rtri->lsmu) ? rtri->ltmu : rtri->lsmu;
    rtra->lrpi = (rtri->ltpi>rtri->lspi) ? rtri->ltpi : rtri->lspi;
    rtra->lrk  = (rtri->ltk >rtri->lsk ) ? rtri->ltk  : rtri->lsk;
    rtra->lrp  = (rtri->ltp >rtri->lsp ) ? rtri->ltp  : rtri->lsp;
    int flt  = ( rtri->flt & FLR_NEW ) ? ( rtri->flt & 0xFFFF ) : rtri->flt ;
    int fls  = ( rtri->fls & FLR_NEW ) ? ( rtri->fls & 0xFFFF ) : rtri->fls ; 
    rtra->hri  = ( flt > fls ) ? flt  : fls;
    rtra->flr = 0x2000;
  }
  // should I normalize?
  return 0;
}



int favorRTRA(int mode) {
  nrtra=ArteTable<RTRI>::size();
  if (!nrtra) return 0;
  // -- favor one of algorithms, if not filled use the other one
  // -- RISE=0, RITER/RIRE=1
  for (unsigned int i=1; i<=nrtra; i++) {
    ArtePointer<RTRI> rtri(i);
    ArtePointer<RTRA> rtra(i);    
    int flt  = ( rtri->flt & FLR_NEW ) ? ( rtri->flt & 0xFFFF ) : rtri->flt ;
    int fls  = ( rtri->fls & FLR_NEW ) ? ( rtri->fls & 0xFFFF ) : rtri->fls ;
    switch (mode) {
    case 0:
      rtra->lre  = (rtri->lte >0) ? rtri->lte  : rtri->lse;
      rtra->lrmu = (rtri->ltmu>0) ? rtri->ltmu : rtri->lsmu;
      rtra->lrpi = (rtri->ltpi>0) ? rtri->ltpi : rtri->lspi;
      rtra->lrk  = (rtri->ltk >0) ? rtri->ltk  : rtri->lsk;
      rtra->lrp  = (rtri->ltp >0) ? rtri->ltp  : rtri->lsp;
      rtra->hri  = (flt >0) ? flt  : fls;
      rtra->flr = 0x4000;
      break;
    case 1:
      rtra->lre  = (rtri->lse >0) ? rtri->lse  : rtri->lte;
      rtra->lrmu = (rtri->lsmu>0) ? rtri->lsmu : rtri->ltmu;
      rtra->lrpi = (rtri->lspi>0) ? rtri->lspi : rtri->ltpi;
      rtra->lrk  = (rtri->lsk >0) ? rtri->lsk  : rtri->ltk;
      rtra->lrp  = (rtri->lsp >0) ? rtri->lsp  : rtri->ltp;
      rtra->hri  = (fls >0) ? fls  : flt;      
      rtra->flr = 0x4000;
      break;
    defaut:
      cout<<"favorRTRA: wrong algorithm specified: "<<mode
	  <<"  must be 0 or 1"<<endl;
      return -1;
      break;
    }
  }
  // should I normalize?
  return 0;
}


void rcResetRichRtra(float x=-1) {
  for (ArteTable<RTRA>::iterator irtra=ArteTable<RTRA>::begin();
       irtra!=ArteTable<RTRA>::end(); ++irtra) {
    ArtePointer<RTRA> rtra=(*irtra);
    rtra->lre = rtra->lrmu = rtra->lrpi = rtra->lrk = rtra->lrp = rtra->lrme = x;
    rtra->hri = int(x);
    rtra->flr = 0;
  }
}





