// ********************************************************************
//
//  Routine:      rcSelectJPSIEE
//  Author:       T.Oest
//
//  Description:  search for J/psi candidates
//
// symbol:    rcSelectJPSIEE
//
// called by: rcSelectDecays
//
// purpose:   J/Psi -> e+ e- AND detached vertex
//
// hints:     selection results are stored in appropriate tables (RTRA, RVER)
//            cuts might be set in classification if not defaults are used
//            classification bits are set using entries in the tables
//            
//  Modification:
//  -- faster J/psi selection by changing order of cuts  
//                                                      01/12/99 T.Oest
//  -- changed the cuts in the J/psi selection, 
//     now each lepton must be identified either by
//     Bremsstrahlung, RICH or OTR                      05/06/00 D.Samtleben
//
//  -- take care of all ECAL clusters assigned to an RTRA entry
//  -- use VDS-ECAL momentum
//                                                      12/07/00 D.Samtleben
//  -- polished code so that even I can understand it   17/07/00 AG 
//
//  -- check the opposite charge of the tracks
//     from VDS-ECAL momentum and not from RTRA momentum
//                                                      21/07/00 D.Samtleben
//  -- fill RTRA-TOPP relation for Jpsi
//                                                      21/07/00 D.Dujmic
//  -- Make an RTRA entry also for Upsilon selection
//       if invariant mass >6.0 GeV 
//       no identification/confirmation is required
//     Mass without added br is stored in the "leh" field 
//                                                      29/11/00 M.Villa
//  -- changed name to rcSelectJPSIEE                   08/12/00 AG
//  -- Changed vertex finding routine: JPSIvertex.
//     Now identified electrons (loose cuts) are passed to the JPSIvertex
//     routine to search for detached VX.
//     Cut on E_t of cluster reduced to 0.9 GeV/c       12/12/00 M.Villa
//  -- added some debug printout                        15/12/00 AG
// ********************************************************************

// standard includes
#include <iostream.h>
#include <stdlib.h>
#include <math.h>

// ARTE includes
#include "arte/arte.hh"
#include "farm/farm.hh"
#include "reco/rc99Areco.hh"
#include "arte/RSEG_cmp.hh"
#include "arte/ArteUtil.h"
#include "reco/rcXkinema.h"
#include "reco/JPSIvertex.hh"

// externals
extern int FADEB; // debug flag steered by runargs=-debug

// globals
float cutJPSIEE     = 2.0; // default di-lepton low mass cut
float cutRECYPS     = 6.0; // default di-lepton high mass cut
float cutRECDVlower = 0.5; // default lower detached vertex cut
float cutRECDVupper = 3.0; // default higher detached vertex cut

// prototypes
float rc99Bremsstrahlung(ArtePointer<RTRA>);
int
JPSIvertex(ArtePointer<RTRA> track1,ArtePointer<RTRA> track2,
	   JPSIVtxType t);

//***************************************************************
// function
//*****************************************************************

int rcSelectJPSIEE()
{
  ARTE_mark( "rcSelectJPSIEE 0", 0 );

  // print
  if( FADEB ){
    cout << "JPSIEE: "
	 << "cutRECYPS="     << cutRECYPS     << " "
	 << "cutJPSIEE="     << cutJPSIEE     << " "
	 << "cutRECDVlower=" << cutRECDVlower << " "
	 << "cutRECDVupper=" << cutRECDVupper << endl;
  }
 
  //==================================================================
  // 1. Cuts
  //==================================================================
  float dyCut         = 3.;
  float pCut          = 5.;
  float ptCut         = 0.0;
  float EtCut         = 0.9;
  float EoverPCut     = 0.36;
  float MassCut       = cutJPSIEE;
  float MassCutUps    = cutRECYPS; // default to 6.0 (can appear in a kuip cmd)  

  //==================================================================
  // 
  //==================================================================
  float px1, py1, pz1, p1, px2, py2, pz2, p2, pt1, pt2, mom1, mom2;
  float Et1, Et2, Ee1, Ee2, Eb1, Eb2;
  float px, py, pz, p, mee, meecorr, pt;
  float dy1, dy2;
  
  // RTRA loop 1
  for( ArteTable<RTRA>::iterator rtra1 = ArteTable<RTRA>::begin();
       rtra1 != ArteTable<RTRA>::end(); ++rtra1 ) {
    
    ArtePointer<RSEG> rVDS1  = 0;
    ArtePointer<RSEG> rRICH1 = 0;
    ArtePointer<RSEG> rOTR1  = 0;
    
    // RSEG loop 1 (many RSEG per RTRA)
    for (ArteRelationIterator<RSEG> rseg2 = (*rtra1)->rseg.begin();
         rseg2 != (*rtra1)->rseg.end(); ++rseg2){
      
      if ((*rseg2)->fit <= 0 ||
	  (*rseg2)->fit == 8 ||
          (*rseg2)->fit == 10){
	
	if ((*rseg2)->cmp & Rsegc::vxd)  rVDS1  = *rseg2;
	if ((*rseg2)->cmp & Rsegc::rich) rRICH1 = *rseg2;
	if ((*rseg2)->cmp & Rsegc::patt) rOTR1  = *rseg2;
      }
    }
    
    // RCCL loop 1 (many RCCL per RTRA)
    for (ArteRelationIterator<RCCL> rccl1=(*rtra1)->rccl.begin();
	 rccl1 != (*rtra1)->rccl.end(); ++rccl1){       
      
      p1=-10000.;
      mom1=-10000.;
      // RSEG (many RSEG per RCCL)
      for(ArteRelationIterator<RSEG> rmatit1=(*rccl1)->rseg.begin();
	  rmatit1!= (*rccl1)->rseg.end();rmatit1++){ 
	
	if ((*rmatit1)->fit==7){
	  
	  for(ArteRelationIterator<RSEG> clvdsit1=(*rmatit1)->rend.begin();
	      clvdsit1!= (*rmatit1)->rend.end();clvdsit1++){ 
	    
	    if ((*clvdsit1)->fit!=7  && ((*clvdsit1)->cmp & Rsegc::vxd) 
		&& (*clvdsit1) == rVDS1){
	      
	      mom1=1./(*rmatit1)->pf;
	      p1=fabs(mom1);
	      //cout << "yes, found it .. " << (*rtra1) << " " << (*rmatit1) << endl;
	    }
	  }
	}
      }
      //p1  = fabs(1./(*rtra1)->pf);
      
      if (p1 > pCut && fabs((*rccl1)->esum/fabs(p1)-1.)< EoverPCut){
	//cout << "found a track ... " << *rtra1 << " " << (*rccl1)->esum << endl;
	
	pz1 = p1 / sqrt(1.+(*rtra1)->txf*(*rtra1)->txf +
                        (*rtra1)->tyf*(*rtra1)->tyf);
	px1 = (*rtra1)->txf * pz1;
	py1 = (*rtra1)->tyf * pz1;
	pt1 = sqrt(px1*px1+py1*py1);
	Et1 = (*rccl1)->esum * 
	  sqrt((*rccl1)->x*(*rccl1)->x+(*rccl1)->y*(*rccl1)->y)/(*rccl1)->z;
	dy1 = (*rccl1)->y -
	  ((*rtra1)->yf + (*rtra1)->tyf *((*rccl1)->z - (*rtra1)->zf));
	
	//
	// particle 1
	//
	if (pt1 > ptCut && Et1 > EtCut && fabs(dy1) < dyCut){
	  
	  // RTRA loop 2
	  for( ArteTable<RTRA>::iterator rtra2 = rtra1;
	       rtra2 != ArteTable<RTRA>::end(); ++rtra2 ) {
	    
	    ArtePointer<RSEG> rVDS2 = 0;
	    ArtePointer<RSEG> rRICH2 = 0;
	    ArtePointer<RSEG> rOTR2 = 0;
	    
	    // RSEG loop 2
	    for (ArteRelationIterator<RSEG> rseg2 = (*rtra2)->rseg.begin();
		 rseg2 != (*rtra2)->rseg.end(); ++rseg2){
	      
	      if ((*rseg2)->fit <= 0 ||
		  (*rseg2)->fit == 8 ||
		  (*rseg2)->fit == 10){
		
		if ((*rseg2)->cmp & Rsegc::vxd)  rVDS2  = *rseg2;
		if ((*rseg2)->cmp & Rsegc::rich) rRICH2 = *rseg2;
		if ((*rseg2)->cmp & Rsegc::patt) rOTR2  = *rseg2;
	      }
	    }
	    
	    // RCCL loop 2
	    for(ArteRelationIterator<RCCL> rccl2=(*rtra2)->rccl.begin();
		rccl2!= (*rtra2)->rccl.end();rccl2++){       
	      
	      p2=-10000.;
	      mom2=-10000.;
	      
	      for(ArteRelationIterator<RSEG> rmatit2=(*rccl2)->rseg.begin();
		  rmatit2!= (*rccl2)->rseg.end();rmatit2++){ 
		
		if ((*rmatit2)->fit==7){
		  
		  for(ArteRelationIterator<RSEG> clvdsit2=(*rmatit2)->rend.begin();
		      clvdsit2!= (*rmatit2)->rend.end();clvdsit2++){ 
		    
		    if ((*clvdsit2)->fit!=7  && ((*clvdsit2)->cmp & Rsegc::vxd) 
			&& (*clvdsit2) == rVDS2){
		      
		      mom2=1./(*rmatit2)->pf;
		      p2=fabs(mom2);
		      //cout << "yes, found it .. " << (*rtra2) << " " << (*rmatit2) << endl; 
		    }
		  }
		}
	      }
	      //p2  = fabs(1./(*rtra2)->pf);
	      
	      //
	      // particle 2
	      //
	      if (mom1 * mom2 < 0 
		  && p2 > pCut && fabs((*rccl2)->esum/fabs(p2)-1.) < EoverPCut){
		
		//cout << "here we go ... " << (*rtra2) << " " << (*rccl2)->esum << endl;
		
		if ((*rccl1) != (*rccl2) && rVDS1 != rVDS2){
		  
		  pz2 = p2 / sqrt(1.+(*rtra2)->txf*(*rtra2)->txf +
				  (*rtra2)->tyf*(*rtra2)->tyf);
		  px2 = (*rtra2)->txf * pz2;
		  py2 = (*rtra2)->tyf * pz2;
		  pt2 = sqrt(px2*px2+py2*py2);
		  Et2 = (*rccl2)->esum * 
		    sqrt((*rccl2)->x*(*rccl2)->x+(*rccl2)->y*(*rccl2)->y)/(*rccl2)->z;
		  dy2 = (*rccl2)->y -
		    ((*rtra2)->yf + (*rtra2)->tyf *((*rccl2)->z - (*rtra2)->zf));
		  
		  //
		  // J/Psi candidate
		  // 
		  if (pt2 > ptCut && Et2 > EtCut && fabs(dy2) < dyCut){
		    
		    px  = px1 + px2;
		    py  = py1 + py2;
		    pz  = pz1 + pz2;
		    p   = sqrt(px*px+py*py+pz*pz);
		    pt  = sqrt(px*px+py*py);
		    
		    //-----------------------------------------------------------------
		    //   electron electron mass
		    //-----------------------------------------------------------------
		    float corr1, corr2;
		    
		    Ee1    = p1;
		    Ee2    = p2;
		    Eb1    = rc99Bremsstrahlung((*rtra1));
		    Eb2    = rc99Bremsstrahlung((*rtra2));
		    
		    mee    = sqrt( 2.*(Ee1*Ee2 - px1*px2 - py1*py2 - pz1*pz2));
		    
		    if (Ee1 != 0.0 && Ee2 != 0.0) {
		      corr1     = 1. + Eb1 / Ee1;
		      corr2     = 1. + Eb2 / Ee2;
		      meecorr   = mee * sqrt(corr1 * corr2);
		    } else {
		      meecorr   = 0.0;
		    }

		    //-----------------------------------------------------------------
		    //   mass cut
		    //-----------------------------------------------------------------
		    //		    cout << "and now? " << meecorr << "  " << Eb1 << " " << rRICH1 << " " <<rOTR1 << " " << Eb2 << " " << rRICH2 << " " << rOTR2 << endl;
		    if (meecorr > MassCutUps ||
 			  (meecorr > MassCut &&
			    (Eb1>0||rRICH1!=0||rOTR1!=0)&&(Eb2>0||rRICH2!=0||rOTR2!=0)) ){
		      //	     &&((Eb1>0||Eb2>0)||(rRICH1!=0&&rRICH2!=0))){
		      //-----------------------------------------------------------------
		      //   vertex cut
		      //-----------------------------------------------------------------
		      
		      int rc = 
			JPSIvertex(*rtra1 ,*rtra2, ELECTRONS_ALWAYS);

		      if( rc !=0 ){

			ArtePointer<RTRA> rtra = ArteTable<RTRA>::size();
			//-----------------------------------------------------------------
			//   fill RTRA and vertex bank
			//-----------------------------------------------------------------
			
			rtra->lee = meecorr;
			rtra->leh = mee;
			rtra->cmp = (*rtra1)->cmp | (*rtra2)->cmp;
                      }
		    } else {

		      int rc = 
			JPSIvertex(*rtra1 ,*rtra2, ELECTRONS_DETACHED);
		      if( rc>0 ){
		        ArtePointer<RTRA> rtra = ArteTable<RTRA>::size();
			rtra->lee = meecorr;
			rtra->leh = mee;
			rtra->cmp = (*rtra1)->cmp | (*rtra2)->cmp;
                      }
                    }
		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }
  ARTE_mark( "rcSelectJPSIEE 1", 0 );
  return 0; 
}
