// **********************************************************************************************************************************
//
//    K0, Lambda and Lambda_bar Selection routine 
//
//     Author F.Sanchez (Nov'00)
//
//      Include "rcSelect_K0S_Lambda.hh"
//
//      ARTE ONLINE VERSION:  call "rcSelect_K0S_Lambda()" after reconstruction. It initialize automatically and 
//                            is set to true if one K0S or lambda is found. Optimal cuts are set, and no ntuple 
//                            is created. The RTRA associated to V0 has a toop :
//
//                                      kinema::findTopp(310) for K short.
//                                      kinema::findTopp(3122) for Lambda.
//                                      kinema::findTopp(-3122) for anti Lambda. 
//
//
//       To initialize: K0LambdaSelectInit(void);                                      -> No cuts, does not produce an ntuple.
//                      K0LambdaSelectInit(char ntuplefile )                           -> No cuts. It does produe an ntuple.
//                      K0LambdaSelectInit(cu1,cu2,cu3)                                -> Cuts, does not produce an ntuple.
//                      K0LambdaSelectInit(cu1,cu2,cu3,char ntuplefile )               -> Cuts, produce an ntuple.
//                         cu1  = minimum distance between tracks. (cm).
//                         cu2  = flight path of K0S in K0S restframe. (cm).
//                         cu3  = Impact parameter to the closest primary (cm).
//
//                     K0LambdaSelectMassWindow(kshortlow,kshorthigh,lambdalow,lambdahigh) -> define mass windows
//                     
//                     K0LambdaSelectVertexPackage(VT) or K0LambdaSelectVertexPackage(GROVER) selects the vertex package.
//                                                                                            Default is GROVER.
//
//       Each event:    K0LambdaSelect() returns the number of K0S in the event. It creates a new VERTEX with 
//                                      k0s particle as an rtra with fid == 310.
//
//       Stop funct:    K0LambdaClose()  close ntuples if opened.
//
//       For each K0 or lambda found a new vertex is created. The new vertex can be identified as follows:
//
//   if( K0LambdaSelect() ) 
//     {    
//       for( ArteTable<RVER>::iterator rverIt = ArteTable<RVER>::begin();
//  		       rverIt != ArteTable<RVER>::end() ; ++rverIt) 
//         {
//  	     ArtePointer<RTRA> part = rverIt->btra;
//  	     switch( (part->topp)->lund ) 
//  	       {
//               case 310:
//                  .................Kshort.............
//                  break;
//               case 3122:
//  	            ................LAMBDA..............
//                  break;
//               case -3122:
//  	            ................LAMBDA_BAR..............
//                  break;
//               default:
//                  .............OTHER VERTECES............
//                  break;
//  	       }
//         }
//     }
//
// Changes:
// 04 Dec 2000 (AG) corrected include path (reco/)
// 12 Dec 2000 (FS) more efficient lambda <-> k0s selection in the kinematically shared region.
// 14 Dec 2000 (FS) RHP histograms were not initialized.
// 19 Dec 2000 (FS) New function to enable the ONLINE mode. 
//  9 Jan 2000 (FS) New mass window for Lambdas.
// 18 Jan 2001 (FS) New track selection for tracks with ITR hits and change of track quality criteria. 
//
// **********************************************************************************************************************************


// includes

#include <stdio.h>
#include <math.h>
#include <vector.h>
#include <map.h>

#include "CLHEP/Vector/LorentzVector.h"
#include "CLHEP/Vector/LorentzRotation.h"
#include "reco/rcSelect_K0S_Lambda.hh"
#include "marple/marple.hh"
#include "arte/ArteUtil.h"
#include "reco/rcFitFlag.hh"

#include "rhp.h"
#include "dq.h"


extern "C" {
  void hgevnt_();
  void dgevnt_();
  void l1evnt_();
  void rcevnt_();
}

extern "C" {
#include "hbook.h"
void hcdir_(char *,char *,int, int);
}

#include "reco/rcXkinema.h"

#include "arte/arte.hh"
#include "arte/ArteFlag.h"
#include "arte/ArteFlag.h"
#include "farm/farm.hh"
#include "farm/fa.hh"

// vt++ package.


#include "vt++/Vt++.hh"
#include "vt++/VtUtil.hh"
#include "vt++/VtDistance.hh"
#include "vt++/VtMatrix.hh"
#include "vt++/VtKalman.hh"


// grover package

#include "grover/grover.hh"

using namespace VERTEX;
using namespace MATRIX;

int trSelect( vector< ArtePointer<RTRA> > &rtra, int &clpos, int &clneg);

void efmasser(int &npar, float ams[], float ppa[][50], float covp[][3][50],
              float &efms, float &errmas, float pout[], float covout[][3]);

void ptrcov (int &ifl, float par[], float covpar[][3],
                       float vec[], float covvec[][3]);

float distance(float vect[3],float pos[3],float poin[3]);

void createRVER(Vertex *rvv,float mass, int part);
void createRVER(GroverVertexList *rvv, int part);
void RVERrelations(ArtePointer<RVER> rver , int fid );

bool MassCut(int cand, float mass);
void SelectNextTry(float *candidate, int Q, float *mass1, float *mass2);

static const int index[5][5] = { {0, 1,  3, 6,10},
                                 {1, 2,  4, 7,11},
                                 {3, 4,  5, 8,12},
                                 {6, 7,  8, 9,13},
                                 {10,11,12,13,14} };

float Pion   = 0.13957;
float Proton = 0.93827;

ntuple evt;

struct {
    int iquest[100];
  } quest_;

static int FillNtuple;
static bool Online = FALSE;
static float DistCut_saved;
static float FlightCut_saved;
static float ImpactParamCut_saved;
static float K0LowMassCut_saved;
static float K0HighMassCut_saved;
static float LambdaLowMassCut_saved;
static float LambdaHighMassCut_saved;

vtxpck pack_save = GROVER;


void rc_Select_K0S_Lambda_OnlineMode()
{
  Online = TRUE;
}

bool rcSelect_K0S_Lambda()
{
  ARTE_mark( "rcSelect_K0S_Lambda 0", 0 );

  static bool first = FALSE;

  if ( !first )
    {  
      rc_Select_K0S_Lambda_OnlineMode();
      K0LambdaSelectInit(0.04,0.13,0.06); 
      K0LambdaSelectVertexPackage(GROVER);
      first = TRUE;
    }
    
  if( K0LambdaSelect() ) return TRUE;


  ARTE_mark( "rcSelect_K0S_Lambda 1", 0 );
  return FALSE;
}


void K0LambdaSelectVertexPackage(vtxpck pack)
{
  pack_save = pack;
}

void K0LambdaSelectMassWindow( float k0low, float k0high, float Lmbdlow, float Lmbdhigh )
{
  K0LowMassCut_saved = k0low;
  K0HighMassCut_saved = k0high;
  LambdaLowMassCut_saved = Lmbdlow;
  LambdaHighMassCut_saved = Lmbdhigh;
}


void K0LambdaSelectClose(void)
{
  int i = 0;

  if( !FillNtuple ) return;

  cout << " Calling HCDIR \n";

  hcdir_("//K0Lambda"," ",strlen("//K0Lambda"),strlen(" "));

  cout << " Calling HROUT \n";

  hrout(0,i,"T");

  cout << " Calling HREND \n";

  hrend("K0S");

}

void K0LambdaSelectInit(void)
{
  K0LambdaSelectInit(0.1,-10000.,10000.,0," ");
}


void K0LambdaSelectInit(float DistCut, float FlightCut, float ImpactParamCut )
{

 K0LambdaSelectInit(DistCut, FlightCut, ImpactParamCut,0," ");

}

void K0LambdaSelectInit(float DistCut, float FlightCut, float ImpactParamCut,char *ntuplefile )
{

 K0LambdaSelectInit(DistCut, FlightCut, ImpactParamCut,1,ntuplefile);

}


void K0LambdaSelectInit(char *ntuplefile)
{
  K0LambdaSelectInit(0.1,-10000.,10000.,1,ntuplefile);
}



void K0LambdaSelectInit(float DistCut, float FlightCut, float ImpactParamCut , int EnableNtuple, char *ntuplefile )
{
  FillNtuple = EnableNtuple;
  DistCut_saved = DistCut;
  FlightCut_saved = FlightCut;
  ImpactParamCut_saved = ImpactParamCut;
  K0LowMassCut_saved = 0.428;
  K0HighMassCut_saved = 0.568;
  LambdaLowMassCut_saved = 1.10;
  LambdaHighMassCut_saved = 1.13;

  cout<< endl << endl ;
  cout<<"---------------------------------------------------------------- \n"; 
  cout<<" Initialization of K short selection \n";
  cout<<"   Cuts : \n";
  cout<<"     Maximum distance between tracks " << DistCut << " cm \n";
  if( FlightCut != -10000. ) 
  cout<<"     Minimum distance of K0 flight path (K0S rest frame) " << FlightCut_saved << " cm \n";
  if(  ImpactParamCut != 10000.  ) 
  cout<<"     Impact parameter to closest primary vertex " << ImpactParamCut_saved << " cm \n";
  cout<<"     Mass(K0S)= (" << K0LowMassCut_saved << " , " << K0HighMassCut_saved << " )  GeV \n";
  cout<<"     Mass(LAMBDA)= (" << LambdaLowMassCut_saved << " , " << LambdaHighMassCut_saved << " )  GeV \n";
  cout<<"   Options: \n";
  if( FillNtuple ) 
  cout<<"     Ntuple output " << ntuplefile <<" is selected. \n"; 
  else
  cout<<"     Ntuple output disable \n " ;
  cout<<"---------------------------------------------------------------- \n\n\n";
  

  if( FillNtuple ) 
    {
      int istat = 0;

      char Ntuple[]="run,evt,time,bx,tcode,part,chi2,dist,distv,angle,mass,mass_corr,mass_err,xs,ys,zs,q1,nhtr1,px1,py1,pz1,px1l,py1l,pz1l,xot1,yot1,x1,y1,z1,q2,nhtr2,px2,py2,pz2,px2l,py2l,pz2l,xot2,yot2,x2,y2,z2,flgt,nprim,txot1,txot2";
      
  
      cout << " Open " << ntuplefile << " for ntuples \n";
      
      quest_.iquest[9] = 65000;

      marpleinit_();
      
      hropen(1,"K0Lambda",ntuplefile,"N",1024,istat);
      
      if( istat ) cout << " Error opening file \n";
      
      cout << " Book " << Ntuple  << " ntuple \n";

      hbook1( 8000 ,"Selected events (1=k0s,2=lamb,3=lamb_b)",  3,  0.5,  3.5 ); 
      hbook1( 8001 ,"K short    mass", 50,  0.45,  0.545 ); 
      hbook1( 8002 ,"Lambda     mass", 25,   1.1,   1.13 ); 
      hbook1( 8003 ,"Lambda bar mass", 25,   1.1,   1.13 ); 

      hbookn(3,"Selection","//K0Lambda",Ntuple);
            
      return;
    }
  else if ( Online )  
    {

      cout<<"     Online version enabled \n " ;

      rhp_hbook1( DQ_ID( COMP_RECO,LVL_REC, 8000 ),"Selected events (1=k0s,2=lamb,3=lamb_b)",  3,  0.5,  3.5 ); 
      rhp_hbook1( DQ_ID( COMP_RECO,LVL_REC, 8001 ),"K short    mass", 50,  0.45,  0.545 ); 
      rhp_hbook1( DQ_ID( COMP_RECO,LVL_REC, 8002 ),"Lambda     mass", 25,   1.1,   1.13 ); 
      rhp_hbook1( DQ_ID( COMP_RECO,LVL_REC, 8003 ),"Lambda bar mass", 25,   1.1,   1.13 ); 

    }


  return;  
}


int K0LambdaSelect()
{
  int n_selected = 0;

  // get VDS-OTR tracks - quality check

  vector< ArtePointer<RTRA> > rtra;
  int pos = 0;
  int neg = 0;

  int ntall = trSelect(rtra, pos, neg);
 
  // At least two track with opposite sign 

  if ( ntall <= 1 || pos == 0 || neg == 0 ) return 0;
 
  // fill kinematic quantities for selected tracks into STL vectors

  int first = 1;

  vector< float > tx;
  vector< float > ty;
  vector< float > px;
  vector< float > py;
  vector< float > pz;
  vector< float > p;
  vector< float > Q;

  vector< ArtePointer<RTRA> > track;
  vector< ArteTable<RVER>::iterator > primary;

  int   ntr = 0;
 
  // Some definitions for kinematic
  
  for( int i = 0; i != ntall ; i++ ) {

    // px/pz
    float Tx    = rtra[i]->txf;
    // py/pz
    float Ty    = rtra[i]->tyf;
    // Q/p  
    float P     = fabs(1./rtra[i]->pf);
    // Pz
    float Pz    = P/sqrt(1.+Tx*Tx+Ty*Ty);
    // Px
    float Px    = Pz*Tx;
    // Py
    float Py    = Pz*Ty;
    
    float Charge = 1.;
    if( rtra[i]->pf < 0 ) Charge = -1.;

    ntr++;
    track.push_back( rtra[i] );
    tx.push_back(Tx);
    ty.push_back(Ty);
    px.push_back(Px);
    py.push_back(Py);
    pz.push_back(Pz);
    p.push_back( P );
    Q.push_back(Charge);
  }

  if ( ntr < 2 ) return 0;

  //
  // 2-track vertexing
  // -----------------
  
  float dz;
  float xs, ys, zs;
   
  int ifl;
  
  int npar; 
  float ams[50], ppa[3][50], covp[3][3][50];
  float efems, errmas, pout[3], covout[3][3];
  int nprimary = 0;
  
  for(int i=1; i!=ntr; ++i) 
    {
      for(int j=0; j!=i; ++j) 
	{
	  Vertex vs;
	  GroverVertexList vtxsel;
	  float mass1,mass2;
          // Look for opposite charge tracks.

	  if( Q[i] == Q[j] ) continue;

	  // Look for candidate 2-track secondary vertices

	  if( pack_save == VT )
	    {
	      Vertex vs(track[i], track[j]);
	      if( !vs.findVertexVt() ) continue;
	      vs.rmsDistAngle();

	      xs = vs.vx();
	      ys = vs.vy();
	      zs = vs.vz();
	      evt.dist = vs.dist();
	  
	      if( evt.dist > DistCut_saved ) continue;

	      // Distance to the target.
	      
	      dz  = zs+5.;

	      if( dz < 2.0 ) continue;

	      // Generic mass spectra for any pair that vertexes
	      
	      px[i] = vs[0]->kalman.px();
	      px[j] = vs[1]->kalman.px();
	      py[i] = vs[0]->kalman.py();
	      py[j] = vs[1]->kalman.py();
	      pz[i] = vs[0]->kalman.pz();
	      pz[j] = vs[1]->kalman.pz();

	      evt.angle = vs.angle();
	      evt.chi2 = vs.chi2();
 
	      // Start with the K0

	      evt.part = 1;
	      mass1 = Pion;
	      mass2 = Pion;

	      evt.mass  = vs.mass(Pion,Pion);

	      if( !MassCut((int)evt.part, evt.mass) )
		{
		  // Lambda 
		  if( Q[i] > 0 ) 
		    { 
		     mass1 = Proton;
		     mass2 = Pion;
		    }
		  else 
		    {
		     mass1 = Pion;
		     mass2 = Proton;  
		    }
 		  evt.part = 2;
		  evt.mass = vs.mass(mass1,mass2);
		  if( !MassCut((int)evt.part, evt.mass) )
		    { 
		      // Lambda Bar
		      if( Q[i] < 0 ) 
			{ 
			  mass1 = Proton;
			  mass2 = Pion;
			}
		      else 
			{
			  mass1 = Pion;
			  mass2 = Proton;  
			}	
		      evt.part = 3;
		      evt.mass = vs.mass(Proton,Pion); 
		      if( !MassCut((int)evt.part, evt.mass) ) continue;
		    } 
		}

	    }
	   else
	    {
	      GroverParticle pi1(track[i]), pi2(track[j]);
	      pi1.GlobalToLocal(); 
	      pi2.GlobalToLocal();

	      // If the momemtum cocient is smaller than 4.5 we start to check the k0 short.
	      if( max(p[i]/p[j],p[j]/p[i]) < 4.5 ) 
		{
		  evt.part = 1;
		  mass1 = Pion;
		  mass2 = Pion;
		}
	      else
		{
	      //  If not, we check first the Lambda or the Lambda_bar.
		  if( p[i] > p[j] )
		    {
		      mass1 = Proton;
		      mass2 = Pion;
		      if( Q[i] > 0 ) 
			evt.part = 2;
		      else
			evt.part = 3;
		    }
		  else
		    {
		      mass1 = Pion;
		      mass2 = Proton;
		      if( Q[i] < 0 ) 
			evt.part = 2;
		      else
			evt.part = 3;
		    }
		}
 
	      pi1.SetMass(mass1);
	      pi2.SetMass(mass2);

	      GroverVertex vtx(&pi1,&pi2);
	      //and make an unconstrained fit
	      vtx.Fit(-1.0);

	      xs = vtx.X();
	      ys = vtx.Y();
	      zs = vtx.Z();
	      dz = zs+5.;
	      
	      if( dz < 2.0 ) continue;

	      evt.dist = pi1.DistanceTo(&pi2);  

	      if( evt.dist > DistCut_saved ) continue;

	      evt.mass = vtx.Mass();

	      evt.chi2 = vtx.Chi2();

	      if( !MassCut((int)evt.part,evt.mass ) ) 
		{
		  SelectNextTry(&evt.part, (int) Q[i],&mass1,&mass2);
		  pi1.SetMass(mass1);
		  pi2.SetMass(mass2);
		  GroverVertex vtx(&pi1,&pi2);
		  vtx.Fit(-1.0);
		  evt.mass = vtx.Mass();
		  evt.chi2 = vtx.Chi2();
		  if( !MassCut((int)evt.part,evt.mass ) ) 
		    {
		      SelectNextTry(&evt.part,(int) Q[i],&mass1,&mass2);
		      pi1.SetMass(mass1);
		      pi2.SetMass(mass2);
		      GroverVertex vtx(&pi1,&pi2);
		      vtx.Fit(-1.0);
		      mass1 = Pion;
		      mass2 = Proton;
		      evt.mass = vtx.Mass();
		      evt.chi2 = vtx.Chi2();
		      if( !MassCut((int)evt.part,evt.mass ) ) 
			continue;
		      else
			vtxsel.push_back(vtx);
		    }
		  else
		    vtxsel.push_back(vtx);
		}
	      else 
		vtxsel.push_back(vtx);
	    }
	  
	  float mom[3],pos[3],vertex[3];
	  int primvtx = 0;
	  
	  evt.q1 = Q[i];
	  evt.q2 = Q[j];
	  
	  // Impact parameter to vertex.
	  
	  mom[0] = px[i]+px[j];
	  mom[1] = py[i]+py[j];
	  mom[2] = pz[i]+pz[j];
	  
	  pos[0] = xs;
	  pos[1] = ys;
	  pos[2] = zs;
	  
	  if( first )
	    {
	      for( ArteTable<RVER>::iterator rverIt = ArteTable<RVER>::begin();
		   rverIt != ArteTable<RVER>::end() ; ++rverIt) 
		{
		  if( rverIt->nt >= 4 )
		    {
		      nprimary++;
		      primary.push_back( rverIt );
		    }
		}
	      
	      evt.nprim =(float)nprimary;
	      first = 0;
	    }
	  
	  float mindist = 10.;
	  
	  for(int  ll=0; ll<nprimary ; ll++) 
	    {
	      float local;
	      
	      vertex[0] =  primary[ll]->xv;
	      vertex[1] =  primary[ll]->yv;
	      vertex[2] =  primary[ll]->zv;
	      
	      local = distance(mom,pos,vertex);
	      
	      if( local < mindist && local > 0.0 )
		{
		  mindist = local;
		  primvtx = ll;
		}
	    }
	  
	  if ( mindist > ImpactParamCut_saved ) continue;
	  
	  // Boost to K0Lambda rest frame.

	  HepLorentzVector plab1(px[i],py[i],pz[i],sqrt(px[i]*px[i]+py[i]*py[i]+pz[i]*pz[i]+mass1*mass1));
	  HepLorentzVector plab2(px[j],py[j],pz[j],sqrt(px[j]*px[j]+py[j]*py[j]+pz[j]*pz[j]+mass2*mass2));

	  HepLorentzVector boost = plab1+plab2;
	  HepLorentzRotation RotBoost(boost.boostVector());
	  
	  if( nprimary != 0 ) 
	    {
	      evt.flight = sqrt((xs-primary[primvtx]->xv)*(xs-primary[primvtx]->xv)+
				(ys-primary[primvtx]->yv)*(ys-primary[primvtx]->yv)+
				(zs-primary[primvtx]->zv)*(zs-primary[primvtx]->zv))/(float)RotBoost.tt();
	    }
	  else
	    evt.flight = -100. ;
	  
	  if ( evt.flight < FlightCut_saved ) continue;
	  
          // -----------------------------------------------
          //    Create RVER
	  if( pack_save == VT )   
	    createRVER(&vs,evt.mass,(int)evt.part);	     
	  else
	    createRVER(&vtxsel,(int)evt.part);

	  // -----------------------------------------------
	  
	  n_selected++;

	  if( Online ) 
	    {
	      rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 8000 ), evt.part , 1. ); 
		       
	      rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 8000+(int)evt.part ), evt.mass , 1. ); 

	    }
	  else
	    {
	      hf1( 8000 , evt.part , 1. ); 
	      hf1( 8000+(int)evt.part , evt.mass , 1. ); 
	    }

	  
	  if( !FillNtuple ) continue;
	  
	  evt.run    =(float)(*ArtePointer<EVHD>(1)).run;
	  evt.evt    =(float)(*ArtePointer<EVHD>(1)).evt;
	  evt.time   =(float)(*ArtePointer<EVHD>(1)).time;
	  evt.bx     =(float)faDDAQphysBX();
	  evt.tcode  =(float)faDDAQFCS_tcode();
	  
	  evt.distv = mindist;    
	  
	  //----------- K^0_s --------------  
	  // Calculate the mass and the mass error
	  
	  float par1[3], vec1[3], covpar1[3][3], covvec1[3][3];
	  float par2[3], vec2[3], covpar2[3][3], covvec2[3][3];
	  
	  par1[0] = tx[i];
	  par1[1] = ty[i];
	  par1[2] = p[i];
	  
	  par2[0] = tx[j];
	  par2[1] = ty[j];
	  par2[2] = p[j];
	  
	  for(int mm=0; mm<3; mm++)
	    for(int ll=0; ll<3; ll++)
	      {
		covpar1[mm][ll] = track[i]->cvf[index[mm][ll]];
		covpar2[mm][ll] = track[j]->cvf[index[mm][ll]];
	      }
	  
	  ifl =0;
	  ptrcov(ifl,par1,covpar1,vec1,covvec1);
	  ptrcov(ifl,par2,covpar2,vec2,covvec2);
	  
	  for(int mm=0; mm<3; mm++)
	    {
	      ppa[mm][0] = vec1[mm];
	      ppa[mm][1] = vec2[mm];
	      for(int ll=0; ll<3; ll++)
		{
		  covp[mm][ll][0] = covvec1[mm][ll];
		  covp[mm][ll][1] = covvec2[mm][ll];
		}
	    }
	  
	  npar = 2;
	  ams[0] = mass1;
	  ams[1] = mass2;
	  efems  = 0.;
	  errmas = 10000.;
	  efmasser(npar, ams, ppa, covp, efems, errmas, pout, covout);
	  
	  //--------------------------------------
	  
	  evt.mass_corr = efems;
	  evt.mass_err =  errmas;
	  
	  evt.nhtr1 = track[i]->hou+1000*track[i]->hin;
	  evt.nhtr2 = track[j]->hou+1000*track[j]->hin;
	  
	  evt.xs = xs;
	  evt.ys = ys;
	  evt.zs = zs;
	  
	  RcTraPoint TrackF1(track[i]);
	  RcTraPoint TrackF2(track[j]);
	  
	  TrackF1.prolongate(700.);
	  TrackF2.prolongate(700.);
	  
	  evt.xot1 = TrackF1.x();
	  evt.yot1 = TrackF1.y();
	  
	  evt.xot2 = TrackF2.x();
	  evt.yot2 = TrackF2.y();
	  
	  evt.txot1 = TrackF1.tx();
	  evt.txot2 = TrackF2.tx();
	  
	  evt.px1 = px[i];
	  evt.px2 = px[j];
	  
	  evt.py1 = py[i];
	  evt.py2 = py[j];
	  
	  evt.pz1 = pz[i];
	  evt.pz2 = pz[j];
	  
	  evt.x1 = track[i]->xf;
	  evt.x2 = track[j]->xf;
	  
	  evt.y1 = track[i]->yf;
	  evt.y2 = track[j]->yf;
	  
	  evt.z1 = track[i]->zf;
	  evt.z2 = track[j]->zf;
	  
	  plab1 *= RotBoost.inverse();
	  plab2 *= RotBoost.inverse();
	  
	  evt.px1lab = plab1.px();
	  evt.py1lab = plab1.py();
	  evt.pz1lab = plab1.pz();
	  
	  evt.px2lab = plab2.px();
	  evt.py2lab = plab2.py();
	  evt.pz2lab = plab2.pz();
	  
	  hfn(3,(float*)&evt);
       
	}
    }
  
return n_selected;
}


bool MassCut( int candidate , float mass )
{

  switch ( candidate ) 
    {
    case 1: 
      if ( mass > K0LowMassCut_saved && mass < K0HighMassCut_saved ) return TRUE; 
      break;
    case 2: 
    case 3: 
      if ( mass > LambdaLowMassCut_saved && mass < LambdaHighMassCut_saved ) return TRUE;
      break;
    }

  return FALSE;
}


void SelectNextTry(float *candidate, int Q, float *mass1, float *mass2) 
{
	
  switch( (int) *candidate )
    {
    case 1:  // Last try was a K0s.
      // Lambda 
      if( Q > 0 ) 
	{ 
	  *mass1 = Proton;
	  *mass2 = Pion;
	}
      else 
	{
	  *mass1 = Pion;
	  *mass2 = Proton;  
	}
      *candidate = 2.;
      break;
      
    case 2:  // Last try was a Lambda.
      // Lambda bar
      if( Q < 0 ) 
	{ 
	  *mass1 = Proton;
	  *mass2 = Pion;
	}
      else 
	{
	  *mass1 = Pion;
	  *mass2 = Proton;  
	}
      *candidate = 3.;
      break; 
    case 3: // Last try was a Lambda_bar.
      *mass1 = Pion;
      *mass2 = Pion;
      *candidate = 1.;
      break;
    }
} 

void efmasser(int &npar, float ams[], float ppa[][50], float covp[][3][50],
              float &efms, float &errmas, float pout[], float covout[][3])

//     ******************************************************************
//     *                                                                *
//     * Calculate invariant mass of NPAR particles with masses         *
//     *      AMS and momenta PPA                                       *
//     *      COVP(i,j,ip) - the input cov matrix of the momenta)       *
//     *                                                                *
//     * Output: EFMS - inv. mass (Gev), POUT - momentum of combin.     *
//     *      ERRMAS - error of the mass,                               *
//     *      COVOUT - the cov of POUT(1..3)                            *
//     ******************************************************************

{

    // Local variables
    int np;
    double dpx, dpy, dpz;
    double defm,  dvar, derms, detot, de, dm;
    double derp[3][50], dcout[3][3], dpt[3];

    np = npar;
    if (npar > 50)
      {
	np = 50;
	cout<<"\n EFFMASER error: sorry, more than 50 particles.\n";
      }
    
    detot  = (float) 0.;
    dpt[0] = (float) 0.;
    dpt[1] = (float) 0.;
    dpt[2] = (float) 0.;

    for (int i = 0; i < np; i++)
      {
        dpx = ppa[0][i];
        dpy = ppa[1][i];
        dpz = ppa[2][i];
        dm  = ams[i];

        de = sqrt(dpx * dpx + dpy * dpy +
                  dpz * dpz + dm * dm);
        dpt[0] += dpx;
        dpt[1] += dpy;
        dpt[2] += dpz;
        detot  += de;
        if (de > 0.)
	  {
            derp[0][i] = dpx / de;
            derp[1][i] = dpy / de;
            derp[2][i] = dpz / de;
	  } 
	else
	  {
	    derp[0][i] = (float)0.;
            derp[1][i] = (float)0.;
            derp[2][i] = (float)0.;
          }
      }
    
    defm = sqrt(detot  * detot  - dpt[0] * dpt[0] -
                dpt[1] * dpt[1] - dpt[2] * dpt[2]);
    
    // --- Calculate the error on the mass
    
    for (int i = 0; i < 3; i++)
      {
        for (int j = 0; j < 3; j++)
	  {
            dcout[i][j] = (float)0.;
	  }
      }
    
    derms = (float)0.;
    if (defm > 1.e-6)
      {
        for (int ip = 0; ip < np; ip++)
	  for (int i = 0; i < 3; i++)
	    for (int j = 0; j < 3; j++)
	      {
		if (j >= i)
		  dvar = (double) covp[j][i][ip];
		else
		  dvar = (double) covp[i][j][ip];
		
		dcout[j][i] += dvar;
		derms += dvar * (detot * derp[i][ip] - dpt[i])
		  * ( detot * derp[j][ip]
		      - dpt[j] ) / (defm * defm);
	      }
      }

// --- Output

    efms   = (float) defm;
    errmas = (float) sqrt( ( fabs (derms) ) );
    for (int i = 0; i < 3; i++)
      {
        pout[i] = (float) dpt[i];
        for (int j = 0; j < 3; j++)
	  
	  covout[j][i] = (float) dcout[j][i];
      }
    
    return;
}


int trSelect( vector< ArtePointer<RTRA> > &rtra, int &clpos, int &clneg)
// *---------------------------------------------------------------------
// * return a vector with ArtePointers into RTRA
// *---------------------------------------------------------------------
{
  static int MINHITS_VDS  =  5;   // required minimum number of VDS hits
  static int MINHITS_MAIN = 10;  // required minimum number of Main tracker hits
  static int MINHITS_ITR  =  7;  // required minimum number of Main tracker hits

  map< float, ArtePointer<RTRA> > mrtra;

  //------------------------------------

  // put all VDS-OTR tracks into a sorted map

  for(ArteTable<RTRA>::iterator itrtra = ArteTable<RTRA>::begin();
      itrtra != ArteTable<RTRA>::end(); ++itrtra) 
  {
    if( (*itrtra)->fit != 0 ) continue;
  
    int nvds  = (*itrtra)->hsi;  //number of hits in inner tracker
    int nmain = (*itrtra)->hou+(*itrtra)->hin;  //number of hits in inner/outer tracker
    int notr  = (*itrtra)->hou;
    int nitr  = (*itrtra)->hin;

    if( nvds  < MINHITS_VDS ) continue;
    if( nmain < MINHITS_MAIN && nitr < MINHITS_ITR && notr < MINHITS_MAIN ) continue;

    float fom = float(nvds+nmain)-(*itrtra)->chi/float(nvds+nmain-5);
    mrtra.insert(pair< float, ArtePointer<RTRA> >(-fom,*itrtra));
  }
  // copy the tracks into the return vector, omitting all clones

  clpos = clneg = 0;
  for(map< float, ArtePointer<RTRA> >::iterator it=mrtra.begin();
      it != mrtra.end(); ++it) 
  {
    int OK = 1;
    for(unsigned int i=0; i!=rtra.size(); ++i) 
    {
      if(rtra[i]->csvd  == it->second->csvd ) OK=0;
      if(rtra[i]->cmain == it->second->cmain) OK=0;
      if(!OK) break;
    }
    if(OK) 
    {
      rtra.push_back(it->second);
      if(it->second->pf < 0) clneg++;
      if(it->second->pf > 0) clpos++;
    }
  }

  return rtra.size();
}

void ptrcov (int &ifl, float par[], float covpar[][3],
                       float vec[], float covvec[][3])
//     ******************************************************************
//     *                                                                *
//     *    Converts the covariance matrix of a momentum                *
//     *  PAR: SlopeX,SlopeY,P  ==>  VEC: Px,Py,Pz    if IFL=0          *
//     *  PAR: SlopeX,SlopeY,P <==   VEC: Px,Py,Pz    if IFL.NE.0       *
//     *                                                                *
//     *  Z coordinate (3) - beam direction                             *
//     *                                                                *
//     ******************************************************************
{
    float p;

    if (ifl == 0)
    {
        p = par[2];

        vec[2] = p / sqrt(par[0] * par[0] + (float)1. + par[1] * par[1]);
        vec[0] = vec[2] * par[0];
        vec[1] = vec[2] * par[1];

        covvec[0][0] =   covpar[2][2];
        covvec[1][0] =   covpar[2][2] * par[0]
                       + covpar[2][0] * p;
        covvec[2][0] =   covpar[2][2] * par[1]
                       + covpar[2][1] * p;
        covvec[1][1] =   covpar[2][2] * (par[0] * par[0])
                       + covpar[2][0] * (float)2. * par[0] * p
                       + covpar[0][0] * (p * p);
        covvec[2][1] =   covpar[2][2] * par[0] * par[1]
                       + covpar[2][0] * par[1] * p
                       + covpar[2][1] * par[0] * p
                       + covpar[1][0] * (p * p);
        covvec[2][2] =   covpar[2][2] * (par[1] * par[1])
                       + covpar[2][1] * (float)2. * par[1] * p
                       + covpar[1][1] * (p * p);
        covvec[0][1] =   covvec[1][0];
        covvec[0][2] =   covvec[2][0];
        covvec[1][2] =   covvec[2][1];

    }
    else
    {
        if ( fabs(vec[2]) > (float) 1.e-10 )
        {
            p = sqrt(vec[0] * vec[0] + vec[1] * vec[1] + vec[2] * vec[2]);
            par[0] = vec[0] / vec[2];
            par[1] = vec[1] / vec[2];
            par[2] = p;

            covpar[0][0] =  (covvec[0][0] * (par[0] * par[0])
                           - covvec[1][0] * (float)2. * par[0]
                           + covvec[1][1]) / (p * p);
            covpar[1][0] =  (covvec[0][0] * par[0] * par[1]
                           - covvec[2][0] * par[1]
                           - covvec[2][0] * par[0]
                           + covvec[2][1]) / (p * p);
            covpar[2][0] = (-covvec[0][0] * par[0]
                           + covvec[1][0]) / p;
            covpar[1][1] =  (covvec[0][0] * (par[1] * par[1])
                           - covvec[2][0] * (float)2. * par[1]
                           + covvec[2][2]) / (p * p);
            covpar[2][1] = (-covvec[0][0] * par[2] + covvec[2][0]) / p;
            covpar[2][2] =   covvec[0][0];
            covpar[0][1] =   covpar[1][0];
            covpar[0][2] =   covpar[2][0];
            covpar[1][2] =   covpar[2][1];
        }

    }

    return;
}


float distance(float vect[3],float pos[3],float poin[3])
{
  float diff;
  float vectn=0.0;
  float dd=0.0;
  float diffn=0.0;
  
  for(int i=0;i<3;i++)
    {
      diff = poin[i]-pos[i];
      dd += diff*vect[i];
      vectn += vect[i]*vect[i];
      diffn += diff*diff;
    }

  if( dd > 0 ) 
    return -sqrt(diffn-dd*dd/vectn);
  else 
    return sqrt(diffn-dd*dd/vectn);

}


//==============================================================================
// createRVER: save VT++ vertex in RVER table
//==============================================================================

void createRVER(Vertex *rvv, float mass, int part) {

  ArtePointer<RVER> rver = rvv->createRVER();

  rver->mass = mass;

  int fid = 0;

   switch( part ) 
    {
    case 1: 
      fid = 310;
      break;
    case 2:
      fid = 3122  ;
      break;
    case 3:
      fid = -3122  ;
      break;
    }    
     
     
  RVERrelations(rver,fid);

  return;
}


//==============================================================================
// createRVER: save GROVER vertex in RVER table
//==============================================================================

void createRVER(GroverVertexList *rvv, int part) {

  GroverAddToRVER(rvv,  FIT_SHORT);
 
  ArtePointer<RVER> rver = ArtePointer<RVER>(ArteTable<RVER>::size());

  int fid = 0;

  switch( part ) 
    {
    case 1: 
      fid = 310;
      break;
    case 2:
      fid = 3122  ;
      break;
    case 3:
      fid = -3122  ;
      break;
    }

  ArtePointer<RTRA> rtra  = rver->btra;

  rtra->fit =  FIT_SHORT;
  
  rtra->topp = kinema::findTopp(fid);

}


void RVERrelations(ArtePointer<RVER> rver, int fid )
 {
  float px=0.0,py=0.0,pz=0.0;
  float Q = 0.0;
  float p,pzl;

  for(ArteRelationIterator<RREL> it=rver->rrel.begin(); it!=rver->rrel.end(); ++it) 
    {
      ArtePointer<RREL> rrel = *it;
      ArtePointer<RTRA> track = rrel->rtra;
 
      if ( track->pf > 0 ) Q++;
      else Q--;
      p = fabs(1./track->pf);

      pzl    = p/sqrt(1.+track->txf*track->txf+track->tyf*track->tyf);
      pz   += pzl;
      px   += pzl*track->txf;
      py   += pzl*track->tyf;
    }
  
  // Store the new particle as an RTRA.
 
  RTRA rt;
  
  p = sqrt(px*px+py*py+pz*pz);
  rt.chi = 0;
  rt.ever = rver;
  rt.fid =   fid ;
  if( Q == 0 ) rt.pf = 1/p;
  else         rt.pf = Q/p;
  rt.txf = px/pz;
  rt.tyf = py/pz;
  rt.fit =   FIT_SHORT;    

  ArtePointer<RTRA> rtrap = ArteTable<RTRA>::push_back(rt);
  
  rver->btra = rtrap;
  rtrap->ever = rver;

  ArtePointer<TOPP> rtop = kinema::findTopp(fid);
  
  createRelation( rtop, rtrap, &TOPP::rtra );
  
  return;
}


	    







