//==============================================================================
//
// Routine to return No. of corresponding PC RSEG for an SLT refit track
//
// Input parameters:
//     int trigger  - No. of trigger (starting from 0)
//     int track    - No. of track in this trigger (0 or 1)
//
// Output parameters:
//     ArtePointer<RSEG> *rseg  - RSEG with minimal value of chi
//     float *chi               - Minimal value of chi
//
// Return value:
//     0          - success
//    -1          - problem extracting SLT information
//    -2          - no RSEGs found in PC region
//
// If chi>100. one can consider that corresponding RSEG is not found.
//
// Alexander Lanyov                              August 2000
//
//============================ Example of usage:
// #include "farm/farm.hh"
// extern "C" {
// #include "sltinfo.h"
//   int rcSltRsegPC(int trigger, int tra, ArtePointer<RSEG> *rseg, float *chi);
// }
// 
// if (faDDAQFCS_tcode()==9) return;
// if (SLTReadDSLT()!=SLT_OK) return;
// SLT_STATUS stat;
// SLT_TRIGGER trig;
// SLT_TRACK track;
// SLT_REFITTRACK refit;
// if (SLTStatus(&stat)!=SLT_OK) return;
// if (!(stat.status&SLTEVENT_ACCEPT)) return;
// for(int trigger=0;trigger<stat.ntriggers;trigger++) {
//   if (SLTTrigInfo(trigger,&trig)!=SLT_OK) continue;
//   if (trig.status==SLTTRIG_REJECT) continue;
//   for(int tra=0;tra<trig.ntracks;tra++) {
//     if (SLTTrackInfo(trig.trackindex[tra],&track)!=SLT_OK) continue;
//     if(track.status==SLTTRACK_ERROR) continue;
//     if(track.status!=SLTTRACK_ACCEPT) continue;
//     if (SLTRefitTrackInfo(trig.trackindex[tra],&refit)!=SLT_OK) continue;
//     ArtePointer<RSEG> rseg_;
//     float chi_;
//     int k = rcSltRsegPC(trigger, tra, &rseg_, &chi_);
//     printf("rcSltRsegPC(trigger=%d, track=%d, rseg=%d, chi=%f) = %d\n",
//              trigger, tra, rseg_.index(), chi_, k);
//   }
// }
// 
// changes:
// 20 Nov 2000 (AG) added math.h
// 01 Dec 2000 (AG) added function name to printout
//
//==============================================================================

// standard includes
#include <stdio.h>
#include <math.h>

// ARTE includes
#include <arte/arte.hh>
#include "arte/RSEG_cmp.hh"
#include "farm/farm.hh"

// includes
extern "C" {
#include "sltinfo.h"  // /afs/desy.de/group/hera-b/SLT/slt/Rev0502/include/sltinfo.h
  void rseg_at_z(ArtePointer<RSEG> rseg, float z,
			  float *x, float *y, float *tx, float *ty);
}

//#define DEBUG 1

////////////////////////////////////////////////////////////////////////////
// Get track estimates at coordinate Z (in lab system)
//
extern "C" void rcRsegZ(ArtePointer<RSEG> rseg, float z,
			  float *x, float *y, float *tx, float *ty) {

  float p;

  float f = (z-rseg->zf)/(rseg->ze-rseg->zf);
  float flimit=0.05;
  if (f < flimit) {
    *x = rseg->xf + rseg->txf * (z-rseg->zf);
    *y = rseg->yf + rseg->tyf * (z-rseg->zf);
    *tx = rseg->txf;
    *ty = rseg->tyf;
    p = rseg->pf;
  }
  else if (f > 1-flimit) {
    *x = rseg->xe + rseg->txe * (z-rseg->ze);
    *y = rseg->ye + rseg->tye * (z-rseg->ze);
    *tx = rseg->txe;
    *ty = rseg->tye;
    p = rseg->pe;
  }
  else {
    //printf("f = %f, zf=%f, ze=%f, z=%f\n",f,zf,ze,z);
    ArtePointer<RHIT> rhit1(0);
    ArtePointer<RHIT> rhit0(0);
    int n=0;
    int found=0;
    for (ArteRelationIterator<RHIT> rhitIt=rseg->rhit.begin() ;
	 rhitIt!=rseg->rhit.end() ; rhitIt++) {
      rhit1=*rhitIt;
      //printf("%3d. z=%f\n",n,rhit1->hitb->z);
      if (n && (rhit1->hitb->z - z) * (rhit0->hitb->z - z) <= 0.) {
	found=1;
	break;
      }
      rhit0=rhit1;
      n++;
    }
    if (found) {
      float z0 = rhit0->hitb->z;
      float z1 = rhit1->hitb->z;
      
      float dz=z1-z0;
      float f = (z-z0)/(z1-z0);
      if (fabs(dz)>50. && f<0.1) {
	f = 0.;
      }
      else if (fabs(dz)>50. && f>0.9) {
	f = 1.;
      }
      else if (z1>700. && z1<800. && z0<700.) {
	f = 1.;
      }
      else if (fabs(dz)<0.2 || fabs(dz)>50.) {
	f = 0.;
      //  printf("f = %f, z0=%f, z1=%f, z=%f\n", f,z0,z1,z);
      //  printf("Strange dz=%f for RSEG %d\n",dz,rseg.index());
      }
      
      *x = rhit0->x + (rhit1->x-rhit0->x)*f;
      *y = rhit0->y + (rhit1->y-rhit0->y)*f;
      *tx = rhit0->tx + (rhit1->tx-rhit0->tx)*f;
      *ty = rhit0->ty + (rhit1->ty-rhit0->ty)*f;

      p  = rhit0->p + (rhit1->p-rhit0->p)*f;
    }
    else {
      printf("rcRsegZ: f = %f, zf=%f, ze=%f, z=%f\n",
	     f, rseg->zf, rseg->ze, z);
      printf("rcRsegZ: Suitable RHITs not found, n=%d\n", n);
    }
  }
}

////////////////////////////////////////////////////////////////////////////

extern "C" int rcSltRsegPC(int trigger, int track,
			   ArtePointer<RSEG> *rseg, float *chi) {

  *chi = 1.e10;
  *rseg = ArtePointer<RSEG>(0);

  //------------------------------------------------------ Basic SLT info
  int daqcode=faDDAQFCS_tcode();
  
  // Detailed trigger definition is given at:
  // http://www-hera-b.desy.de/subgroup/shiftinfo/sub/DAQSHI.html

#ifdef DEBUG
  cout<<"rcSltRsegPC: Trigger Info:"<<endl;
  if (daqcode>=8) cout<<"rcSltRsegPC: Random (4Hz) triggered event"<<endl;
  if (daqcode==0||daqcode==8) cout<<"rcSltRsegPC: FLT triggered event"<<endl;
  if (daqcode==6||daqcode==14) cout<<"rcSltRsegPC: ''Test triggger'' event"<<endl;
#endif
  if (daqcode==9) return -1; /* No SLT information in non triggered events */

#ifdef DEBUG
  cout<<"rcSltRsegPC: *** Slt Info:***"<<endl;
#endif
  if (SLTReadDSLT()!=SLT_OK) {
    cout <<"rcSltRsegPC: No SLT information found"<<endl;
    return -1;
  }
  
  SLT_STATUS stat;
  SLT_TRIGGER trig;
  SLT_TRACK trac;
  SLT_REFITTRACK refit;

  if (SLTStatus(&stat)!=SLT_OK) {
#ifdef DEBUG
    cout <<"rcSltRsegPC: No status information"<<endl;
#endif
    return -1;
  }
  if (stat.status&SLTEVENT_BYPASS) {
#ifdef DEBUG
    cout <<"rcSltRsegPC: Event was a SLT BYPASS event!"<<endl;
#endif
  }
  if (!(stat.status&SLTEVENT_ACCEPT)) {
#ifdef DEBUG
    cout <<"rcSltRsegPC: Event rejected"<<endl;
#endif
    return -1;
  }
#ifdef DEBUG
  cout<<"rcSltRsegPC: Event accepted with "<<stat.ntracks<<" tracks in ";
  cout<<stat.ntriggers<<" trigger pairs"<<endl;
#endif

  if (trigger >= stat.ntriggers) {
#ifdef DEBUG
    cout << "rcSltRsegPC: SLT trigger " << trigger << "was not found" << endl;
#endif
    return -1;
  }

  if (SLTTrigInfo(trigger,&trig)!=SLT_OK) {
#ifdef DEBUG
    cout<<"rcSltRsegPC: Error getting info on trigger number "<<trigger<<endl;
#endif
    return -1;
  }
  if (trig.status==SLTTRIG_REJECT) {
#ifdef DEBUG
    cout << endl <<"rcSltRsegPC: Trigger "<<trigger<<" rejected"<<endl;
#endif
    return -1;
  }
#ifdef DEBUG
  cout<<"rcSltRsegPC: Trigger "<<trigger<<" accepted"<<endl;
#endif
  
  if (track >= trig.ntracks) {
#ifdef DEBUG
    cout << "rcSltRsegPC: SLT track " << track << "was not found" << endl;
#endif
    return -1;
  }

  if (SLTTrackInfo(trig.trackindex[track],&trac)!=SLT_OK) {
#ifdef DEBUG
    cout<<"rcSltRsegPC: Error get info on track "<<trig.trackindex[track]<<endl;
#endif
    return -1;
  }
  switch(trac.status) {
  case SLTTRACK_ERROR:
#ifdef DEBUG
    cout<<"rcSltRsegPC: Error in data of track "<<track<<endl;
#endif
    break;
  case SLTTRACK_ACCEPT:
#ifdef DEBUG
    cout<<"rcSltRsegPC: Track "<<track<<" accepted:"<<endl;
#endif
    break;
  case SLTTRACK_REJECT_SILI:
#ifdef DEBUG
    cout<<"rcSltRsegPC: Track "<<track<<" not accepted L2Sili:"<<endl;
#endif
    break;
  case SLTTRACK_REJECT_OTHER:
#ifdef DEBUG
    cout<<"rcSltRsegPC: Track "<<track<<" rejected in OTR:"<<endl;
#endif
    break;
  }
  
  if(trac.status==SLTTRACK_ERROR) return -1;
  
  if(trac.status!=SLTTRACK_ACCEPT) return -1;
  
  if (SLTRefitTrackInfo(trig.trackindex[track],&refit)!=SLT_OK) {
    cout<<"rcSltRsegPC: Error getting Refit track for track";
    cout<<trig.trackindex[track]<<endl;
    return -1;
  }

#ifdef DEBUG
  cout<<"rcSltRsegPC:   Main Tracker track information:"<<endl;
  cout<<"rcSltRsegPC:     # of hits and chi square: ";
  cout<<refit.nhits<<" "<<refit.chisq<<endl;
  cout<<"rcSltRsegPC:     Slope and intercept in x: ";
  cout<<refit.tx<<" "<<refit.bx<<endl;
  cout<<"rcSltRsegPC:     Slope and intercept in y: ";
  cout<<refit.ty<<" "<<refit.by<<endl;
#endif
  
  float zpc1=700.;
  float xslt = refit.bx + zpc1*refit.tx;
  float yslt = refit.by + zpc1*refit.ty;
  
  float sig_x = 0.08;
  float sig_y = 12.*sig_x;
  float sig_tx = sig_x / 70.;
  float sig_ty = sig_tx*sig_y/sig_x;
  
  //=========================================================== Analyze RSEGs
  float chi_max = 1.e10;
  ArtePointer<RSEG> rseg0;

  for ( ArteTableIterator<RSEG> rsegIt; rsegIt.valid(); rsegIt.next() ) {
    if (rsegIt->cmp & Rsegc::bitsign) continue;
    if (! (rsegIt->cmp & (Rsegc::wch | Rsegc::msg))) continue;
    if (! (rsegIt->cmp & Rsegc::patt)) continue;
    
    float x1,y1,tx1,ty1;
    rcRsegZ(rsegIt, zpc1, &x1, &y1, &tx1, &ty1);
    
    // Calculate chi squared at zpc1=700
    float chi_x  = pow((x1-xslt)/sig_x,2);
    float chi_y  = pow((y1-yslt)/sig_y,2);
    float chi_tx = pow((tx1-refit.tx)/sig_tx,2);
    float chi_ty = pow((ty1-refit.ty)/sig_ty,2);
    
    float chi2 = chi_x + chi_y + chi_tx + chi_ty;
    if (chi2 < chi_max) {
      chi_max = chi2;
      rseg0 = rsegIt;
    }
  }
  
  if (chi_max == 1.e10) return -2;

  *chi = chi_max;
  *rseg = rseg0;
  return 0;
}
