
// ****************************************************************************
//
// Routine:      rcStandAlone
// Author:       T.Oest/D.Samtleben
//
// Description:  stand alone reconstruction
//
//               reco level 0: no reconstruction
//                          1: hitb only
//                          2: reconstruction only
//                          3: hitb + reconstruction
//
// Modification:  
// drop RICH hits
// recoTARGET,... variables no longer static                        27/05/99 TO
// VDS==SVD for kumac-steering error-message if given subdetector
// is not known                                                     10/06/99 DS
// if at least one reconstruction package of a subdetector
// is called the Rxxx-tables are cleared                            01/07/99 DS
// bug fix: proper termination of input[]
// include CATS: recoVDSmodule
// correct calling sequenz for rc99ATARGET                          26/07/99 TO
// global alignment transformation added                            06/09/99 TO
// RICH hits are no longer removed since they are needed for
// the particle ID                                                  14/09/99 TO 
// move VDS global alignment transformation to rc99AVDS             20/09/99 TO
// filling of GTAR bank moved to acReadTARGET                       24/09/99 TO
// include global VDS alignment transformation                      29/10/99 TO
// ITR added                                                        07/02/00 TO
// use ARTE_mark()                                                  12/05/00 AG
// introduce timing                                                 17/05/00 AG
// timing info into globals                                         18/05/00 AG
// creation of chrono jobs only once per job                        24/05/00 DS
// initialize timers to 0                                           07/06/00 AG
// use more ARTE_mark()                                             08/06/00 AG
// polishing                                                        16/06/00 AG
// split ECAL reco in stand-alone and matching part                 21/07/00 AG
// added debug flag                                                 26/07/00 AG
// initialized timers correctly                                     27/07/00 AG
// improved printout                                                08/08/00 AG
// added more debug printout for OTR and RICH cuts                  25/08/00 AG
// introduced rcRANGER and removed ITR and OTR                      13/11/00 AG
// changed the muon reconstruction steering                         29/11/00 BF
// polished printout                                                29/11/00 AG
// polished printout                                                01/12/00 AG
// polished printout                                                06/12/00 AG
// moved rc99AStandAlone to rcStandAlone                            02/01/01 AG
// changed time names                                               02/01/01 AG
// cured typo                                                       03/01/01 AG
//
// ****************************************************************************

// standard includes
#include <iostream.h>
#include <stdlib.h>
#include <stdio.h>

// ARTE includes
#include "arte/arte.hh"
#include "arte/Chrono.hh"
#include "arte/ChronoManager.hh"
#include "farm/farm.hh"
#include "reco/reco.hh"
#include "reco/rcflags.hh"
#include "rang/zfpar.hh"
#include "rise/RingSearchAttrib.h"

// timers
float time_rcVDS    = 0;
float time_rcHPT    = 0;
float time_rcRANGER = 0;
float time_rcRICH   = 0;
float time_rcTRD    = 0;
float time_rcECAL   = 0;
float time_rcECAL2  = 0;
float time_rcMUON   = 0;

// externals
extern int FATIME;
extern int FADEB;

// prototypes
void acRSEGinGlobalCoordinates( const int CMPnumber );

// reco
void            pt2hitb();
int             rcRANGER( int recoITR, int recoOTR );
extern "C" void trdevn();
extern "C" void careevnt1_();
extern "C" void careevnt2_();
extern "C" void musini_();
extern "C" void mureca_();
void            matchRsegCare();
int             rcModSel();

//********************************************************************
// rcStandAlone
//********************************************************************
int
rcStandAlone()
{
  // mark
  ARTE_mark( "rcStandAlone 0", 0 );
  
  //===================================================================
  //     remove the Rxxx tables (for real data)
  //===================================================================
  
  //------  remove Rxxx-tables
  if( recoVDS  ||
      recoITR  ||
      recoHPT  || 
      recoOTR  ||
      recoRICH || 
      recoTRD  || 
      recoECAL || 
      recoMUON    ){
    ArteTable<RREL>::clear();
    ArteTable<RPNT>::clear();
    ArteTable<RVER>::clear();
    ArteTable<RHIT>::clear();
    ArteTable<RSEG>::clear();
    ArteTable<RTRA>::clear();
    ArteTable<RCCL>::clear();
  }

  //===================================================================
  //     initializations
  //===================================================================
  
  // timers
  static Chrono *cron_rcVDS    = 0;
  static Chrono *cron_rcHPT    = 0;
  static Chrono *cron_rcRANGER = 0;
  static Chrono *cron_rcRICH   = 0;
  static Chrono *cron_rcTRD    = 0;
  static Chrono *cron_rcECAL   = 0;
  static Chrono *cron_rcECAL2  = 0;
  static Chrono *cron_rcMUON   = 0;

  static bool first = 1;
  if( first ){
    first = 0;
    ChronoManager CM;
    CM.instance();
    
    cron_rcVDS    = CM.createChrono( "rcVDS"   );
    cron_rcHPT    = CM.createChrono( "rcHPT"   );
    cron_rcRANGER = CM.createChrono( "rcRANGER"  );
    cron_rcRICH   = CM.createChrono( "rcRICH"  );
    cron_rcTRD    = CM.createChrono( "rcTRD"   );
    cron_rcECAL   = CM.createChrono( "rcECAL"  );
    cron_rcECAL2  = CM.createChrono( "rcECAL2" );
    cron_rcMUON   = CM.createChrono( "rcMUON"  );
  }

  // initialization
  int VDSflag    = 0;
  int RANGERflag = 0;
  int RICHflag   = 0;

  //===================================================================
  //     stand-alone modules
  //===================================================================

  // VDS
  time_rcVDS = 0;
  if( recoVDS ){
    ARTE_mark( "rcVDS", 0 );
    cron_rcVDS->start();
    VDSflag = rcVDS( recoVDS );
    cron_rcVDS->stop();
    time_rcVDS = cron_rcVDS->lastTime();
  }

  // ECAL
  time_rcECAL = 0;
  if( recoECAL ){
    ARTE_mark( "careevnt1_", 0 );
    cron_rcECAL->start();
    careevnt1_();
    cron_rcECAL->stop();
    time_rcECAL = cron_rcECAL->lastTime();
  }

  //==================================================================
  // module selection
  //==================================================================
  ARTE_mark( "rcModSel" , 0 );
  int flagModSel = rcModSel();
  if( FADEB ) cout << " rcStandAlone: flagModSel=" << flagModSel << endl;

  // HPT
  time_rcHPT = 0;
  if( recoHPT ){
    ARTE_mark( "rcHPT", 0 );
    cron_rcHPT->start();
    pt2hitb();
    cron_rcHPT->stop();
    time_rcHPT = cron_rcHPT->lastTime();
  }

  // RANGER
  time_rcRANGER = 0;
  if( recoITR || recoOTR ){
    if( FADEB ) cout << " rcStandAlone: OTR: qpmax=" << zfpar_.qpmax << endl;
    ARTE_mark( "rcRANGER", 0 );
    cron_rcRANGER->start();
    RANGERflag = rcRANGER( recoITR, recoOTR );
    cron_rcRANGER->stop();
    time_rcRANGER = cron_rcRANGER->lastTime();
  }

  // RICH
  time_rcRICH = 0;
  if( recoRICH ){
    if( FADEB ) cout << " rcStandAlone: RICH: " 
		     << "occupancyCut="     << RingSearchAttrib::overwriteOccupancyCut << " "
		     << "seedOccupancyCut=" << RingSearchAttrib::seedOccupancyCut      << endl;
    ARTE_mark( "rcRICH", 0 );
    cron_rcRICH->start();
    RICHflag = rcRICH(recoRICH);
    cron_rcRICH->stop();
    time_rcRICH = cron_rcRICH->lastTime();
  }

  // TRD
  time_rcTRD = 0;
  if( recoTRD ){
    ARTE_mark( "rcTRD", 0 );
    cron_rcTRD->start();
    trdevn();
    cron_rcTRD->stop();
    time_rcTRD = cron_rcTRD->lastTime();
  }

  // MUON
  time_rcMUON = 0;
  musini_();
  if( recoMUON ){
    ARTE_mark( "rcMUON", 0 );
    cron_rcMUON->start();
    int one = 1;
    mureca_( &one, &one );
    cron_rcMUON->stop();
    time_rcMUON = cron_rcMUON->lastTime();
  }

  //===================================================================
  //     transformation of segments/RHIT to global coordinates
  //===================================================================  
  ARTE_mark( "acRSEGinGlobalCoordinates", 0 );
  if( recoVDSmodule & 1 ) acRSEGinGlobalCoordinates( 1 );
  if( recoVDSmodule & 2 ) acRSEGinGlobalCoordinates( 65537 );
  if( recoRICH )          acRSEGinGlobalCoordinates( 67108864 );
  
  //===================================================================
  //     ECAL needs tracks and trafo for the 2. reconstruction step
  //===================================================================
  time_rcECAL2 = 0;
  if( recoECAL ){
    ARTE_mark( "rcECAL2", 0 );
    cron_rcECAL2->start();
    //ECALflag = rcECAL(recoECAL); // 1. and 2. step
    //ARTE_mark( "careevnt1_", 0 );
    //careevnt1_();    // ECAL 1
    ARTE_mark( "matchRsegCare", 0 );
    matchRsegCare(); // Select the tracks (RSEG entries) for VDS-OTR
    ARTE_mark( "careevnt2_", 0 );
    careevnt2_();    // ECAL reconstruction using tracks
    cron_rcECAL2->stop();
    time_rcECAL2 = cron_rcECAL2->lastTime();
  }

  //===================================================================
  //     print
  //===================================================================
  if( FATIME ){

    float sum =
      time_rcVDS    +
      time_rcHPT    +
      time_rcRANGER +
      time_rcRICH   +
      time_rcTRD    +
      time_rcECAL   +
      time_rcECAL2  +
      time_rcMUON;
    
    printf( "\033[1m TIME: vds=%4.2f hpt=%4.2f track=%4.2f rich=%4.2f trd=%4.2f ecal1=%4.2f ecal2=%4.2f muon=%4.2f sum=%4.2f \033[0m \n",
	    time_rcVDS,
	    time_rcHPT,
	    time_rcRANGER,
	    time_rcRICH,
	    time_rcTRD,
	    time_rcECAL,
	    time_rcECAL2,
	    time_rcMUON,
	    sum );
  }
  
  // mark
  ARTE_mark( "rcStandAlone 1", 0 );
  
  return 0;  
}
