#ifndef __SINGLESPHERICALMIRROR_H__
#define __SINGLESPHERICALMIRROR_H__

#include "rich/ricutil.h"

// ------------------------------------------------------------------------
/// description of a single spherical mirror
/**
  A single spherical mirror have to have a polygon as boundary.
  The number of corners is limited to six points. The mirror
  have SingleSphericalMirror's as neighbors on the left,right, top
  and bottom side. All top/bottom neighbors defining a column, the
  left/right neighbors a row. 
  */
class SphericalMirror;

class SingleSphericalMirror : public CoordinateSystem {

friend class SphericalMirror;
Table  refl;     // mirror reflectifity

//ArtePointer<CSSM> a;

char   nname[21];
char   ttype[9];


protected:
Vec3   coc;      // center of curvature (=center of sphere);
double rr;       // radius of the mirror curvature
double dd;       // radius of the smallest circle surounding the mirror
double xdim;     // horizontal dimension
int    r,c;      // row and column

Vec3   corner[6];
Vec3   nbound[6];
double dbound[6];
int    numcorner;

/// definition of the neighborhood. (not for public use)
SingleSphericalMirror  *top,*bottom,*left,*right; // neighbouring mirrors

/// set the serial number
  inline void name(const char* s) {
    size_t l=strlen(s);
    if (l>sizeof(nname)-1)l=sizeof(nname)-1;
    memcpy(nname,s,l);
    nname[l]='\0';
    }
/// set the mirror type 
  inline void type(const char* s) {
    size_t l=strlen(s);
    if (l>sizeof(ttype)-1)l=sizeof(ttype)-1;
    memcpy(ttype,s,l);
    ttype[l]='\0';
    }
// constructor (only for special purposes) 
  inline SingleSphericalMirror() { name("undefined");}

/// returns 1, if the photon hits the mirror
  int  inside(const Vec3& v);

public:

/// returns the center of curvature
  inline  const Vec3& CoC() const { return coc; } 
/// returns the row, where the mirror is placed
  inline       int     row()    const { return r; }
/// returns the column, where the mirror is placed
  inline       int     col()    const { return c; }
/// returns the type of the mirror (HEX,FIVE,...)
  inline       char*   type()               { return ttype;};
/// returns the name of the mirror (i.e. serial no)
  inline       char*   name()               { return nname;};
/// returns the radius of curvature
  inline double radius()      const  { return rr; }
/// returns half of the diameter of the mirror
  inline double halfdiameter() const { return dd; }
/// returns corner coordinate
  inline Vec3 Corner(int i) const { return TransformToOld(corner[i]); } 
/// returns number of corners
  inline  int numCorner() const { return numcorner; }


//@Man: reflection on the mirror
//@{
/// returns 0, if the photon hit the mirror, else  error code
  int  intersection(const StraightLine& p,double *path);
/// call reflect, hit inside mirror, reflexion loss. returns 0, if ok.
  int  reflect(Photon* p);
/// only used for the calculation of the ring center
  int  reflect(StraightLine* t);
/// returns the reflectivity-curve of the mirror
  inline const Table&  reflectivity() const { return refl; }
/// set the reflectivity curve
  inline       Table&  reflectivity()       { return refl; }
//@}

/// produce paw command to plot the mirror boundary
  void xyview(FILE* fp);
/// print the contents of the class to an ASCII file.
  void print(FILE *fp=stdout);
/// constructor reading the information from an ASCII-FILE. See also page:
/**
 constructor reading the information from an ASCII-FILE produce by'print'.
  In addition a Table providing the reflectivity curve has to be
  supplied using the second argument.
*/
  SingleSphericalMirror(ScriptFile& f,Table& reflectivy);
/// constructor using the ARTE-Table CSSM
  SingleSphericalMirror(ArtePointer<CSSM> a);
  };

#endif
