//--*-C++-*--
//
//  Created: Apr 04 2001
//
//  Modifications:
//
//   23 May, 2000 (DD) RRing::flag added
//
//
#ifndef _ROBJECT_H_
#define _ROBJECT_H_
#include "arte/arte.hh"
#include <vector.h>
#include <math.h>
#include "rich/PhiLambda.h"

#define TWOPI 6.283185307 // 2*pi
#define MPI   3.141592654 // pi
#define PI32  0.063493636 // 1/(2*pi)^(3/2)




/**
   RICH photon description
*/
class RHit {
public:
  /// Photon x/y coordinate in hit space
  float  x,y;
  /// Global assignment counter
  int    assigned;
  /// Assignment to a mc track
  int    mctrack;
  /// Detector plane
  int    plane;
  /// Hit cell size
  float  area;
  /// Hit cell sigma
  float  sigma;
  /// Hit cell weight
  float  weight;
  /// Index in global hit bank (e.g. HITB)
  int    index;
  /// Print
  inline void print() {
    cout<<"x/y="<<x<<"/"<<y<<" plane="<<plane<<" assign="<<assigned<<endl;
  }
};





/**
   Description of photon that belongs to ring ROI
*/
class RCorrHit {
public:
  /// Coordinates in hit space after optical corrections 
  float  X,Y;
  /// Pointer to original photon
  vector<RHit>::iterator ihit;
  /// Assignment flag to ring 
  bool   assigned;
  /// Background parameter due to other rings and random hits
  float  b;
  /// Background due to other rings
  float beta;
  /// Flux due to signal ring
  float flux;
  /// Weight with respect to signal ring
  float w;
  /// Print
  inline void print() {
    cout<<"x/y="<<X<<"/"<<Y<<"  old x/y="<<(*ihit).x<<"/"<<(*ihit).y<<endl;
  }
};






/**
   Structure for track description
*/
class RTrack {
public:
  /// Track coordinates
  float x,y,z;
  /// Track angles
  float tx,ty;
  /// Covariance matrix
  float cov[15];
  /// Q/p
  float ip;
  /// Coordinates in hit space
  float X,Y;
  /// Monte carlo index
  int   mctrack;
  /// Track index in a global bank (e.g. RTRA)
  int   index;
  /// Point index in a global bank (e.g. RPNT)
  int   point;
  /// Print
  inline void print() {
    cout<<"+-----------------------------------------------------"<<endl;
    cout<<"| x/y/z="<<z<<"/"<<y<<"/"<<z<<" tx/ty="<<tx<<"/"<<ty<<" ip="<<ip<<endl;
    cout<<"! X/Y="<<X<<"/"<<Y<<endl;
    cout<<"+-----------------------------------------------------"<<endl;
  }
  /// Impact x parameter at z=0
  inline float bx() { return x-z*tx; }
  /// Impact y parameter at z=0
  inline float by() { return y-z*ty; }
  /// Error in horizontal diraction
  inline float stx() { return (cov[5]>0) ? sqrt(cov[5]) : .001; }
  /// Error in vertical diraction
  inline float sty() { return (cov[9]>0) ? sqrt(cov[9]) : .001; }
};



/**
   List of possible hypothesis
*/
enum RHypothesis {
  ELECTRON,
  MUON,
  PION,
  KAON,
  PROTON,
  GHOST,
  MULTI_ELECTRON
};

/**
   Dummy function for shadowing, always returns 1.0
 */
float shadowDummy(float *track);

/**
   Class for ring description, contains
   initialization, fitting and final results
*/
class RRing {
public:
  /// Empty constructor
  RRing() { ; }
  /// Constructor from a track
  RRing (vector<RTrack>::iterator it) {
    itrack=it;
    X=it->X;
    Y=it->Y;
    accept = false;
    chi2   = 1e5;
    nIt    = 0;
    weight = 0;
    Delta  = 1e5;
    Dr     = 1e5;
    Dh     = 1e5;
    L      = -1e5;
    flag   = 1;
  }
  /// Copy constructor
  //RRing(const RRing &ring) {
  //memcpy(this,&ring,sizeof(ring));
  //}
  /// Ring center in hit space
  float X,Y;
  /// Ring radius
  float radius;
  /// Background Rb paramter from random flux
  float Rbs;
  /// Background Rb paramter from other rings
  float Rbb;
  /// Total photon flux thru ROI
  float flux;
  /// Total search area
  float SArea;
  /// Number of signal photons
  float S;
  /// Total weight of photons
  float weight;
  /// Signal/sqrt(background)
  float SB;
  /// Chi2 like quality of the fit
  float chi2;
  /// Ring accepted(true)
  bool  accept;
  /// Minimum radius^2 for hit acceptance into ROI
  float r2min;
  /// Maximum radius^2 for hit acceptance into ROI
  float r2max;
  /// Total photons in ROI
  int   nROI;
  /// ID likelihoods: e, mu, pi, k, p, multi e
  float lr[6];
  /// Fitting likelihood
  float L;
  /// Minimum distance from R2P2 line (in terms of sigmas)
  float Dh;
  /// Distance from hypothetical radius in terms of sigma^2
  float Dr;
  /// Brightness width (Delta=delta/sigma)
  float Delta;
  /// Sigma of the ring
  float sigma;
  /// Track reference
  vector<RTrack>::iterator itrack;
  /// Hypothetical radius
  float radiusH;
  /// Hypothetical mass
  float mass;
  /// Error^2 of the radius^2 for given hypothesis/mass
  float R2P2error2;
  /// Number of iterations used in fitting
  int   nIt;
  /// Fraction of photons inside acceptance
  float Fs;
  /// Ring flag
  int   flag;
  /// Print ring
  inline void print(int verb=0) {
    cout<<"+-----------------------------------------------------------"<<endl;
    cout<<"|  Mass = "<<mass<<"    p="<<1/itrack->ip<<endl;
    cout<<"|  Fit: X="<<X<<" Y="<<Y<<" R="<<radius<<"  Fs="<<Fs<<endl;
    if (verb==0) { // only initial conditions
      cout<<"|  Hyp: R="<<radiusH<<"  R2error2="<<R2P2error2<<endl;
      cout<<"|  ROI: R2min="<<r2min<<"  R2max="<<r2max<<endl;
    }
    if (verb==1) { // only results
      cout<<"|  L="<<L<<"   Dh="<<Dh<<"  Dr="<<Dr<<endl;
      cout<<"|  lr: e="<<lr[0]<<" mu="<<lr[1]<<" pi="<<lr[2]<<" K="<<lr[3]<<" p="<<lr[4]<<endl;
      cout<<"|  Res: accept="<<accept<<"  S="<<S<<"  Rb="<<Rbs<<"/"<<Rbb<<"  chi2="<<chi2<<endl;
      cout<<"|       n iter="<<nIt<<"  Weight="<<weight<<"  nROI="<<nROI<<endl;
    }
    cout<<"+-----------------------------------------------------------"<<endl;
  }
  /// Set R2P2 resolution ^2
  inline void setR2P2error2() {
    float ip          = itrack->ip;
    float sigma2_rr   = 1.5e-5 * 1.5e-5; /// this should come from RICH description!
    float aa          = .03*.03;         /// this from tracker dp/p
    float bb          = .0025*.0025;     ///    -ii-                   itrack->cov[14]*itrack->p*itrack->p
    float m2          = mass*mass;
    float ip2         = ip*ip;
    float sigma2_ipip = 4 * (ip2*ip2) * (aa+bb/ip2);
    R2P2error2        = sigma2_rr + m2*m2 * sigma2_ipip;
  }
  /// Set search area
  inline void setSArea() {
    SArea = (r2max-r2min)*MPI;
  }
  /// Pointer to shadowing/acceptance function
  static float (*shadow)(float *track);
  /// Set shadowing/acceptance function
  static void setShadowFunction(float (*myShadow)(float *track));
  /// Calculate acceptance using shadowing/acceptance function
  inline void setShadow() {
    float track[]={itrack->x, itrack->y, itrack->z, itrack->tx, itrack->ty, radius};
    Fs=shadow(track);
  }
};





/**
  Main parameters for RICH detector
*/
class RRICH {
public:
  /// N0T=(Number of photons)/(radiation length)/(Cherenkov angle)^2
  float N0T;
  /// Cherenkov angle for Beta=1 particles 
  float radius0;
};





/**
   Dummy class for object list filling (no action),
   can be replaced with user function.
*/
void fillListDummy();
void saveListDummy();


/**
   Template containers for RICH objects
*/
template<class T> class RichObject {
public:
  /// Constructor
  RichObject() { fillList = fillListDummy; saveList = saveListDummy; }
  /// Destructor
  ~RichObject() {;}
  /// List of tracks
  static vector<T> list;
  /// Use custom function to fill STL vector RichObject<T>::list 
  void                  setFillFunction(void (*myFillList)() ) {
    fillList = myFillList;
  }
  /// Print list
  void print() {
    for (vector<T>::iterator it=list.begin(); it!=list.end(); ++it) {
      it->print();
    }
  }
  /// Function pointer for filling list
  void                  (*fillList)();
  /// Clean up  list
  inline void clear() {
    if (list.size()) list.clear();
  }
  /// Set function pointer for saving list
  void setSaveFunction( void (*mySaveFunction)() ) { 
    saveList=mySaveFunction; 
  }
  /// Function pointer for saving list
  void (*saveList)();
};




#endif
