// ********************************************************************
//
//  Routine:      acReadVDS
//  Author:       T.Glebe, I.Kisel
//
//  Description:  call the VDS initialization to get the VDS constants
//                - if key not assigned, use run number as major number
//                - returns <0 error
//                           0 no VDS update necessary
//                           1 update done using key
//                           2 update done using date/time or run number
//                
//
//  Modification:	21.03.2000 - added code used by L2FORMARTE (fg sciacca)
//			02.04.2000 - fixed handling of keytable mechanism (fg sciacca)
//			11.04.2000 - bugs removed
//				   - L2FORMARTE code integrated with VDS code (fg sciacca)	
//			30.06.2000 - bu fix: now ok with unsparsified data (fg sciacca)
//                      29.11.2000 - (TG) call ARTE_exit() in case DB could not
//                                   be read (for reprocessing, request by A. Gellrich)
//			14.12.2000 - DB access now allows writing (pwd hardcoded) (fg sciacca)
//				     (for new and old LEDA compatibility - cmp. V. Vagnoni)
//			04.04.2001 - GEDE update and SVDstartup skipped when in MC mode
//				     (for use within SLT simulation) (fg sciacca)
// ********************************************************************
//!!!!!!!!!!!!!!!!!!!!!!
using namespace std;//!!
//!!!!!!!!!!!!!!!!!!!!!!

#include <stdio.h>

#include "arte/arte.hh"
#include "database/keytable.h"
#include "farm/farm.hh"
#include "grr.h"
#include "silarte/silarte.hh"
#include "vdsleda.h"

// sparse event definition
#include "event.h"
#include "etb.h"

// prototypes
int SvdStartup();
//int  vdsleda_fetchtable( char *filename, int major, int minor);
extern "C" {
  void fill_cats(void);
}

static int VDSserver =	-1;
static int VDStable =	-1;
static int VDStable_side =	-1;
static int VDStable_chip =	-1;
static int VDStable_bondmap =	-1;
static int VDStable_fed =	-1;
static int VDStable_chanstat =	-1;
static int VDStable_chipbond =	-1;
static int VDStable_chipchans =	-1;
static int VDStable_chipfedside =	-1;


// *****************************************************************
// function
// *****************************************************************

int acReadVDS(int KTstatus) 
{
  
//==================================================================
// no action for MC
//==================================================================
  
  //  if (ArteTable<MTRA>::size() > 0 ) return 0;

//==================================================================
// database server (default: /VDS_CONFIG)
//                  can be changed by setting DB_VDS
//==================================================================
 
  char defaultserver[13];
  sprintf(defaultserver,"/VDS_CONFIG");
  char * dbserver;
  
  static bool firsttime = true;

  if (getenv("DB_VDS") == NULL){
    dbserver = defaultserver;
  } else {
    dbserver = getenv("DB_VDS");
  };
 
  
  switch (KTstatus) {

    case 2: 
      // structure of keytable changed, so loop over the keytable and get 
      // your serverid and tableids 

      VDSserver = KeyFindServer(dbserver); 
      if (VDSserver == -1) { 
	ARTE_exit(GRR_SFATAL, "acReadVDS: Keytable not found!",0);
        cout<<"acReadVDS: KeyTable Server not found!"<<endl; 
        //return -1; 
        break;  // use old method
      }; 

//      if(firsttime) {
//        SvdStartup();
//        firsttime = false;
//      }
      cout << "acReadVDS: Looking for Keytable entries..." << endl;
      VDStable_side =		KeyFindTable(VDSserver, "VDS/Side"); 
      VDStable_chip =		KeyFindTable(VDSserver, "VDS/Chip"); 
      VDStable_bondmap =	KeyFindTable(VDSserver, "VDS/Bondmap"); 
      VDStable_fed =		KeyFindTable(VDSserver, "VDS/Fed"); 
      VDStable_chanstat =	KeyFindTable(VDSserver, "VDS/Chanstat"); 
      VDStable_chipbond =       KeyFindTable(VDSserver, "VDS/ChipBondmap");
      VDStable_chipchans =      KeyFindTable(VDSserver, "VDS/ChipChanstat");
      VDStable_chipfedside =    KeyFindTable(VDSserver, "VDS/ChipFed");

      if ((VDStable_side == -1) || (VDStable_chip == -1) || 
	  (VDStable_bondmap == -1) || (VDStable_fed == -1) || 
          (VDStable_chanstat == -1) || (VDStable_chipbond  == -1) ||
          (VDStable_chipchans == -1) || (VDStable_chipfedside  == -1) ) { 
	ARTE_exit(GRR_SFATAL, "acReadVDS: Keytable entries not found!",0);
        cout<<"acReadVDS: Keytable entries not found!"<<endl; 
	break;  // use old method
      }; 
      // NO _BREAK_ HERE! 

    case 1: 
      // an update happened, look if your server is involved 
      if (!KeyServerNew(VDSserver)) {
        cout<<"acReadVDS: no VDS update necessary, server not involved"<<endl;      
        cout<<"acReadVDS: returning with rc = 0"<<endl;
      return 0; //no VDS update necessary
      };

      // now for the tables - look whether they have changed 
      if (KeyTableNew(VDStable)) { 
        // using LEDA to get the Tables x:
        MizziDatabase(mydb);
        Tables x;

	cout<<"acReadVDS: keytable found: opening VDS database "<< VDSserver <<endl;
        mydb.open(KeyServerName(VDSserver),21186); //hardcode password for new and old LEDA compatibility
        x=mydb.get_id();

        if(VDS_Side     ->book(mydb) == 0) {
          ARTE_exit(GRR_SFATAL, "acReadVDS: VDS_Side table not found!",0); 
        }
        if(VDS_Chip     ->book(mydb) == 0) {
          ARTE_exit(GRR_SFATAL, "acReadVDS: VDS_Chip table not found!",0);
        }
        if(VDS_Bondmap  ->book(mydb) == 0) {
          ARTE_exit(GRR_SFATAL, "acReadVDS: VDS_Bondmap table not found!",0);
        }
        if(VDS_Fed      ->book(mydb) == 0) {
          ARTE_exit(GRR_SFATAL, "acReadVDS: VDS_Fed table not found!",0);
        }
        if(VDS_Chanstat ->book(mydb) == 0) {
          ARTE_exit(GRR_SFATAL, "acReadVDS: VDS_Chanstat table not found!",0);
        }

        VDS_Side	->remove-> all();
        VDS_Chip	->remove-> all();
        VDS_Bondmap	->remove-> all();
        VDS_Fed		->remove-> all();
        VDS_Chanstat	->remove-> all();


        cout<<"acReadVDS: reading VDS/Side information"<<endl;
        KeyTableStamp(x,KeyServerName(VDSserver),KeyTableName(VDStable_side));
        VDS_Side	->fetch_data( DBeq );
        VDS_Side	->set_keys();

        cout<<"acReadVDS: reading VDS/Chip information"<<endl;
        KeyTableStamp(x,KeyServerName(VDSserver),KeyTableName(VDStable_chip));
        VDS_Chip	->fetch_data( DBeq );
        VDS_Chip	->set_keys();

        cout<<"acReadVDS: reading VDS/Bondmap information"<<endl;
        KeyTableStamp(x,KeyServerName(VDSserver),KeyTableName(VDStable_bondmap));
        VDS_Bondmap	->fetch_data( DBeq );
        VDS_Bondmap	->set_keys();
  
        cout<<"acReadVDS: reading VDS/Fed information"<<endl;
        KeyTableStamp(x,KeyServerName(VDSserver),KeyTableName(VDStable_fed));
        VDS_Fed		->fetch_data( DBeq );
        VDS_Fed		->set_keys();

        cout<<"acReadVDS: reading VDS/Chanstat information"<<endl;
        KeyTableStamp(x,KeyServerName(VDSserver),KeyTableName(VDStable_chanstat));
        VDS_Chanstat	->fetch_data( DBeq );
        VDS_Chanstat	->set_keys();

	VDS_Side	->fetch_relations(DBeq); /* does not load any table */
	VDS_Fed		->fetch_relations(DBeq); /* does not load any table */ 
        KeyTableStamp(x,KeyServerName(VDSserver),
                        KeyTableName(VDStable_chipbond ));
	VDS_Bondmap	->fetch_relations(DBeq);
        KeyTableStamp(x,KeyServerName(VDSserver),
                        KeyTableName(VDStable_chipchans ));
	VDS_Chanstat	->fetch_relations(DBeq);
        KeyTableStamp(x,KeyServerName(VDSserver),
                        KeyTableName(VDStable_chipfedside ));
	VDS_Chip	->fetch_relations(DBeq); /* loads fed and side relations */

        VDS_Side	->unbook();
        VDS_Chip	->unbook();
        VDS_Bondmap	->unbook();
        VDS_Fed		->unbook();
        VDS_Chanstat	->unbook();
        cout<<"acReadVDS: VDS tables loaded: closing VDS database "<< VDSserver <<endl;

	// Only update GEDE and initialize Marvin in Data mode; skip if MC
	if (ArteTable<MTRA>::size() == 0 ){ 
	  silarte_updateGEDE();

	  fill_cats(); //fill cats with new geometry
	  if(firsttime) {		//silraw and marvin init must occurr after DB access
	    SvdStartup();
	    firsttime = false;
	  }
	}
	
	cout<<"acReadVDS: returning with rc = 1"<<endl;
        return 1; // update using keytable
      }  

      //cout<<"acReadVDS: no VDS update necessary"<<endl;
      cout<<"acReadVDS: returning with rc = 0"<<endl;
      return 0;  // no VDS update necessary

      //break; 
  } // of switch 

  // No keytable information available, using old method...

    if(firsttime) {
    cout << "acReadVDS: Key NOT assigned --> using old method" << endl; 
//    SvdStartup();
//    firsttime = false;
    }

  static int VDS_Side_major = -1;
  static int VDS_Side_minor = -1;
  static bool findmajmin = true;
  static MizziDatabase( vdsdb_setup );

  // get event pointer
  EVENT_HDR *event = ARTE_get_event();
  if ( !event ) {
    cout << "acReadVDS: Could not find event!" << endl;
    return -1; // error
  }      

  //    Get constants from database
  //
  // Time stamp is evaluated in this order:
  //  1. date/time   from vdsleda IOL
  //  2. major/minor from vdsleda IOL
  //  3. event time from DDAQ (if all IOL parameters are =0)
  int major = 0;
  int minor = 0;
  int ndaq;
  
  EVT_DATA_RECORD daq;
  ETB_get_daqs( event, 1, &daq, &ndaq );
  if ( ndaq > 0 ) {
    major = daq.data.event_time;
    minor = 0;
  }

  if(VDS_Side_major == -1 || 
     (major > VDS_Side_major && !findmajmin)) {

    vdsleda_fetchdata(major, minor);
    // Only update GEDE and initialize Marvin in Data mode; skip if MC
    if (ArteTable<MTRA>::size() == 0 ){ 
      silarte_updateGEDE();
      
      fill_cats(); //fill cats with new geometry

      if(firsttime) {		//silraw and marvin init must occurr after DB access
	SvdStartup();
	firsttime = false;
      }
    }
  } else {
      cout<<"acReadVDS: OLDMETHOD; no VDS update necessary"<<endl;
      cout<<"acReadVDS: OLDMETHOD; returning with rc = 0"<<endl;
      return 0; // no update necessary
  }

  // find major minor until which dataset is valid
  if(VDS_Side_major == -1  || 
     findmajmin            ||
     major > VDS_Side_major ) {

    //    vdsdb_setup.open( vdsleda_iol.setup_database.value, 0 );
    vdsdb_setup.open( dbserver, 0 );
    vdsdb_setup.stamp( major, minor );
    VDS_Side    ->book( vdsdb_setup );
    if( VDS_Side->match( DBgt ) == 1 ) {
      // successfull match
      VDS_Side_major = TableMajor( vdsdb_setup.get_id() );
      VDS_Side_minor = TableMinor( vdsdb_setup.get_id() );
      findmajmin = false;
    } else {
      //if there is no newer table, lookup would take place in each event!
      //      findmajmin = true;
      VDS_Side_major = major + 200;
      findmajmin = false;
    }

    VDS_Side->unbook();
    vdsdb_setup.close();
  }

  cout<<"acReadVDS: returning with rc = 2"<<endl;
  return 2;  // constants read in using old method 
}
