#ifndef __VTVERTEX_HH
#define __VTVERTEX_HH
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   21. Aug 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Vertex class
//
// changes:
// 21 Aug 2000 (TG) creation
// 25 Aug 2000 (TG) remove_worst added
// 04 Sep 2000 (TG) member function ndf() added. chi2() returns now chi2
//                  and not chi2/ndf any more
// 04 Sep 2000 (TG) Modified argument of VtRemoveTrack()
// 05 Sep 2000 (TG) declared some member functions as inline
// 05 Sep 2000 (TG) erase() member function added
// 12 Sep 2000 (TG) private member function bigdim() added
// 12 Sep 2000 (TG) member function bigcov() added
// 13 Sep 2000 (TG) added xmass member and add() member function
// 14 Sep 2000 (TG) added v_CINV member
// 14 Sep 2000 (TG) bug fix in documentation
// 02 Okt 2000 (TG) added chi2l() member function
// 04 Okt 2000 (TG) removed v_use_momentum data member
// 17 Okt 2000 (TG) modified comments
// 18 Okt 2000 (TG) MassConstr() member function added
// 24 Okt 2000 (TG) invalid() member function added, v_covn, v_Mvalid member added,
//                  changed return type of bigcov()
// 25 Okt 2000 (TG) covn() member function added
// 26 Okt 2000 (TG) #ifndef NO_ARTE added (for integration into CLUE)
//
// *****************************************************************************

#include <iosfwd>
#include <vector>
#include "CMatrix.hh"
#include "VtVector.hh"
#include "VtTrack.hh"
#include "VtMassC_list.hh"

template <class T> class ArtePointer;
#ifndef NO_ARTE
class RTRA;
class RSEG;
class MTRA;
class RVER;
class TOPP;
#endif // NO_ARTE

namespace VERTEX {
  //
  // Some Declarations
  //
  class Track;
  class Vertex;

  typedef std::vector<double>                 Vector_d;

  typedef std::vector<Track*>                 Track_v;
  typedef std::vector<Track*>::iterator       Track_it;
  typedef std::vector<Track*>::const_iterator Track_cit;

  /** Vertex class */
  //============================================================================
  // Class Vertex
  //============================================================================
  class Vertex: public Track {
  public:
    
    /**@name Constructors */
    ///
    Vertex();
    ///
    Vertex(const Vertex& rhs);
#ifndef NO_ARTE
    ///
    Vertex(const ArtePointer<RVER>& v);
    ///
    Vertex(const ArtePointer<RTRA>& r1, const ArtePointer<RTRA>& r2);
    ///
    Vertex(const ArtePointer<RSEG>& r1, const ArtePointer<RSEG>& r2);
    ///
    Vertex(const ArtePointer<MTRA>& m1, const ArtePointer<MTRA>& m2);
#endif // NO_ARTE
    ///
    Vertex(const vector<Track*>& v);
    ///
    Vertex(Track& t1, Track& t2);
    ///
    ~Vertex();

    /**@name Access functions */
    /// $x$ of vertex
    inline double vx()      const { return kal_xvs[0];  }
    /// $y$ of vertex
    inline double vy()      const { return kal_xvs[1];  }
    /// $z$ of vertex
    inline double vz()      const { return kal_xvs[2];  }
    /// available after call to rmsDistAngle()
    inline double angle()   const { return v_angle; }
    /// available after call to rmsDistAngle()
    inline double dist()    const { return v_dist;  } 
    /// no of tracks in vertex $n$
    inline int    ntracks() const { return size(); }
    /// $\chi^2$ of vertex fit
    inline double chi2()    const { return v_chi2; }
    /// degrees of freedom of vertex fit
    inline int    ndf()     const { return 2*ntracks()-3; }
    /// $\sqrt{\sigma_{vx}^2}$ vertex $x$-error
    inline double vxerr()   const { return sqrt(v_CS(0,0)); }
    /// $\sqrt{\sigma_{vy}^2}$ vertex $y$-error
    inline double vyerr()   const { return sqrt(v_CS(1,1)); }
    /// $\sqrt{\sigma_{vz}^2}$ vertex $z$-error
    inline double vzerr()   const { return sqrt(v_CS(2,2)); }
    /// $\sigma_{vx}^2$
    inline double vxerr2()  const { return v_CS(0,0); }
    /// $\sigma_{vy}^2$
    inline double vyerr2()  const { return v_CS(1,1); }
    /// $\sigma_{vz}^2$
    inline double vzerr2()  const { return v_CS(2,2); }

    /// $3\times3$ Vertex covariance matrix
    inline const MATRIX::VtSymMatrix&  CS()   const { return v_CS; }
    /// $(3\cdot n+3)\times(3\cdot n+3)$ general covariance matrix
    inline const MATRIX::VtSymMatrix&  covn() const { return *v_covn; }
    /// is vertex valid?
    bool   valid()   const { return v_valid; }

    /** @name Mass calculation routines 
	@memo $m = \sqrt{\sum_i p_i^2 + m_i^2 - \sum_i \vec{p}_i^2}$
     */
    /// rest-mass 0, 
    double mass(const bool use = false) const;
    /// single rest-mass calc.,
    double mass(double m1) const;
    /// 2-track mass with particle rest-masses
    double mass(double m1, double m2) const;
    /// 3-track mass with rest-masses
    double mass(double m1, double m2, double m3) const;
    /// 4-track mass with rest-masses
    double mass(double m1, double m2, double m3, double m4) const;
    /// specify $n$ rest masses
    double mass(const Vector_d& m02, const bool use = false)  const;

    /**@name Vertexing routines */
    /// unweighted 2-D min. dist.
    const bool findVertex2D(void);
    /// unweighted 3-D min. dist.
    const bool findVertex3D(void);
    /// Vt Kalman-filter
    const bool findVertexVt    ();
    /// weighted 2-D min. dist.
    const bool VtEstimateVertex();
    /// Kalman filter with mass constraints
    const bool VtMass();

    /**@name Adding tracks */
    ///
    void   push_back(Track& track);
#ifndef NO_ARTE
    ///
    void   push_back(const ArtePointer<RTRA>& rtra);
    ///
    void   push_back(const ArtePointer<RSEG>& rseg);
    ///
    void   push_back(const ArtePointer<MTRA>& mtra);
#endif // NO_ARTE

    /**@name Misc functions */
#ifndef NO_ARTE
    /// write to RVER table
    const ArtePointer<RVER> createRVER(void) const;
#endif // NO_ARTE
    /// calc rms dist and rms angle
    const double rmsDistAngle();
    /// calc track params of mother particle
    bool calc_mother();
    /// calc cov. matrix of mother particle
    bool calc_mother_cov();

    /// remove last track from vertex, do inverse Kalman filter
    const bool remove_last();
    /// remove track with biggest $\chi^2$ contribution, do inverse Kalman filter
    void remove_worst();
    /// set for all tracks whether momentum should be used or not
    void  use_momentum(const bool use);
    /// use refitted track parameters or not
    inline void  use_kalman(const bool use) { v_use_kalman = use; }
    ///
    inline bool  use_kalman() const { return v_use_kalman; }
    /// $\chi^2$ distance to space point, $ndf = 3$
    double distance(double x, double y, double z) const;
    /// $\chi^2$ distance to track, $ndf = 2$
    double distance(Track& t) const;
    /// $\chi^2$ distance to vertex, $ndf = 3$
    double distance(const Vertex& rhs) const;

    /** @name Expert routines */
    ///
    const bool VtFilter        ();
    ///
    const bool VtInverseFilter () const;
    /// smooth vertex position
    void       VtSmoothX       ();
    /// smooth momenta
    void       VtSmoothQ       ();
    /// remove track using inverse Kalman filter
    const bool VtRemoveTrack   (Relation& track);
    /// Kalman filter + smoother
    const bool VtFit           ();
    ///
    const MATRIX::VtVector& xv()  const { return kal_xv; }
    /// vector of vertex-positions
    const MATRIX::VtVector& xvs() const { return kal_xvs; }
    /// called by operator<<()
    void   print(std::ostream& os) const;
    /// $\chi^2$ after last filter step
    double chi2n() const;
    /// $\chi^2$ contribution of last filter step
    double chi2l() const;
    /// unbook relation
    virtual void remove(Relation* r);
    /// erase a relation without refit
    virtual const iterator erase(const iterator& pos);
    /// compute general $(3\cdot n+3) \times (3\cdot n+3)$ covariance matrix 
    const MATRIX::VtSymMatrix& bigcov();

    /** @name Mass Constraint routines */
    /// add a mass constraint for all tracks, $m$ = mass of mother particle
    void MassConstr(double m);
    /// add a mass constraint
    MassC& addMassConstr(double m=0.);
    /// delete the mass constraints
    void clearMassConstr();
    /// number of mass constraints
    inline unsigned int nMassConstr() { return xmass.size(); }

    //
    // Overloaded operators
    //
    const Vertex& operator= (const Vertex& rhs);
    const bool    operator==(const Vertex& rhs) const;
    const double  operator- (const Vertex& rhs) const;

  private:
    inline const unsigned int bigdim() const { return 3 * ntracks() + 3; }
    // flag vertex as invalid 
    void invalid();                       
    
    bool   v_use_kalman;                  // use refitted track parameters
    bool   v_valid;                       // is vertex information valid?
    bool   v_Mvalid;                      // is mass constrained vertex information valid?
    bool   v_mother;                      // is the mother particle calculated?
    double v_angle;                       // rms of opening angle
    double v_dist;                        // rms of distance to vertex
    double v_bk13;                        // vtx-z at begin of iteration step
                                          // (in Vt: Bk13(iv), b13)
    double v_chi2;                        // chi2 of vertex-fit
    MATRIX::VtVector     kal_xv;          // x,y,z of vtx from kalman filter
    MATRIX::VtVector     kal_xvs;         // pos of Vertex after smooth
                                          // (xv-smoothed) (in Vt: xvs(1..3,...))
    MATRIX::VtSymMatrix  v_CS;            // v_C smoothed (=v_C of last
                                          // filtered track) (in Vt: CS)
    MATRIX::VtSymMatrix  v_CINV;          // inverse of CS at initialization time
    MATRIX::VtSymMatrix* v_covn;          // (3*n+3)x(3*n+3) big cov. matrix
    Track_v    tracks;                    // tracks owned by vertex
    MassC_v    xmass;                     // vector of mass constraint pointer
  }; // class Vertex


  //==============================================================================
  // operator<<
  //==============================================================================
  inline std::ostream& operator<<(std::ostream& s, const Vertex& a) {
    a.print(s);
    return s;
  }

} // namespace VERTEX
#endif
